unit DiskName;

interface

uses Windows, SysUTils, ComCtrls, Classes, Util;

function GetDriveManufacturerAndModel(DriveID: Byte; var serialnostr:string; List: TListView=nil): String;

implementation

uses main;

const
  MAX_IDE_DRIVES              = 16;
  IOCTL_STORAGE_QUERY_PROPERTY = $2D1400;

type
  STORAGE_PROPERTY_QUERY = packed record
    PropertyId: DWORD;
    QueryType: DWORD;
    AdditionalParameters: array[0..3] of Byte;
  end;

  STORAGE_DEVICE_DESCRIPTOR = packed record
    Version: short;
    Size: short;
    DeviceType: Byte;
    DeviceTypeModifier: Byte;
    RemovableMedia: Boolean;
    CommandQueueing: Boolean;
    VendorIdOffset: ULONG;
    ProductIdOffset: ULONG;
    ProductRevisionOffset: ULONG;
    SerialNumberOffset: ULONG;
    STORAGE_BUS_TYPE: DWORD;
    RawPropertiesLength: ULONG;
    RawDeviceProperties: array[0..511] of Byte;
  end;

  DUMMYDESCRIPTOR = packed record
    Version: short;
    Size: short;
    DeviceType: Byte;
    DeviceTypeModifier: Byte;
    RemovableMedia: Boolean;
    CommandQueueing: Boolean;
    VendorIdOffset: ULONG;
    ProductIdOffset: ULONG;
    ProductRevisionOffset: ULONG;
    SerialNumberOffset: ULONG;
    STORAGE_BUS_TYPE: DWORD;
    RawPropertiesLength: ULONG;
  end;


function ByteToChar(const B: Byte): char;
begin
  Result := Chr(B + $30)
end;

function SerialNumberToCardinal (SerNum: String): Cardinal;
begin
  HexToBin(PChar(SerNum), PChar(@Result), SizeOf(Cardinal));
end;

function SerialNumberToString(SerNum: String): String;
var
  I, StrLen: Integer;
  Pair: string;
  B: Byte;
  Ch: Char absolute B;

begin
  Result := '';
  StrLen := Length(SerNum);

  if Odd(StrLen) then Exit;

  I := 1;

  while I < StrLen do
  begin
    Pair := Copy (SerNum, I, 2);
    HexToBin(PChar(Pair), PChar(@B), 1);
    Result := Result + Chr(B);
    Inc(I, 2);
  end;

  I := 1;

  while I < Length(Result) do
  begin
    Ch := Result[I];
    Result[I] := Result[I + 1];
    Result[I + 1] := Ch;
    Inc(I, 2);
  end;
end;

function GetStringFromOffset(buffer:array of byte; offset,len:integer): string;
var i:integer;
begin
    result:='';
    if offset=0 then exit;
    for i := Offset to Offset+len do
    begin
      //if buffer[i] = 0 then break;
      if buffer[i] >= 32 then
      begin
        //if (result='') or
        //   ((result<>'') and (result[length(result)-1]<>' '))
        //then
        result:=result+chr(buffer[i]);
      end;
    end;
    result:=trim(result);
    result:=ReplaceAllStringsInString(result,'  ',' ');
end;

function HexTwins2String(s:string;IgnoreSpaces:boolean):string;
var i, l, z1, z2: longint;
begin
    result:='';
    l:=Length(s);
    i:=1;
    while i<=l do
    begin
      z1:=StrToIntDef('$'+Copy(s,i+2,2),0);
      z2:=StrToIntDef('$'+Copy(s,i,2)  ,0);
      if z1>20 then
      result:=result+Chr(z1);
      if z2>20 then
      result:=result+Chr(z2);
      Inc(i,4);
    end;
end;

function GetDriveManufacturerAndModel(
  DriveID: Byte; var serialnostr:string;
  List: TListView=nil): String;
type
  PCharArray = ^TCharArray;
  TCharArray = array[0..32767] of ansichar;

var
  DriveNumber: Byte;
  FFileHandle: Cardinal;
  FProductRevision: string;
  FProductId: string;
  FSerialNumber: string;
  FVendorId: string;

  Returned: Cardinal;
  Status: LongBool;
  PropQuery: STORAGE_PROPERTY_QUERY;
  DeviceDescriptor: STORAGE_DEVICE_DESCRIPTOR;
  dummy: DUMMYDESCRIPTOR;
  PCh: PChar;

begin
  FFileHandle := INVALID_HANDLE_VALUE;
  try
    FFileHandle := CreateFile(
                     PChar('\\.\PhysicalDrive' + ByteToChar(DriveID)),
                     0,
                     FILE_SHARE_READ or FILE_SHARE_WRITE,
                     nil,
                     OPEN_EXISTING,
                     0,
                     0
                   );

    if FFileHandle = INVALID_HANDLE_VALUE then
      RaiseLastOSError;

    ZeroMemory(@PropQuery, SizeOf(PropQuery));
    ZeroMemory(@DeviceDescriptor, SizeOf(DeviceDescriptor));

    DeviceDescriptor.Size := SizeOf(DeviceDescriptor);

    Status := DeviceIoControl(
                FFileHandle,
                IOCTL_STORAGE_QUERY_PROPERTY,
                @PropQuery,
                SizeOf(PropQuery),
                @DeviceDescriptor,
                DeviceDescriptor.Size,
                Returned,
                nil
              );

    if not Status then
      RaiseLastOSError;

    DeviceDescriptor.VendorIdOffset:=1;
    while DeviceDescriptor.RawDeviceProperties[DeviceDescriptor.VendorIdOffset]<32 do
      inc(DeviceDescriptor.VendorIdOffset);

    DeviceDescriptor.ProductIdOffset:=DeviceDescriptor.VendorIdOffset+16;

    if DeviceDescriptor.VendorIdOffset <> 0 then
    begin
      PCh := @PCharArray(@DeviceDescriptor)^[DeviceDescriptor.VendorIdOffset];
      FVendorId := GetStringFromOffset( DeviceDescriptor.RawDeviceProperties,DeviceDescriptor.VendorIdOffset,32);
      //FVendorId := PCh;
    end;

    if DeviceDescriptor.ProductIdOffset <> 0 then
    begin
      PCh := @PCharArray(@DeviceDescriptor)^[DeviceDescriptor.ProductIdOffset];
      FProductId := GetStringFromOffset( DeviceDescriptor.RawDeviceProperties,DeviceDescriptor.ProductIdOffset,8);
      //FProductId := PCh;
      FVendorId:='"'+FVendorId+'-'+FProductId+'"';
    end;

    if DeviceDescriptor.ProductRevisionOffset <> 0 then
    begin
      PCh := @PCharArray(@DeviceDescriptor)^[DeviceDescriptor.ProductRevisionOffset];
      FProductRevision := PCh;
    end;

    if DeviceDescriptor.SerialNumberOffset <> 0 then
    begin
      PCh := @PCharArray(@DeviceDescriptor)^[DeviceDescriptor.SerialNumberOffset];
      FSerialNumber := PCh;
    end;
  finally
    if FFileHandle <> INVALID_HANDLE_VALUE then
      CloseHandle(FFileHandle);
  end;

            if (List<>nil) then begin
                with List.Items.Add do
                begin Caption:='Vendor';
                      SubItems.Add(FVendorId); end;
                with List.Items.Add do
                begin Caption:='Model';
                      SubItems.Add(FProductId); end;
                with List.Items.Add do
                begin Caption:='Version';
                      SubItems.Add(FProductRevision); end;
                with List.Items.Add do
                begin Caption:='Serial';
                      SubItems.Add(FSerialNumber); end;
            end;
  result:=FVendorId;
end;

end.
