{*******************************************************************************
NT Disk access unit
allows physical disks to be accessed as TStream
by Miika Sihvola (smiika@mbnet.fi)
*******************************************************************************}
unit Disks;

interface

uses Classes, Windows, Dialogs, Util;

const
  PartitionTableOffset              = $1BE; // 446 location inside master boot record

  PARTITION_STYLE_MBR               = 0;
  PARTITION_STYLE_GPT               = 1;
  PARTITION_STYLE_RAW               = 2;

  IOCTL_DISK_GET_DRIVE_LAYOUT_EX    = $00070050;
  //IOCTL_DISK_GET_DRIVE_LAYOUT       = CTL_CODE(IOCTL_DISK_BASE, 0x0003, METHOD_BUFFERED, FILE_READ_ACCESS)

  IOCTL_DISK_GET_DRIVE_GEOMETRY     = $70000; //(7 shl 16);
  IOCTL_DISK_GET_DRIVE_GEOMETRY_EX  = $700a0;
  //    CTL_CODE(IOCTL_DISK_BASE, 0x0000, METHOD_BUFFERED, FILE_ANY_ACCESS)
  // ex CTL_CODE(IOCTL_DISK_BASE, 0x0028, METHOD_BUFFERED, FILE_ANY_ACCESS)

  IOCTL_STORAGE_CHECK_VERIFY        = 2967552;

  MediaTypes : array [0..25] of string =
      ('Unknown',                // Format is unknown
       'F5_1Pt2_512',            // 5.25", 1.2MB,  512 bytes/sector
       'F3_1Pt44_512',           // 3.5",  1.44MB, 512 bytes/sector
       'F3_2Pt88_512',           // 3.5",  2.88MB, 512 bytes/sector
       'F3_20Pt8_512',           // 3.5",  20.8MB, 512 bytes/sector
       'F3_720_512',             // 3.5",  720KB,  512 bytes/sector
       'F5_360_512',             // 5.25", 360KB,  512 bytes/sector
       'F5_320_512',             // 5.25", 320KB,  512 bytes/sector
       'F5_320_1024',            // 5.25", 320KB,  1024 bytes/sector
       'F5_180_512',             // 5.25", 180KB,  512 bytes/sector
       'F5_160_512',             // 5.25", 160KB,  512 bytes/sector
       'REMOVABLE', //'RemovableMedia',         // Removable media other than floppy
       'FIXED', //FixedMedia',             // Fixed hard disk media
       'F3_120M_512',            // 3.5", 120M Floppy
       'F3_640_512',             // 3.5" ,  640KB,  512 bytes/sector
       'F5_640_512',             // 5.25",  640KB,  512 bytes/sector
       'F5_720_512',             // 5.25",  720KB,  512 bytes/sector
       'F3_1Pt2_512',            // 3.5" ,  1.2Mb,  512 bytes/sector
       'F3_1Pt23_1024',          // 3.5" ,  1.23Mb, 1024 bytes/sector
       'F5_1Pt23_1024',          // 5.25",  1.23MB, 1024 bytes/sector
       'F3_128Mb_512',           // 3.5" MO 128Mb   512 bytes/sector
       'F3_230Mb_512',           // 3.5" MO 230Mb   512 bytes/sector
       'F8_256_128',             // 8",     256KB,  128 bytes/sector
       'F3_200Mb_512',           // 3.5",   200M Floppy (HiFD)
       'F3_240M_512',            // 3.5",   240Mb Floppy (HiFD)
       'F3_32M_512');            // 3.5",   32Mb Floppy

type

  TDiskGeometry = packed record
    Cylinders: LARGE_INTEGER;
    MediaType: DWORD;
    TracksPerCylinder: dword;
    SectorsPerTrack: dword;
    BytesPerSector: dword;
  end;

  TDiskGeometryEx = packed record
    Geometry: TDiskGeometry;                                 // Standard disk geometry: may be faked by driver.
    DiskSize: int64;                                 // Must always be correct
    Data    : array[0..1] of byte;                                                  // Partition, Detect info
  end;

  TDriveLayoutInformationMbr = record
    Signature: DWORD;
  end;

  TDriveLayoutInformationGPT = record
    DiskId: TGuid;
    StartingUsableOffset: Int64;
    UsableLength: Int64;
    MaxPartitionCount: DWORD;
  end;
  PDriveLayoutInformationGPT = ^TDriveLayoutInformationGPT;

  TPartitionInformationMBR = record
    PartitionType: Byte;
    BootIndicator: Boolean;
    RecognizedPartition: Boolean;
    HiddenSectors: DWORD;
  end;
  PPartitionInformationMBR = ^TPartitionInformationMBR;

  TPartitionInformationGPT = record
    PartitionType: TGuid;
    PartitionId: TGuid;
    Attributes: Int64;
    Name: array [0..35] of WideChar;
  end;
  PPartitionInformationGPT = ^TPartitionInformationGPT;

  TPartitionInformationEx = record
    PartitionStyle: Integer;
    StartingOffset: Int64;
    PartitionLength: Int64;
    PartitionNumber: DWORD;
    RewritePartition: Boolean;
    case Integer of
      0: (Mbr: TPartitionInformationMbr);
      1: (Gpt: TPartitionInformationGpt);
  end;
  PPartitionInformationEx = ^TPartitionInformationEx;

	TGUIDPartitionTableEntry = record
		PartitionTypeGUID   : array[0..15] of byte;
		//Pad                 : byte;
		UniquePartitionGUID : array[0..15] of byte;
		StartingLBA         : int64;
		EndingLBA           : int64;
		AttributeBits       : array[0..7] of byte;
		PartitionName       : array[0..35] of WideChar;
  end;
  PGUIDPartitionTableEntry = ^TGUIDPartitionTableEntry;

  TDriveLayoutInformationEx = record
    PartitionStyle: DWORD;
    PartitionCount: DWORD;
    DriveLayoutInformation: record
      case Integer of
      0: (Mbr: TDriveLayoutInformationMbr);
      1: (Gpt: TDriveLayoutInformationGpt);
    end;
    PartitionEntry: array [0..15] of TPartitionInformationGpt;
    //hard-coded maximum of 16 partitions
  end;

  TDisk = class(TStream)
  public
    Name: string;
    Number: Integer;
    Physical: boolean;
    DiskHandle: THandle;
    DiskGeometry: TDiskGeometry;
    DiskGeometryEx: TDiskGeometryEx;
    DiskSerial:String;

    Size: Int64;
    TotalSectors: Int64;
    SurPlusSectors: Int64;
    SecBuf     : array [0..2047] of Byte;
    DriveLayout: TDriveLayoutInformationEx;

    constructor Create(DiskNumber: Byte); overload;
    constructor Create(DriveLetter: string; DiskNumber:integer); overload;
    destructor Destroy(); override;

    function OpenVolumePhysical:boolean;
    function OpenVolumeByLetter(Letter:string):boolean;
    procedure CloseVolume;

    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function SizeStr: string;
  end;

  //partition entry inside partition table
  TPartDesc = record
    BootIndicator: Byte;  //one of partitions have this value at 80 (which means partition is avtive), which is the partition that MBR code boots
    BeginCHS: array [1..3] of Byte;
    PartitionType: Byte;  //important byte, tells the filesystem type
    EndCHS: array [1..3] of Byte;
    StartSect: Cardinal;  //physical sector location where partition begins
    SizeSect: Cardinal;   //how many sectors partition has (size in bytes = 512 * SizeSect)
  end;
  //partiton table structure inside MBR
  TPartitionTable = array [1..4] of TPartDesc;

  //MBR, the first sector of every disk, and in begin of extended partitions
  TMBR = packed record
    BootCode: array [$0..$1BD] of Byte; //x86 16-bit realmode code loaded to address 7E00 by BIOS, no boot code in extended partition's MBR
    Partitions: TPartitionTable;        //starts at $1BE
    Signature: Word;                    //must be 55 AA
  end;

  TPARTITION_INFORMATION = packed record
    StartingOffset:LARGE_INTEGER;
    PartitionLength:LARGE_INTEGER;
    HiddenSectors:DWORD;
    PartitionNumber:DWORD;
    PartitionType:BYTE;
    BootIndicator:BOOLEAN;
    RecognizedPartition:BOOLEAN;
    RewritePartition:BOOLEAN;
  end;

  TSTORAGE_DEVICE_NUMBER = packed record
    // The FILE_DEVICE_XXX type for this device.
    DeviceType:DWORD;
    // The number of this device
    DeviceNumber:DWORD;
    // If the device is partitionable, the partition number of the device. Otherwise -1
    PartitionNumber:DWORD;
  end;

  //application specfic structure
  TPartInfo = record
    Disk: TDisk;
    StartSec: Int64;
    SizeSec:  Int64;
    PartitionTable: TPartitionTable;
    sLetter: string;
    partno:Byte;
    sVolName: string;
    sFSName: string;
    sVolSerial: string;
  end;
  PPartInfo = ^TPartInfo;


procedure GetPartitionInfo(driveletter: string; var sVolName: string; var sFSName: string; var sVolSerial: string);
function GetFSNameString(PartitionType: Byte): string;
function GetFSName(pDesc: TPartDesc): string;

//var DiskHandle : THandle;
//var Disk       : TDisk;
//var I          : Byte;

implementation

uses SysUtils, DiskName, Main,
     NTFS, FAT32;

(*
procedure GetDiskLayout;
const MAX_IDE_DRIVES = 16; // Max number of drives assuming master/slave topology
var i: Integer;
    DiskType: Byte;
    RawMBR: array[0..511] of Byte;
    btsIO: DWORD;
    hDevice: THandle;
    s: string;
begin
   s := '';
   for i := 0 to MAX_IDE_DRIVES - 1 do
   begin
      hDevice := CreateFile(PChar('\\.\PhysicalDrive' + IntToStr(i)), GENERIC_READ,
         FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
      if hDevice <> INVALID_HANDLE_VALUE then
      begin
         SetFilePointer(hDevice, 512, nil, FILE_BEGIN); //sector 1 for GPT
         if not ReadFile(hDevice, RawMBR[0], 512, btsIO, nil) then
         begin
            CloseHandle(hDevice);
            Continue;
         end;
         if (IntToHex(RawMBR[0], 2) + IntToHex(RawMBR[1], 2) +
             IntToHex(RawMBR[2], 2) + IntToHex(RawMBR[3], 2) +
             IntToHex(RawMBR[4], 2) + IntToHex(RawMBR[5], 2) +
             IntToHex(RawMBR[6], 2) + IntToHex(RawMBR[7], 2)) =
            '4546492050415254' then //EFI PART
            DiskType := 1 //GPT
         else
         begin
            DiskType := 0; //MBR
            SetFilePointer(hDevice, 0, nil, FILE_BEGIN); //sector 0 for MBR
            if not ReadFile(hDevice, RawMBR[0], 512, btsIO, nil) then
            begin
               CloseHandle(hDevice);
               Continue;
            end;
         end;
         CloseHandle(hDevice);
         if DiskType = 0 then
            s := s + 'Disk ' + IntToStr(i) + ' = ' + IntToHex(RawMBR[443], 2) + ' ' +
               IntToHex(RawMBR[442], 2) + ' ' + IntToHex(RawMBR[441], 2) +
               ' ' + IntToHex(RawMBR[440], 2) + #13#10
         else
            s := s + 'Disk ' + IntToStr(i) + ' = ' + IntToHex(RawMBR[59], 2) +
               ' ' + IntToHex(RawMBR[58], 2) + ' ' + IntToHex(RawMBR[57], 2) +
               ' ' + IntToHex(RawMBR[56], 2) + ' - ' + IntToHex(RawMBR[61], 2) +
               ' ' + IntToHex(RawMBR[60], 2) + ' - ' + IntToHex(RawMBR[63], 2) +
               ' ' + IntToHex(RawMBR[62], 2) + ' - ' + IntToHex(RawMBR[64], 2) +
               ' ' + IntToHex(RawMBR[65], 2) + ' - ' + IntToHex(RawMBR[66], 2) +
               ' ' + IntToHex(RawMBR[67], 2) + ' ' + IntToHex(RawMBR[68], 2) +
               ' ' + IntToHex(RawMBR[69], 2) + ' ' + IntToHex(RawMBR[70], 2) +
               ' ' + IntToHex(RawMBR[71], 2) +
               #13#10;
      end;
   end;
   //ShowMessage(s);
end;
*)

constructor TDisk.Create(DiskNumber: Byte);
var s1            : string;
begin
    Physical:=true;
    Number:=DiskNumber;

    // get actual device name in future
    //s1:=ReadPhysicalDriveInNT(DiskNumber, nil); // diskname.pas
    DiskSerial:='-';
    s1:=GetDriveManufacturerAndModel(DiskNumber,DiskSerial,nil); // diskname.pas
    insert('-',DiskSerial,5);
    Name := s1;
    OpenVolumePhysical;
end;

constructor TDisk.Create(DriveLetter: string; DiskNumber: integer);
var VolName, FSName, VolSerial: string;
begin
    Physical:=false;
    Number:=DiskNumber;
    Name := DriveLetter + ': nonPhysicalDrive' + IntToStr(Number);
    GetPartitionInfo(DriveLetter, VolName, FSName, VolSerial);
    OpenVolumeByLetter(Driveletter);
end;

destructor TDisk.Destroy;
begin
  CloseVolume;
  inherited Destroy;
end;

procedure TDisk.CloseVolume;
begin
  if Win32Platform=VER_PLATFORM_WIN32_NT then
  begin
    if (Physical) and (DiskHandle <> INVALID_HANDLE_VALUE) and (DiskHandle<>0) and (DiskHandle+1<>0) then
    begin
      CloseHandle(DiskHandle);
      DiskHandle:=0;
    end;
  end else
    DeleteFile('\\.\INT13EXT');
  DiskHandle:=0;
end;

function TDisk.OpenVolumePhysical:boolean;
var i: cardinal;
begin
    result:=false;
    if (DiskHandle=0) or (DiskHandle+1=0) then
    begin

      if Win32Platform=VER_PLATFORM_WIN32_NT then
      begin
        DiskHandle := CreateFile(pchar('\\.\PhysicalDrive' + IntToStr(Number)), GENERIC_READ,
            FILE_SHARE_READ	OR FILE_SHARE_WRITE, nil, OPEN_EXISTING,
            0, //FILE_FLAG_SEQUENTIAL_SCAN OR FILE_FLAG_NO_BUFFERING, //0,
            0)
      end else begin
      //Handle := CreateFile(pchar('\\.\PhysicalDrive' + IntToStr(DiskNumber)), GENERIC_READ, FILE_SHARE_READ	, nil, OPEN_EXISTING, 0, 0);
      //Handle := CreateFile(pchar('\\.\INT13EXT.VXD'), 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
        DiskHandle := CreateFile(pchar('\\.\vwin32'), 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
      end;
    end else begin

    end;

    if not DeviceIoControl(DiskHandle, IOCTL_STORAGE_CHECK_VERIFY, nil, 0, nil, 0, I, nil)
    then begin
      // make sure media is loaded
      CloseHandle(DiskHandle);
      DiskHandle:=0;
      Name:=Name+ ' not loaded';
      exit;
      //SysErrMsg(GetLastError, 'checkdisk');
    end;

    if (DiskHandle <> INVALID_HANDLE_VALUE) then
    begin
      result:=true;
      //if not DeviceIoControl(DiskHandle, IOCTL_DISK_GET_DRIVE_LAYOUT_EX,
      //cClustersTotal/cSectorsPerCluster

      if DeviceIoControl(DiskHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, nil, 0, @DiskGeometry, SizeOf(DiskGeometry), I, nil) then
      begin
        TotalSectors:=DiskGeometry.cylinders.QuadPart *
                      DiskGeometry.trackspercylinder *
                      DiskGeometry.sectorspertrack;

        //DiskGeometry:=DiskGeometryEx.Geometry;
        //DiskGeometryEx.Geometry.BytesPerSector := 512;
        //Size := DiskGeometry.DiskSize; // TotalSectors*512; //DiskGeometry.BytesPerSector;
        //SurPlusSectors := trunc(Size / 512) - TotalSectors;
        //TotalSectors := TotalSectors + SurPlusSectors;
      end else begin
        CloseHandle(DiskHandle);
        raise EInOutError.CreateFmt({s1}'', ['1']); // ('Failed getting disk geometry.');
      end;

      if DeviceIoControl(DiskHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, nil, 0, @DiskGeometryEx, SizeOf(DiskGeometryEx), I, nil) then
      begin
        //TotalSectors := DiskGeometry.cylinders.lowpart * DiskGeometry.trackspercylinder * DiskGeometry.sectorspertrack;
        TotalSectors:=DiskGeometryEx.Geometry.cylinders.QuadPart *
                      DiskGeometryEx.Geometry.trackspercylinder *
                      DiskGeometryEx.Geometry.sectorspertrack;

        //DiskGeometry:=DiskGeometryEx.Geometry;
        //DiskGeometryEx.Geometry.BytesPerSector := 512;
        Size := DiskGeometryEx.DiskSize; // TotalSectors*512; //DiskGeometry.BytesPerSector;
        SurPlusSectors := trunc(Size / 512) - TotalSectors;
        TotalSectors := TotalSectors + SurPlusSectors;
      end else begin
        CloseHandle(DiskHandle);
        raise EInOutError.CreateFmt({s1}'', ['1']); // ('Failed getting disk geometry.');
      end;

      (*
      if DeviceIoControl(DiskHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, nil, 0, @DiskGeometry, SizeOf(DiskGeometry), I, nil) then
      begin
        TotalSectors := DiskGeometry.cylinders.lowpart * DiskGeometry.trackspercylinder * DiskGeometry.sectorspertrack;
        Size := TotalSectors*512; //DiskGeometry.BytesPerSector;
      end else begin
        CloseHandle(DiskHandle);
        raise EInOutError.CreateFmt({s1}'', ['1']); // ('Failed getting disk geometry.');
      end;
      *)
    end else begin

      CloseHandle(DiskHandle);
      raise EInOutError.CreateFmt({s1}'', ['1']); // ('Failed getting disk device handle.');

    end;
end;

function TDisk.OpenVolumeByLetter(Letter:string):boolean;
var I             : Cardinal;
begin
    result:=false;
    if Win32Platform=VER_PLATFORM_WIN32_NT then
    begin
      DiskHandle := CreateFile(pchar('\\.\'+Letter+':'), GENERIC_READ,
        FILE_SHARE_READ	OR FILE_SHARE_WRITE, nil, OPEN_EXISTING,
        FILE_FLAG_SEQUENTIAL_SCAN OR FILE_FLAG_NO_BUFFERING, 0)
    end else begin
        //Handle := CreateFile(pchar('\\.\PhysicalDrive' + IntToStr(DiskNumber)), GENERIC_READ, FILE_SHARE_READ	, nil, OPEN_EXISTING, 0, 0);
        //Handle := CreateFile(pchar('\\.\INT13EXT.VXD'), 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
        //Handle := CreateFile(pchar('\\.\vwin32'), 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);
    end;

    if not DeviceIoControl(DiskHandle, IOCTL_STORAGE_CHECK_VERIFY, nil, 0, nil, 0, I, nil) then
    begin
      //CloseHandle(Handle);
      exit;
    end;

    if (DiskHandle <> INVALID_HANDLE_VALUE) then
    begin
      if DeviceIoControl(DiskHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, nil, 0, @DiskGeometry, SizeOf(DiskGeometry), I, nil) then
      begin
        TotalSectors := DiskGeometry.cylinders.QuadPart * DiskGeometry.trackspercylinder * DiskGeometry.sectorspertrack;
        Size := TotalSectors * 512; //DiskGeometry.BytesPerSector;
        result:=true;
      end else begin
        CloseHandle(DiskHandle);
        raise EInOutError.CreateFmt(name, ['1']); // ('Failed getting disk geometry.');
      end;
    end else begin
      CloseHandle(DiskHandle);
      raise EInOutError.CreateFmt(name, ['1']); // ('Failed getting disk device handle.');
    end;
end;

////////////////////////////////////////////////////////////////////////////////

function TDisk.Read(var Buffer; Count: Integer): Longint;
var BytesRead: Cardinal;
begin
  if Count = 0 then begin result:=0; exit; end;
  if Count mod 512 <> 0 then
	  raise EInOutError.Create('Number of bytes to read must be aligned to 512 bytes.');
  if not ReadFile(DiskHandle, Buffer, Count, BytesRead, nil) then
	  //RaiseLastWin32Error;
    SysErrMsg(GetLastError, 'DiskRead: ');
  Result := BytesRead;
end;

function TDisk.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
var
HighOffset: Integer;
begin
if Offset mod 512 <> 0 then
raise EInOutError.Create('Number of bytes to read must be aligned to 512 bytes.');
HighOffset := (Offset shr 32) and $00000000FFFFFFFF;
 case Origin of
  soBeginning:
   Result := SetFilePointer(DiskHandle, Integer(Offset and $00000000FFFFFFFF), @HighOffset, FILE_BEGIN);
  soCurrent:
   Result := SetFilePointer(DiskHandle, Integer(Offset and $00000000FFFFFFFF), @HighOffset, FILE_CURRENT);
  soEnd:
   Result := SetFilePointer(DiskHandle, Integer(Offset and $00000000FFFFFFFF), @HighOffset, FILE_END);
 else
   Result := 0;
  end;
end;

function TDisk.SizeStr: string;
begin
Result := IntToSize(Size);
end;

function TDisk.Write(const Buffer; Count: Integer): Longint;
begin
raise EInOutError.Create('Writing to physical disks is not implemented.');
end;

procedure GetPartitionInfo(driveletter: string; var sVolName, sFSName, sVolSerial: string);
var MaximumComponentLength,
    VOLSERIAL, FileSystemFlags: DWORD;
    VOLNAME: array[0..31] of char;
    FSNAME: array[0..15] of char;
    Size : integer;
begin
    Size := SizeOf(FSNAME);
    FillChar(FSNAME, Size, 0);
    GetVolumeInformation(PWideChar(driveletter + ':\'),
                       VOLNAME,32,
                       @VOLSERIAL,MaximumComponentLength,
                       FileSystemFlags,
                       FSNAME, 16);
    sVolName:=StrPas(VOLNAME);
    sFSName:=StrPas(FSNAME);
    sVolSerial := IntToHex(HiWord(VOLSERIAL), 4) + '-' +
                  IntToHex(LoWord(VOLSERIAL), 4);
end;

function GetFSNameString(PartitionType: Byte): string;
begin
  case PartitionType of

    $00: Result := 'Free';
    $05: Result := 'Extended DOS-partition';
    $0F: Result := 'Extended XINT13';

    $01: Result := 'FAT12';
    $02: Result := 'Xenix root';
    $03: Result := 'Xenix user';
    $04: Result := 'FAT16 < 32M';
    $06: Result := 'FAT16 > 32M MSDOSV4 HUGE';

    $07: Result := 'NTFS'; //|ExFat|ReFS';
    $08: Result := 'AIX';
    $09: Result := 'AIX boot';
    $0A: Result := 'OS/2 boot Manager';

    $0B: Result := 'FAT32';
    $0C: Result := 'FAT32 XINT13';
    $0E: Result := 'FAT16 XINT13';
    $11: Result := 'FAT12 Hidden';
    $27: Result := 'NTFS WinRE'; // 'Windows RE'; // Windows REcovery mostly hidden
    $A4: Result := 'NTFS WinRE ';
    $28: Result := 'FAT32 uefi';


    $14: Result := 'FAT16 Hidden (<32MB)';
    $16: Result := 'FAT16B Hidden (>=32MB)';

    $17: Result := 'NTFS Hidden';

    $1B: Result := 'FAT32 Hidden';
    $1C: Result := 'FAT32 Hidden XINT13';
    $1E: Result := 'FAT16 Hidden XINT13';

    $12: Result := 'Compaq diagnostics';
    $2C: Result := 'WildFile/Adaptec GOBack';
    $3C: Result := 'PowerQuest Recoverable Partition';
    $40: Result := 'VENIX 80286';
    $41: Result := 'PowerPC BOOT';
    $42: Result := 'Veritas Logical Disk Manager';
    $51: Result := 'Novell';
    $52: Result := 'MICROPORT';
    $63: Result := 'Unix';
    $64: Result := 'Novell NetWare 286';
    $65: Result := 'Novell NetWare (3.11 and 4.1)';
    $66: Result := 'Novell NetWare 386';

    $75: Result := 'PC/IX';
    $80: Result := 'OLD Minix';

    $81: Result := 'Linux/Minix v1.4b+';
    $82: Result := 'Linux Swap Partition';
    $83: Result := 'Linux Ext2/3/Rsr';
    $84: Result := 'OS/2 hiding type 04h partition';
    $85: Result := 'Linux Extended';

    $86: Result := 'NT FAT volume set';
    $87: Result := 'NT IFS volume set';

    $93: Result := 'Amoeba/Hidden Linux native FS (Reiser/Ext3/2)';
    $94: Result := 'Amoeba BBT';

    $A5: Result := 'BSD / 386';
    $AF: Result := 'APPLE_HFS';
    $B7: Result := 'BSDI fs';
    $B8: Result := 'BSDI swap';

    $C6: Result := 'SYRINX / Disabled NT FAT volume set';
    $C7: Result := 'SYRINX / Disabled NT IFS volume set';

    $DB: Result := 'CP/M';
    $DE: Result := 'Dell Corporation diagnostic partition';

    $E1: Result := 'DOS ACCESS';

    $EE: Result := 'MS Reserved';

    $F2: Result := 'DOS3.3+Secondary';
    $FE: Result := 'LANstep';
    $FF: Result := 'BBT';
  else Result := 'Unknown (' + IntToHex(PartitionType,2) + ')';
  end;
end;

function GetFSName(pDesc: TPartDesc): string;
begin
  result:=GetFSNameString(pDesc.PartitionType);
end;

end.


