{*******************************************************************************
Application-level (read only) FAT32 filesytem driver v0.1
            by Miika Sivola (smiika@mbnet.fi)
references
http://home.freeuk.net/foxy2k/disk/disk5.htm
*******************************************************************************}
unit FAT32;

interface

uses Classes,SysUtils,Windows,Dialogs,Main,Disks;

type

  TFAT32BootRecord = packed record
    BootCode: array [0..2] of Byte;
    OemName: array [0..7] of ansichar;
    BytesPerSec: Word;
    SecPerClust: Byte;
    ReservedSectors: Word;
    NumOfFatCopies: Byte;
    MaxRootEntries: Word;
    SectorsOld: Word;
    MediaDescriptor: Byte;
    SectorsPerFatOld: Word;
    SectorsPerTrack: Word;
    NumOfHeads: Word;
    HiddenSectors: Cardinal;
    SectorsSize: Cardinal;

    SectorsPerFat: Cardinal;
    Flags: Word;
    Version: Word;
    RootDirStartClust: Cardinal;
    FileSystemInfoSec: Word;
    BackupSec: Word;
    Reserved: array [0..11] of Byte;
    DriveNumber: Word;
    ExtraSignature: Byte;
    SerialNumber: Cardinal;
    VolumeName: array [0..10] of ansichar;
    FATName: array [0..7] of ansichar;

    Code: array [0..419] of Byte;
    BootSignature: Word;
  end;

  TFAT16BootRecord = packed record
    BootCode: array [0..2] of Byte;
    OemName: array [0..7] of ansichar;
    BytesPerSec: Word;
    SecPerClust: Byte;
    ReservedSectors: Word;
    NumOfFatCopies: Byte;
    MaxRootEntries: Word;
    SectorsOld: Word;
    MediaDescriptor: Byte;
    SectorsPerFatOld: Word;
    SectorsPerTrack: Word;
    NumOfHeads: Word;
    HiddenSectors: Cardinal;
    SectorsSize: Cardinal;

    LogicalDriveNumberofPartition: Word;
    ExtendedSignature: Byte;
    SerialNumber: Cardinal;
    VolumeName: array [0..10] of ansichar;
    FATName: array [0..7] of ansichar;

    Code: array [0..447] of Byte;
    BootSignature: Word;
  end;

  TFAT32FileSystemInfo = packed record
    Signature: Cardinal;
    Unknown: array [0..479] of Byte;
    FSInfoSignature: Cardinal;
    FreeClusters: Integer;
    RecentlyAllocatedClust: Cardinal;
    Reserved: array [0..13] of Byte;
    BootSignature: Word;
  end;

  TFAT32Attributes = set of (aReadOnly,aHidden,aSystem,aLabel,aDirectory,aArchive,aReserved1,aReserved2);

  TFAT32DirEntry = packed record
    FileName: array [0..7] of ansichar;
    FileExt: array [0..2] of ansichar;
    Attributes: TFAT32Attributes;
    NT: Byte;
    TenthOfCrtSec: Byte;
    TimeCreated: Word;
    DateCreated: Word;
    DateAccessed: Word;
    StartClustHigh: Word;
    Time: Word;
    Date: Word;
    StartClust: Word;
    Size: Cardinal;
  end;

  pFAT32DirEntry = ^TFAT32DirEntry;
  TFAT32DirEntryLFN = packed record
    Part: Byte;
    Name1: array [0..4] of WideChar;
    Attributes: Byte;
    Reserved: Byte;
    Checksum: Byte;
    Name2: array [0..5] of WideChar;
    Reserved2: Word;
    Name3: array [0..1] of WideChar;
  end;

  TFAT32 = class
  public
    PartInfo: TPartInfo;
    Disk: TDisk;
    bCloseHandleOnDestroy:Boolean;
    PartName:string;

    BootSec16: TFAT16BootRecord;
    BootSec: TFAT32BootRecord;
    InfoSec: TFAT32FileSystemInfo;
    FATStartSec: Cardinal;//begin of file allocation table
    DataStartSec: Cardinal;//begin of data area
    FatType: (ftFAT12,ftFAT16,ftFAT32);
    constructor Create(APartInfo: TPartInfo);
    destructor Destroy; override;
    function ReadCluster(Number: Cardinal): TMemoryStream;
    function NextCluster(CurrClust: Cardinal): Cardinal;
    function ReadClusterChain(StartClust: Cardinal; MaxSize: Cardinal = $FFFFFFFF): TMemoryStream;
    function ReadFile(DirEnt: TFAT32DirEntry): TMemoryStream;
  end;

//helper functions
function FAT32DateToStr(Date: Word): string;
function FAT32TimeToStr(Time: Word): string;
function FAT32AttribsToStr(Attribs: TFAT32Attributes): string;
function FAT32ShortFilename(DirEnt: TFAT32DirEntry): string;
function FAT32LongFilename(var DirEnt: TFAT32DirEntry; Stream: TMemoryStream): WideString;
function FAT32FileBeginClust(DirEnt: TFAT32DirEntry): Cardinal;

implementation

function FAT32DateToStr(Date: Word): string;
begin
Result := IntToStr(1980 + (Date shr 9));//Year
Result := Result + Format('-%.2d',[(Date shr 5) and $F]);//Month
Result := Result + Format('-%.2d',[(Date and $1F)]);//Day of month
end;

function FAT32TimeToStr(Time: Word): string;
begin
Result := Format(':%.2d',[Time and $1F]);//seconds
Result := Format(':%.2d',[(Time shr 5) and $3F]) + Result;//minutes
//Result := Format('%.2d',[Time shr 11]) + Result;//hours
end;

function TrimEndSpaces(S: string): string;
begin
while (Length(S) > 0) and (S[Length(S)] = ' ') do
Delete(S,Length(S),1);
Result := S;
end;

function FAT32ShortFilename(DirEnt: TFAT32DirEntry): string;
var
Name,Ext: string;
begin
Name := TrimEndSpaces(DirEnt.FileName);
Ext := TrimEndSpaces(DirEnt.FileExt);
if Ext <> '' then
Result := Name + '.' + Ext
else
Result := Name;
if (Length(Result) > 1) then if Result[1] = #229 then Result[1] := 'x';
while ((Length(Result) > 0) and (Result[Length(Result)] = #$FFFF) ) do
Delete(Result,Length(Result),2);
end;

function FAT32LongFilename(var DirEnt: TFAT32DirEntry; Stream: TMemoryStream): WideString;
var
LFNEnt: TFAT32DirEntryLFN;
begin
Move(DirEnt,LFNEnt,32);
Result := '' + LFNEnt.Name1 + LFNEnt.Name2 + LFNEnt.Name3;
Stream.ReadBuffer(LFNEnt,32);
while LFNEnt.Attributes = $0F do
 begin
 Result := '' + LFNEnt.Name1 + LFNEnt.Name2 + LFNEnt.Name3 + Result;
 Stream.ReadBuffer(LFNEnt,32);
 end;
Move(LFNEnt,DirEnt,32);
while ((Length(Result) > 0) and (Result[Length(Result)] = #$3F) ) do
Delete(Result,Length(Result),2);
while ((Length(Result) > 0) and (Result[Length(Result)] = #$FFFF) ) do
Delete(Result,Length(Result),2);
end;

function FAT32FileBeginClust(DirEnt: TFAT32DirEntry): Cardinal;
begin
Result := DirEnt.StartClust or (DirEnt.StartClustHigh shl 16);
end;

function FAT32AttribsToStr(Attribs: TFAT32Attributes): string;
begin
if aReadOnly in Attribs then
Result := 'read only, ';
if aHidden in Attribs then
Result := Result + 'hidden, ';
if aSystem in Attribs then
Result := Result + 'system, ';
if aLabel in Attribs then
Result := Result + 'volume label, ';
if aDirectory in Attribs then
Result := Result + 'directory, ';
if aArchive in Attribs then
Result := Result + 'archive, ';
Delete(Result,Length(Result)-1,2);
end;

{ TFAT32 }

destructor TFAT32.Destroy;
begin
  if bCloseHandleOnDestroy then
    Disk.CloseVolume;
  inherited Destroy;
end;

constructor TFAT32.Create(aPartInfo: TPartInfo);
var i:integer;
    FATTotSec:DWORD;
    FATDataSec:LONGINT;
    FATSz:longint;
    FATRootDirSectors:DWORD;
    FATCountofClusters: DWORD;
begin
  PartInfo := aPartInfo;
  Disk     := PartInfo.Disk;

  bCloseHandleOnDestroy:=false;
  if (Disk.DiskHandle=0) or (Disk.DiskHandle=INVALID_HANDLE_VALUE) then
  begin
    if Disk.Physical then
      Disk.OpenVolumePhysical
    else
      Disk.OpenVolumeByLetter(PartInfo.sLetter);
    bCloseHandleOnDestroy:=true;
  end;

  // read bootsector
  i:=sizeof(BootSec);
  Disk.Position := PartInfo.StartSec * Int64(512);
  if Disk.Read(BootSec,512)=512 then
  begin
    Disk.Position := (PartInfo.StartSec + Int64(BootSec.FileSystemInfoSec)) * Int64(512);
    Disk.Read(InfoSec,512);

    if BootSec.FATName = 'FAT32   ' then
      FatType := ftFAT32
    else
      FatType := ftFAT16;


    (*
    Disk.Position := PartInfo.StartSec * Int64(512);
    Disk.Read(BootSec16,512);
    FATRootDirSectors := Trunc(((BootSec.MaxRootEntries  * 32) + (BootSec.BytesPerSec - 1)) / BootSec.BytesPerSec);
    if (BootSec.SectorsOld <> 0) then FATSz := BootSec.SectorsPerFatOld // BPB_FATSz16
                                 else FATSz := BootSec.SectorsPerFat; // BPB_FATSz32;
    if (BootSec.SectorsOld <> 0) then FATTotSec := BootSec.SectorsOld
                                 else FATTotSec := BootSec.SectorsSize;

    FATDataSec := integer(FATTotSec) -
                 (BootSec.ReservedSectors +
                 (BootSec.NumOfFatCopies * FATSz) + integer(FATRootDirSectors));

    //Now we determine the count of clusters:
    FATCountofClusters := Trunc(FATDataSec / BootSec.SecPerClust);

    {
    Please note that this computation rounds down.
    Now we can determine the FAT type. Please note carefully or you will commit an off-by-one error!
    In the following example, when it says <, it does not mean <=. Note also that the numbers are correct.
    The first number for FAT12 is 4085; the second number for FAT16 is 65525. These numbers and the
    }

    if (FATCountofClusters < 4085) then
        FatType := ftFAT12  // Volume is FAT12
    else if(FATCountofClusters < 65525) then
        FatType := ftFAT16  // Volume is FAT16
    else
        FatType := ftFAT32; // Volume is FAT32

		if (FatType = ftFAT32) and (BootSec.Version<>0) then
        ;
        //name:=name+''
        //BootSec.Version
        //sprintf(cFSVersion, "%d.%d", HIBYTE(gBOOTSECFAT.Fat32.BPB_FSVer), LOBYTE(gBOOTSECFAT.Fat32.BPB_FSVer));
    //if BootSec.SectorsPerFat>PartInfo.SizeSec then
    *)


    if FatType = ftFAT32 then
    begin
      //FATStartSec := BootSec.HiddenSectors + BootSec.ReservedSectors;
      //DataStartSec := BootSec.HiddenSectors + BootSec.ReservedSectors + (BootSec.NumOfFatCopies * BootSec.SectorsPerFat);

      FATStartSec := PartInfo.StartSec + BootSec.ReservedSectors;
      DataStartSec := PartInfo.StartSec + BootSec.ReservedSectors +
                    (BootSec.NumOfFatCopies * BootSec.SectorsPerFat);
    end else begin
      Disk.Position := PartInfo.StartSec * Int64(512);
      Disk.Read(BootSec16,512);

      FATStartSec  := PartInfo.StartSec + BootSec.ReservedSectors;
      DataStartSec := PartInfo.StartSec + BootSec.ReservedSectors +
                      (BootSec.NumOfFatCopies * BootSec.SectorsPerFatOld) +
                      (BootSec.MaxRootEntries * 32 div 512);

      BootSec.SerialNumber:=BootSec16.SerialNumber;
      for i:=0 to sizeof(BootSec.VolumeName)-1 do
        BootSec.VolumeName[i]:=BootSec16.VolumeName[i];

      for i:=0 to sizeof(BootSec.FATName)-1 do
        BootSec.FATName[i]:=BootSec16.FATName[i];

      //BootSec.VolumeName:=BootSec16.VolumeName;
      //BootSec.FATName:=BootSec16.FATName;

      //FATStartSec := BootSec.HiddenSectors + BootSec.ReservedSectors;
      //DataStartSec := BootSec.HiddenSectors + BootSec.ReservedSectors + (BootSec.NumOfFatCopies * BootSec.SectorsPerFatOld) + (BootSec.MaxRootEntries * 32 div 512);
    end;
  end;
end;

function TFAT32.NextCluster(CurrClust: Cardinal): Cardinal;
var
Sector: Cardinal;
Offset: Cardinal;
Buf: array [0..511] of Byte;
begin
if FatType = ftFat32 then
 begin
 Sector := CurrClust div 128;
 Offset := (CurrClust mod 128) * 4;
 Disk.Position := Int64(FATStartSec + Sector) * Int64(512);
 Disk.Read(Buf,512);
 //DiskRead(PartInfo.DiskID,FATStartSec + Sector,1,@Buf);
 Move(Buf[Offset],Result,4);
 end
else
 begin
 Sector := CurrClust div 256;
 Offset := (CurrClust mod 256) * 2;
 Disk.Position := Int64(FATStartSec + Sector) * Int64(512);
 Disk.Read(Buf,512);
 //DiskRead(PartInfo.DiskID,FATStartSec + Sector,1,@Buf);
 Move(Buf[Offset],Result,2);
 end;
end;

function TFAT32.ReadCluster(Number: Cardinal): TMemoryStream;
begin
Result := TMemoryStream.Create;
Result.SetSize(Int64(BootSec.SecPerClust) * 512);
Disk.Position := Int64(DataStartSec + (Number * BootSec.SecPerClust) - (2 * BootSec.SecPerClust)) * 512;
Result.CopyFrom(Disk,Int64(BootSec.SecPerClust) * 512);
Result.Position := 0;
//Result := DiskReadStream(PartInfo.DiskID,DataStartSec + (Number * BootSec.SecPerClust) - (2 * BootSec.SecPerClust),BootSec.SecPerClust);
end;

function TFAT32.ReadClusterChain(StartClust: Cardinal; MaxSize: Cardinal = $FFFFFFFF): TMemoryStream;
var
CurClust: Cardinal;
Clust: TMemoryStream;
begin
Result := TMemoryStream.Create;
CurClust := StartClust;
if FatType = ftFAT32 then
 repeat
 if (CurClust>1000000) then
 begin
  CurClust:=CurClust;
 end;
 Clust := ReadCluster(CurClust);
 Result.Write(Clust.Memory^,Clust.Size);
 Clust.Free;
 CurClust := NextCluster(CurClust);
 until (CurClust >= $0FFFFFF8) or (CurClust < 2) or (Result.Size >= MaxSize)
else
 repeat
 Clust := ReadCluster(CurClust);
 Result.Write(Clust.Memory^,Clust.Size);
 Clust.Free;
 CurClust := NextCluster(CurClust);
 until (CurClust >= $FFF8) or (CurClust < 2) or (Result.Size >= MaxSize);
Result.Position := 0;
end;

function TFAT32.ReadFile(DirEnt: TFAT32DirEntry): TMemoryStream;
begin
Result := ReadClusterChain(FAT32FileBeginClust(DirEnt),DirEnt.Size);
Result.SetSize(DirEnt.Size);
end;

end.
