unit FAT32Browse;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Grids, MPHexEditor, Main, ImgList, StdCtrls, Menus, FAT32,
  ToolWin, Util;

type

  TFAT32BrowseForm = class(TForm)
    View: TListView;
    ImageList1: TImageList;
    PopupMenu1: TPopupMenu;
    Info1: TMenuItem;
    ViewinDiskEditor1: TMenuItem;
    ViewFATentryinDiskEditor1: TMenuItem;
    N1: TMenuItem;
    SaveAs1: TMenuItem;
    RecoveryWizard1: TMenuItem;
    SaveDlg: TSaveDialog;
    ImageList2: TImageList;
    CoolBar1: TCoolBar;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    Open1: TMenuItem;
    OpeninNewWindow1: TMenuItem;
    N2: TMenuItem;
    Status: TStatusBar;
    Edit1: TEdit;
    reloadbtn: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);

    procedure Exit1Click(Sender: TObject);
    procedure ReadDirectory(Clust: Cardinal);
    procedure ViewDblClick(Sender: TObject);
    procedure Info1Click(Sender: TObject);
    procedure ViewinDiskEditor1Click(Sender: TObject);
    procedure ViewFATentryinDiskEditor1Click(Sender: TObject);
    procedure SaveAs1Click(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure OpeninNewWindow1Click(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure RecoveryWizard1Click(Sender: TObject);
    procedure ViewCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
    procedure ViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure ViewKeyUp(Sender: TObject; var Key: Word;Shift: TShiftState);

    procedure FormShow(Sender: TObject);
    procedure OneDirUp;
    procedure FormResize(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;Shift: TShiftState);
    procedure reloadbtnClick(Sender: TObject);
  public
    CurClust: Cardinal;
    FAT32: TFAT32;
    PartName:string;
    BackHistory: TList;
    NextHistory: TList;
  end;

var
  FAT32BrowseForm: TFAT32BrowseForm;
  ColumnToSortFound        : Integer;
  reverseorder             : boolean;

implementation

uses FAT32FileProps, DiskEdit, FAT32Search;

{$R *.dfm}

procedure TFAT32BrowseForm.FormCreate(Sender: TObject);
begin
BackHistory := TList.Create;
NextHistory := TList.Create;
ColumnToSortFound := 2;
reverseorder:= false;
end;

procedure TFAT32BrowseForm.Exit1Click(Sender: TObject);
begin
Close;
end;

procedure TFAT32BrowseForm.ReadDirectory(Clust: Cardinal);
var M: TMemoryStream;
  DirEnt: pFAT32DirEntry;
  Item: TListItem;
  FileName: string;
  stype   : string;
  lastsort: integer;
begin
  View.SortType:=stNone;
  View.Clear;
  Status.Panels[0].Text := 'Searching, please wait...';
  if Clust = 0 then
  begin
    if FAT32.FatType = ftFAT32 then begin
      if FAT32.BootSec.RootDirStartClust=0 then
        FAT32.BootSec.RootDirStartClust:=FAT32.FATStartSec;
      CurClust := FAT32.BootSec.RootDirStartClust;
      M := FAT32.ReadClusterChain(CurClust);
    end else begin
      CurClust := 0;
      M := TMemoryStream.Create;
      M.SetSize((FAT32.BootSec16.MaxRootEntries * 32));

      FAT32.Disk.Position :=
          Int64(FAT32.BootSec16.ReservedSectors +
                FAT32.BootSec16.NumOfFatCopies * FAT32.BootSec16.SectorsPerFatOld) * 512;
      if (FAT32.BootSec16.MaxRootEntries<FAT32.BootSec.MaxRootEntries) then
      M.CopyFrom(FAT32.Disk,(FAT32.BootSec.MaxRootEntries * 32))
      else
      M.CopyFrom(FAT32.Disk,(FAT32.BootSec16.MaxRootEntries * 32));
      M.Position := 0;
    end;
  end else begin
     CurClust := Clust;
     M := FAT32.ReadClusterChain(CurClust);
  end;

  repeat
    GetMem(DirEnt,SizeOf(TFAT32DirEntry));
    M.ReadBuffer(DirEnt^,SizeOf(TFAT32DirEntry));
    if Byte(DirEnt.Attributes) = $0F then FileName := FAT32LongFilename(DirEnt^,M)
                                     else FileName := FAT32ShortFilename(DirEnt^);
    if (DirEnt.FileName[0] in [#0]) then begin
      if (FAT32.FatType = ftFAT16) and (CurClust = 0) then 
        Break 
      else 
        Continue;
    end;

    if FileName<>'.' then
    begin
      Item := View.Items.Add;
      Item.Data := DirEnt;
      Item.Caption := trim(FileName);
      if DirEnt.FileName[0] = #229 then begin
        if aDirectory in DirEnt.Attributes then begin
          stype:='dir, deleted';
          Item.ImageIndex := 3;
        end else begin
          stype:='file, deleted';
          Item.ImageIndex := 2;
        end;
      end else begin
        if aDirectory in DirEnt.Attributes then begin
          stype:='dir';
          Item.ImageIndex := 0;
        end else begin
          stype:='file';
          Item.ImageIndex := 1;
        end;
      end;
      //if directory entry is file, show it's size
      if not (aDirectory in DirEnt.Attributes)
      then Item.SubItems.Add(IntToStr(DirEnt.Size))
      else Item.SubItems.Add('-');
      Item.SubItems.Add(stype);
    end;
  until M.Size - M.Position <= 32;
  FreeMem(DirEnt);

  M.Free;
  Status.Panels[0].Text := IntToStr(View.Items.Count) + ' items.';
  View.SortType:=stText;
  lastsort:=ColumnToSortFound;
  if lastsort<>0 then begin
   ColumnToSortFound:=0;
   View.AlphaSort;
  end;
  ColumnToSortFound:=lastsort;
  View.AlphaSort;
  View.SetFocus;
  View.Items[0].Selected:=TRUE;
end;

procedure TFAT32BrowseForm.ViewDblClick(Sender: TObject);
var DirEnt: TFAT32DirEntry;
begin
 if View.ItemIndex > -1 then
 begin
   //if aDirectory in DirEnt.Attributes then
   if View.Selected.ImageIndex = 0 then
   begin
     if View.Selected.Caption='..' then
       OneDirUp
     else begin
       if length(Edit1.Text)>=3 then
         Edit1.Text := Edit1.Text+View.Selected.Caption+'\';

       BackHistory.Add(Pointer(CurClust));
       DirEnt := TFAT32DirEntry(View.Selected.Data^);
       ReadDirectory(FAT32FileBeginClust(DirEnt));

       //Edit1.Text:=; //NTFS.GetReferencedDir(CurMFT, '');
       Application.ProcessMessages;
       View.Items[0].Selected:=TRUE;
       View.Items[0].Focused:=TRUE;
     end;
   end else
     Info1Click(Sender);
 end;
end;

procedure TFAT32BrowseForm.Info1Click(Sender: TObject);
begin
with TFAT32FilePropsForm.Create(Application) do
 begin
 FAT32 := Self.FAT32;
 DirEnt := TFAT32DirEntry(View.Selected.Data^);
 Caption := View.Selected.Caption;
 Show;
 end;
end;

procedure TFAT32BrowseForm.ViewinDiskEditor1Click(Sender: TObject);
var
Clust: Cardinal;
begin
Clust := FAT32FileBeginClust(TFAT32DirEntry(View.Selected.Data^));
with TDiskEditor.Create(Application) do
 begin
 Partition := True;
 Disk := FAT32.PartInfo.Disk;
 MinSec := FAT32.PartInfo.StartSec;
 MaxSec := FAT32.PartInfo.StartSec + FAT32.PartInfo.SizeSec;
 CurSec := FAT32.DataStartSec + (Clust * FAT32.BootSec.SecPerClust) -
           (2 * FAT32.BootSec.SecPerClust);
 BytesPerSector := FAT32.BootSec.BytesPerSec;
 BytesPerCluster := Fat32.BootSec.SecPerClust*FAT32.BootSec.BytesPerSec;
 UpdateSec;
 end;
end;

procedure TFAT32BrowseForm.ViewFATentryinDiskEditor1Click(Sender: TObject);
var
FATSec: Cardinal;
FATOffset: Word;
Clust: Cardinal;
begin
Clust := FAT32FileBeginClust(TFAT32DirEntry(View.Selected.Data^));
if FAT32.FatType = ftFAT32 then
 begin
 FATSec := Clust div 128;
 FATOffset := (Clust mod 128) * 4;
 end
else
 begin
 FATSec := Clust div 256;
 FATOffset := (Clust mod 256) * 2;
 end;
with TDiskEditor.Create(Application) do
 begin
 Partition := True;
 Disk := FAT32.PartInfo.Disk;
 MinSec := FAT32.PartInfo.StartSec;
 MaxSec := FAT32.PartInfo.StartSec + FAT32.PartInfo.SizeSec;
 CurSec := FAT32.FATStartSec + FATSec;
 BytesPerSector := FAT32.BootSec.BytesPerSec;
 BytesPerCluster := Fat32.BootSec.SecPerClust*FAT32.BootSec.BytesPerSec;
 UpdateSec;
 Editor.SelStart := FATOffset;
 Editor.SelEnd := FATOffset + 3;
 end;
end;

procedure TFAT32BrowseForm.SaveAs1Click(Sender: TObject);
var DirEnt: TFAT32DirEntry;
    M: TMemoryStream;
    i:integer;
begin
  SaveDlg.FileName := '[ --- name will be ignored --- ]';// View.Selected.Caption;
  if SaveDlg.Execute then
  begin
    for i:=0 to View.Items.Count-1 do
    begin
      if View.Items[i].Selected then
      begin
        DirEnt := TFAT32DirEntry(View.Items[i].Data^);
        M := FAT32.ReadFile(DirEnt);
        M.SaveToFile(ExtractFilePath(SaveDlg.FileName)+View.Items[i].Caption);
        M.Free;
      end;
    end;
  end;
  SaveDlg.FileName := '';
end;

procedure TFAT32BrowseForm.ToolButton1Click(Sender: TObject);
var
Clust: Cardinal;
begin
if BackHistory.Count > 0 then
 begin
 Clust := Cardinal(BackHistory[BackHistory.Count - 1]);
 BackHistory.Delete(BackHistory.Count - 1);
 NextHistory.Add(Pointer(CurClust));
 ReadDirectory(Clust);
 end
else
Beep;
end;

procedure TFAT32BrowseForm.ToolButton2Click(Sender: TObject);
var
Clust: Cardinal;
begin
if NextHistory.Count > 0 then
 begin
 Clust := Cardinal(NextHistory[NextHistory.Count - 1]);
 NextHistory.Delete(NextHistory.Count - 1);
 BackHistory.Add(Pointer(CurClust));
 ReadDirectory(Clust);
 end
else
Beep;
end;

procedure TFAT32BrowseForm.OneDirUp;
var Item: TListItem;
    DirEnt: TFAT32DirEntry;
    i:integer;
    lastone:string;
begin
  i:=Length(edit1.Text)-1;
  while (edit1.text[i]<>'\') and (i>3) do Dec(i);
  lastone:=copy(edit1.Text,i+1,Length(edit1.Text)-i-1);
  edit1.Text:=copy(edit1.Text,1,i);

  Item := View.FindCaption(0,'..',false,true,false);
  if Item <> nil then
  begin
    DirEnt := TFAT32DirEntry(Item.Data^);
    BackHistory.Add(Pointer(CurClust));
    ReadDirectory(FAT32FileBeginClust(DirEnt));
  end else
    Beep;

  Application.ProcessMessages;
  for i:=0 to View.Items.Count-1 do
  begin
    if View.Items[i].Caption=lastone then
    begin
      View.Selected:=nil;
      Application.ProcessMessages;
      View.Items[i].Selected:=TRUE;
      View.Items[i].Focused:=TRUE;
      View.Items[i].MakeVisible(false);
      break;
    end;
  end;
end;


procedure TFAT32BrowseForm.ToolButton3Click(Sender: TObject);
begin
  OneDirUp;
end;

procedure TFAT32BrowseForm.Open1Click(Sender: TObject);
var
DirEnt: TFAT32DirEntry;
begin
BackHistory.Add(Pointer(CurClust));
DirEnt := TFAT32DirEntry(View.Selected.Data^);
ReadDirectory(FAT32FileBeginClust(DirEnt));
end;

procedure TFAT32BrowseForm.OpeninNewWindow1Click(Sender: TObject);
var DirEnt: TFAT32DirEntry;
    sname:string;
begin
DirEnt := TFAT32DirEntry(View.Selected.Data^);
sname:=Edit1.Text+View.Selected.Caption+'\';
with TFAT32BrowseForm.Create(Application) do
 begin
  FAT32 := TFAT32.Create(Self.FAT32.PartInfo);
  Show;
  edit1.text:=sname;
  ReadDirectory(FAT32FileBeginClust(DirEnt));
  end;
end;

procedure TFAT32BrowseForm.PopupMenu1Popup(Sender: TObject);
var
I: Integer;
begin
if View.ItemIndex = -1 then
for I := 0 to PopupMenu1.Items.Count - 1 do
 begin
 PopupMenu1.Items[I].Visible := True;
 PopupMenu1.Items[I].Enabled := False;
 end
else
 begin
 for I := 0 to PopupMenu1.Items.Count - 1 do
 PopupMenu1.Items[I].Enabled := True;
 Open1.Visible            := View.Selected.ImageIndex = 0;
 OpeninNewWindow1.Visible := View.Selected.ImageIndex = 0;
 Info1.Visible                    := True;
 ViewinDiskEditor1.Visible        := True;
 ViewFATentryinDiskEditor1.Visible:= true;
 SaveAs1.Visible          := true; //View.Selected.ImageIndex = 1;
 RecoveryWizard1.Visible  := View.Selected.ImageIndex = 2;
 end;
end;

procedure TFAT32BrowseForm.RecoveryWizard1Click(Sender: TObject);
var DirEnt: TFAT32DirEntry;
begin
DirEnt := TFAT32DirEntry(View.Selected.Data^);

with TFAT32SearchForm.Create(Application) do
 begin
 FAT32 := TFAT32.Create(Self.FAT32.PartInfo);
 Show;
 end;
end;

function CompareNumbers(s1:string; s2: string): integer;
var n1, n2 : Int64;
begin
  if s1='-' then begin if reverseorder then result:=1 else  result:=-1; exit; end;
  if s2='-' then begin if reverseorder then result:=1 else  result:=-1; exit; end;
	result:=0;
	n1:=StrToInt64(s1); n2:=StrToInt64(s2);
	if n1<n2 then result:=1
	else if n1>n2 then result:=-1;
end;

procedure TFAT32BrowseForm.ViewCompare(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);
var ix   : Integer;
begin
   if Status.Panels[0].Text = 'Searching, please wait...' then exit;
   if Item1.Caption='..' then begin Compare:=-1; exit; end;
   if Item2.Caption='..' then begin Compare:= 1; exit; end;

   ix := ColumnToSortFound - 1;
   if (ix < 0) then
	 if not reverseorder then
	   Compare := CompareStr(Item1.Caption,Item2.Caption)
	 else
	   Compare := -CompareStr(Item1.Caption,Item2.Caption)
   else
	   if (ix = 0) then
		 if not reverseorder then
		   Compare := CompareNumbers(Item1.SubItems[ix],Item2.SubItems[ix])
		 else
		   Compare := -CompareNumbers(Item1.SubItems[ix],Item2.SubItems[ix])
	   else
  	   if not reverseorder then
	  	   Compare := CompareStr(Item1.SubItems[ix],Item2.SubItems[ix])
		   else
		     Compare := -CompareStr(Item1.SubItems[ix],Item2.SubItems[ix])
end;

procedure TFAT32BrowseForm.ViewColumnClick(Sender: TObject;Column: TListColumn);
begin
   if Status.Panels[0].Text = 'Searching, please wait...' then exit;
   if (ColumnToSortFound=Column.Index) then reverseorder:= not reverseorder;
   ColumnToSortFound := Column.Index;
   (Sender as TCustomListView).AlphaSort;
end;

procedure TFAT32BrowseForm.ViewKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
	if (Key=VK_RETURN) then
		ViewDblClick(Sender);
  if key=VK_Back then
    ToolButton3Click(sender);
end;

procedure TFAT32BrowseForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  fat32.Free;
end;

procedure TFAT32BrowseForm.FormShow(Sender: TObject);
begin
  ClearMessageBuffers;
  edit1.Text:=copy(PartName,1,2)+'\';
end;

procedure TFAT32BrowseForm.FormResize(Sender: TObject);
begin
  edit1.Width:=width-150;
end;

procedure TFAT32BrowseForm.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key=VK_F5 then reloadbtnClick(sender);
end;

procedure TFAT32BrowseForm.reloadbtnClick(Sender: TObject);
begin
  ReadDirectory(CurClust);
end;

end.
