unit FAT32Recovery;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, FAT32, ComCtrls;

type
  TFAT32RecoverForm = class(TForm)
    Bevel1: TBevel;
    Shape2: TShape;
    NextButton: TButton;
    Notebook1: TNotebook;
    Label3: TLabel;
    ClustCheck: TRadioButton;
    CopyAheadCheck: TRadioButton;
    CheckBox1: TCheckBox;
    Label1: TLabel;
    DetailsButton: TButton;
    DetailMemo: TMemo;
    ProgressBar1: TProgressBar;
    Label2: TLabel;
    SaveDlg: TSaveDialog;
    procedure DetailsButtonClick(Sender: TObject);
    procedure NextButtonClick(Sender: TObject);
    procedure Recover;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  public
    FAT32: TFAT32;
    DirEnt: TFAT32DirEntry;
    M: TMemoryStream;
  end;

var
  FAT32RecoverForm: TFAT32RecoverForm;

implementation

uses Math;

{$R *.dfm}

procedure TFAT32RecoverForm.DetailsButtonClick(Sender: TObject);
begin
DetailMemo.Visible := not DetailMemo.Visible;
if DetailMemo.Visible then
DetailsButton.Caption := '<< Hide Details'
else
DetailsButton.Caption := 'Show Details >>';
end;

procedure TFAT32RecoverForm.NextButtonClick(Sender: TObject);
begin
case Notebook1.PageIndex of
0:
 begin
 NextButton.Enabled := False;
 DetailMemo.Lines.Add('-File to recover: ' + FAT32ShortFilename(DirEnt));
 DetailMemo.Lines.Add('-Filesize: ' + IntToStr(DirEnt.Size) + ' bytes');
 Recover;
 NextButton.Enabled := True;
 end;
1:
 begin
 NextButton.Caption := 'Save...';
 end;
2:
 begin
 SaveDlg.FileName := FAT32ShortFilename(DirEnt);
 if SaveDlg.Execute then
  begin
  M.SaveToFile(SaveDlg.FileName);
  Close;
  end;
 end;
end;
Notebook1.PageIndex := Notebook1.PageIndex + 1;
end;

procedure TFAT32RecoverForm.Recover;
var
CurClust: Cardinal;
Clust: TMemoryStream;
begin
CurClust := FAT32FileBeginClust(DirEnt);
ProgressBar1.Max := DirEnt.Size;
if ClustCheck.Checked then
 begin
 M := TMemoryStream.Create;
 if FAT32.FatType = ftFAT32 then
  repeat
  Clust := FAT32.ReadCluster(CurClust);
  M.Write(Clust.Memory^,Clust.Size);
  Clust.Free;
  DetailMemo.Lines.Add('-Cluster ' + IntToStr(CurClust) + ' (sectors ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust)) + ' - ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust) + FAT32.BootSec.SecPerClust) + ')');
  CurClust := FAT32.NextCluster(CurClust);
  ProgressBar1.Position := M.Size;
  Application.ProcessMessages;
  until (CurClust >= $0FFFFFF8) or (CurClust < 2)
 else
  repeat
  Clust := FAT32.ReadCluster(CurClust);
  M.Write(Clust.Memory^,Clust.Size);
  Clust.Free;
  DetailMemo.Lines.Add('-Cluster ' + IntToStr(CurClust) + ' (sectors ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust)) + ' - ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust) + FAT32.BootSec.SecPerClust) + ')');
  CurClust := FAT32.NextCluster(CurClust);
  ProgressBar1.Position := M.Size;
  Application.ProcessMessages;
  until (CurClust >= $FFF8) or (CurClust < 2);
 end
else
 begin
 DetailMemo.Lines.Add('-Reading clusters ' + IntToStr(CurClust) + '-' + IntToStr(Round(DirEnt.Size / (512 * FAT32.BootSec.SecPerClust)) + CurClust) + ' (sectors ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust)) + ' - ' + IntToStr(FAT32.DataStartSec + ((CurClust-2) * FAT32.BootSec.SecPerClust) + (FAT32.BootSec.SecPerClust * Round(DirEnt.Size / (512 * FAT32.BootSec.SecPerClust)))) + ')');
 //M := DiskReadStream(FAT32.PartInfo.DiskID,FAT32.DataStartSec + (CurClust * FAT32.BootSec.SecPerClust) - (2 * FAT32.BootSec.SecPerClust),FAT32.BootSec.SecPerClust * Round(DirEnt.Size / (512 * FAT32.BootSec.SecPerClust)));
 M := TMemoryStream.Create;
 FAT32.Disk.Position := Int64(FAT32.DataStartSec + (CurClust * FAT32.BootSec.SecPerClust) - (2 * FAT32.BootSec.SecPerClust)) * 512;
 M.CopyFrom(FAT32.Disk,Int64(FAT32.BootSec.SecPerClust * Round(DirEnt.Size / (512 * FAT32.BootSec.SecPerClust)) * 512));
 M.Position := 0;
 ProgressBar1.Position := ProgressBar1.Max;
 end;
if M.Size < DirEnt.Size then
 begin
 DetailMemo.Lines.Add('-Recovery stopped! Cluster chain is damaged and file is only partially recovered.');
 MessageBeep(MB_ICONERROR);
 if DetailMemo.Visible = False then DetailsButton.Click;
 end
else
 begin
 DetailMemo.Lines.Add('-Recovery completed!');
 MessageBeep(MB_ICONINFORMATION);
 M.SetSize(DirEnt.Size);
 end;
M.Position := 0;
end;

procedure TFAT32RecoverForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
FAT32.Free;
if M <> nil then M.Free;
Release;
end;

end.
