unit FAT32Search;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ToolWin, FAT32, ImgList, StdCtrls, StrUtils;

type
  TFAT32SearchForm = class(TForm)
    StatusBar1: TStatusBar;
    CoolBar1: TCoolBar;
    ToolBar1: TToolBar;
    StartButton: TToolButton;
    PauseButton: TToolButton;
    ToolButton3: TToolButton;
    SaveButton: TToolButton;
    ToolButton1: TToolButton;
    FindButton: TToolButton;
    FindNextButton: TToolButton;
    View: TListView;
    ImageList1: TImageList;
    InfoButton: TToolButton;
    SaveDlg: TSaveDialog;
    Filter: TEdit;
    procedure StartButtonClick(Sender: TObject);
    procedure Search;
    procedure FindButtonClick(Sender: TObject);
    procedure FindNextButtonClick(Sender: TObject);
    procedure PauseButtonClick(Sender: TObject);
    procedure ViewClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure InfoButtonClick(Sender: TObject);
    procedure ViewDblClick(Sender: TObject);
    procedure ViewCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
    procedure ViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  public
    partname: string;
    CurrClust: Cardinal;
    FAT32: TFAT32;
    ToFind: string;
    paused: Boolean;
    Stop: Boolean;
    closing: Boolean;
  end;

const
  cWaitString = 'Searching, please wait...';

implementation

uses Main, FAT32FileProps;

{$R *.dfm}

var
  ColumnToSortFound: Integer;
  reverseorder: Boolean;

procedure TFAT32SearchForm.Search;
var
  ext: string;
  runs: longint;

  procedure ScanDir(StartClust: Cardinal; DirName: string);
  var
    Buffer: TMemoryStream;
    DirEnt: pFAT32DirEntry;
    Item: TListItem;
    FileName: string;
  begin
    runs := 0;
    Buffer := FAT32.ReadClusterChain(StartClust);
    CurrClust := StartClust;
    StatusBar1.Panels[1].Text := 'Scanning ' + DirName;

    GetMem(DirEnt, SizeOf(TFAT32DirEntry));
    while (Buffer.Size - Buffer.Position > SizeOf(TFAT32DirEntry)) and
      (not paused) do
    begin
      inc(runs);
      if (runs mod 50 = 0) then
      begin
        View.Items.endUpdate;
        View.Refresh;
        Application.ProcessMessages;
        View.Items.BeginUpdate;
      end;

      Buffer.ReadBuffer(DirEnt^, SizeOf(TFAT32DirEntry));
      if (aDirectory in DirEnt.Attributes) and (DirEnt.FileName[0] <> '.') and
        (DirEnt.FileName[0] <> #229) and
        ((Pos(Filter.Text, FileName) <> 0) or (FileName = '') or
        (Filter.Text = '') or (Filter.Text = '*.*')) then
      begin
        if Byte(DirEnt.Attributes) = $0F then
          FileName := FAT32LongFilename(DirEnt^, Buffer)
        else
          FileName := FAT32ShortFilename(DirEnt^);
        ScanDir(FAT32FileBeginClust(DirEnt^), DirName + FileName + '\');

      end
      else if DirEnt.FileName[0] = #229 then
      begin
        if Byte(DirEnt.Attributes) = $0F then
          FileName := FAT32LongFilename(DirEnt^, Buffer)
        else
          FileName := FAT32ShortFilename(DirEnt^);
        if FileName[1] = #229 then
          FileName[1] := '?';
        Item := View.Items.Add;
        Item.Data := DirEnt; // we still need this dir entry in recovery
        Item.Caption := FileName;
        ext := LowerCase(ExtractFileExt(FileName));
        Item.SubItems.Add(RightStr(ext, Length(ext) - 1));
        Item.SubItems.Add(IntToStr(DirEnt.Size));
        // Item.SubItems.Add(IntToSize(DirEnt.Size));
        Item.SubItems.Add(DirName);
        Item.SubItems.Add(FAT32DateToStr(DirEnt.Date) + '-' +
          FAT32TimeToStr(DirEnt.Time));

        if aDirectory in DirEnt.Attributes then
          Item.ImageIndex := 1;
      end;

    end; // while
    FreeMem(DirEnt);
    Buffer.Free;

  end; // ScanDir

begin
  View.Items.BeginUpdate;
  ScanDir(CurrClust, '\');
  View.Items.endUpdate;

  Sleep(5);
  Application.ProcessMessages;
  if (not Stop) and (not closing) then
  begin
    if not paused then
    begin
      StatusBar1.Panels[0].Text := 'Search complete (' +
        IntToStr(View.Items.Count) + ' found)';
      PauseButton.Enabled := False;
      StartButton.Enabled := True;
      paused := False;
    end
    else
    begin
      StatusBar1.Panels[0].Text := 'Search aborted (' +
        IntToStr(View.Items.Count) + ' found)';
      PauseButton.Enabled := True;
    end;
    StatusBar1.Panels[1].Text := '';
  end;
end;

procedure TFAT32SearchForm.StartButtonClick(Sender: TObject);
begin
  StatusBar1.Panels[0].Text := cWaitString;
  View.Clear;
  CurrClust := FAT32.BootSec.RootDirStartClust;
  Stop := False;
  paused := False;
  StartButton.Enabled := False;
  PauseButton.Enabled := True;
  Application.ProcessMessages;
  Search;
end;

procedure TFAT32SearchForm.PauseButtonClick(Sender: TObject);
begin
  if not paused then
  begin
    paused := True;
    PauseButton.Down := True;
    PauseButton.Caption := 'Resume';
    PauseButton.Enabled := True;
    StatusBar1.Panels[0].Text := 'Paused, click "Resume" to continue';
  end
  else
  begin
    StatusBar1.Panels[0].Text := cWaitString;
    paused := False;
    PauseButton.Caption := 'Pause';
    PauseButton.Down := False;
    Search;
  end;
end;

procedure TFAT32SearchForm.FindButtonClick(Sender: TObject);
var
  S: string;
  I: Integer;
begin
  if InputQuery('Find', 'Find filename containing:', S) then
  begin
    ToFind := LowerCase(S);
    for I := 0 to View.Items.Count - 1 do
    begin
      if Boolean(Pos(ToFind, LowerCase(View.Items[I].Caption))) then
      begin
        View.Selected := View.Items[I];
        View.Selected.MakeVisible(False);
        MessageBeep(MB_ICONINFORMATION);
        Exit;
      end;
    end;
    MessageDlg('Not found.', mtInformation, [mbOK], 0);
  end;
end;

procedure TFAT32SearchForm.FindNextButtonClick(Sender: TObject);
var
  I: Integer;
begin
  if (ToFind <> '') and (View.Items.Count > 1) and (View.ItemIndex > -1) then
  begin
    for I := View.Selected.Index + 1 to View.Items.Count - 1 do
    begin
      if Boolean(Pos(ToFind, LowerCase(View.Items[I].Caption))) then
      begin
        View.Selected := View.Items[I];
        View.Selected.MakeVisible(False);
        MessageBeep(MB_ICONINFORMATION);
        Exit;
      end;
    end;
    MessageDlg('Not found.', mtInformation, [mbOK], 0);
  end
  else
  begin
    FindButton.Click;
  end;
end;

procedure TFAT32SearchForm.ViewClick(Sender: TObject);
begin
  SaveButton.Enabled := View.ItemIndex > -1;
  InfoButton.Enabled := View.ItemIndex > -1;
end;

procedure TFAT32SearchForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  paused := True;
  Stop := True;
  closing := True;
  Sleep(10);
  Application.ProcessMessages;
  FAT32.Free;
  Release;
end;

procedure TFAT32SearchForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  paused := True;
  Stop := True;
  closing := True;
  Sleep(10);
  Application.ProcessMessages;
end;

procedure TFAT32SearchForm.FormShow(Sender: TObject);
begin
  reverseorder := False;
end;

procedure TFAT32SearchForm.InfoButtonClick(Sender: TObject);
begin
  with TFAT32FilePropsForm.Create(Application) do
  begin
    FAT32 := Self.FAT32;
    DirEnt := TFAT32DirEntry(View.Selected.Data^);
    Caption := View.Selected.Caption;
    Show;
  end;
end;

procedure TFAT32SearchForm.ViewDblClick(Sender: TObject);
begin
  if View.ItemIndex > -1 then
    InfoButton.Click;
end;

procedure TFAT32SearchForm.SaveButtonClick(Sender: TObject);
var
  DirEnt: TFAT32DirEntry;
  M: TMemoryStream;
  J, k, r, x: Integer;
  Path: string;
  fname: string;

begin

  SaveDlg.FileName := '(no name)';
  if SaveDlg.Execute then
    Path := ExtractFilePath(SaveDlg.FileName)
  else
    Exit;
  k := 0;
  r := 0;
  x := 0;

  for J := 0 to View.Items.Count - 1 Do
    if View.Items[J].Selected then
    begin
      StatusBar1.Panels[0].Text := 'Recovering file ' + IntToStr(r) + '/' +
        IntToStr(View.SelCount + x - 1);
      StatusBar1.Refresh;
      inc(x);
      DirEnt := TFAT32DirEntry(View.Selected.Data^);
      try
        M := FAT32.ReadFile(DirEnt);
        fname := View.Items[J].Caption;
        if FileExists(Path + fname) or (Length(fname) = 1) then
        begin
          inc(k);
          fname := Path + IntToStr(k) + fname;
        end
        else
          fname := Path + fname;

        inc(r);

        (* while FileExists(fname) do begin
          i := Length(fname)-1;
          while not Isnumeric(fname[i]) do Dec(i);
          j := i;
          while Isnumeric(fname[j]) do Dec(j);
          Val(Copy(fname, j+1, i-j), num, dum);
          fname := Copy(fname, 1, j) + IntToStr(num+1) + Copy(fname, j+i-j+1, Length(fname)-1);
          end; *)

        M.SaveToFile(fname);
        M.Free;

        View.Items[J].Selected := False;
        View.Items[J].Caption := '[recovered]-' + View.Items[J].Caption;
        View.Refresh;

      except
        on E: Exception do
          MessageDlg('File recovery failed. ' + E.Message, mtError, [mbOK], 0);
      end;
    end;

  StatusBar1.Panels[0].Text := IntToStr(r) + '/' + IntToStr(x) +
    ' files Recovered.';
  { with TFAT32RecoverForm.Create(Application) do
    begin
    FAT32 := TFAT32.Create(Self.FAT32.PartInfo);
    DirEnt := TFAT32DirEntry(View.Selected.Data^);
    Show;
    end; }
end;

function CompareNumbers(s1: string; s2: string): Integer;
var
  n1, n2: Int64;
begin
  if s1 = '-' then
  begin
    if reverseorder then
      result := 1
    else
      result := -1;
    Exit;
  end;
  if s2 = '-' then
  begin
    if reverseorder then
      result := 1
    else
      result := -1;
    Exit;
  end;
  result := 0;
  n1 := StrToInt64(s1);
  n2 := StrToInt64(s2);
  if n1 < n2 then
    result := 1
  else if n1 > n2 then
    result := -1;
end;

procedure TFAT32SearchForm.ViewCompare(Sender: TObject; Item1, Item2: TListItem;
  Data: Integer; var Compare: Integer);
var
  ix: Integer;
begin
  ix := ColumnToSortFound - 1;
  if (ix < 0) then
    if not reverseorder then
      Compare := CompareStr(Item1.Caption, Item2.Caption)
    else
      Compare := -CompareStr(Item1.Caption, Item2.Caption)
  else if (ix = 1) then
    if not reverseorder then
      Compare := CompareNumbers(Item1.SubItems[ix], Item2.SubItems[ix])
    else
      Compare := -CompareNumbers(Item1.SubItems[ix], Item2.SubItems[ix])
  else if not reverseorder then
    Compare := CompareStr(Item1.SubItems[ix], Item2.SubItems[ix])
  else
    Compare := -CompareStr(Item1.SubItems[ix], Item2.SubItems[ix])
end;

procedure TFAT32SearchForm.ViewColumnClick(Sender: TObject;
  Column: TListColumn);
begin
  if closing or paused or Stop then
    Exit;
  if (ColumnToSortFound = Column.Index) then
    reverseorder := not reverseorder;
  ColumnToSortFound := Column.Index;
  (Sender as TCustomListView).AlphaSort;
end;

end.
