{ *******************************************************************************
  Application-level (read only) NTFS filesytem driver v0.1
  by Miika Sivola (smiika@mbnet.fi)
  references
  http://linux-ntfs.sourceforge.net/ntfs/index.html
  ******************************************************************************* }
{ *******************************************************************************
  Application-level (read only) NTFS filesytem driver v0.2
  by blume
  references
  http://linux-ntfs.sourceforge.net/ntfs/index.html
  ******************************************************************************* }

unit NTFS;

interface

uses Classes, SysUtils, Windows, Dialogs,
     Main, Disks, ProgressBar, Util, Forms;

type

  TNTFSBootRecord = packed record
    BootCode: array [0 .. 2] of Byte;
    OemName: array [0 .. 7] of Byte;
    BytesPerSec: Word;
    SecPerClust: Byte;
    ReservedSectors: Word;

    Unused0: array [0 .. 4] of Byte;
    MediaDescriptor: Byte;
    Unused1: Word;
    SectorsPerTrack: Word;
    NumOfHeads: Word;

    HiddenSectors: Cardinal; // PartitionOffset
    Unused2: array [0 .. 7] of Byte;
    SectorSize: Int64; // TotalSectorSize
    MFTClust: Int64;
    MFTMirrorClust: Int64;

    ClustersPerFileRec: Cardinal;
    ClustersPerIndexBlock: Cardinal;
    SerialNumber: Int64;
    Checksum: Cardinal;
    Code: array [0 .. 425] of Byte;
    BootSignature: Word;
  end;

  // NTFS volume information
  TNTFS_VOLUME_DATA_BUFFER = packed record
    SerialNumber: LARGE_INTEGER;
    NumberOfSectors: LARGE_INTEGER;
    TotalClusters: LARGE_INTEGER;
    FreeClusters: LARGE_INTEGER;
    Reserved: LARGE_INTEGER;
    BytesPerSector: ULONG;
    BytesPerCluster: ULONG;
    BytesPerMFTRecord: ULONG;
    ClustersPerMFTRecord: ULONG;
    MFTLength: LARGE_INTEGER;
    MFTStart: LARGE_INTEGER;
    MFTMirrorStart: LARGE_INTEGER;
    MFTZoneStart: LARGE_INTEGER;
    MFTZoneEnd: LARGE_INTEGER;
  end;
  PNTFS_VOLUME_DATA_BUFFER = ^TNTFS_VOLUME_DATA_BUFFER;

  TNTFSMFTRecFlags = set of (fInUse, fDir, fAlign = 15);

  // TNTFS_VOLUMEFLAGS = set of (IsDirty, ResizeLogFile, UpgradeOnMount,
  // MOUNTED_ON_NT4, DELETE_USN_UNDERWAY,
  // REPAIR_OBJECT_ID, CHKDSK_UNDERWAY=16384,
  // MODIFIED_BY_CHKDSK=32768, FLAGS_MASK=49215);

  TNTFSVolumeInformation = packed record
    Reserved: array [0 .. 1] of ULONG;
    Major_Version: UCHAR;
    Minor_Version: UCHAR;
    VolumeFlags: SHORT; // TNTFS_VOLUMEFLAGS;
  end;

  // TFileRecord = array of byte; //Array[0..1023] of byte;
  TFileRecord = Array [0 .. 1023] of Byte;
  PFileRecord = ^TFileRecord;

  TNTFS_FILE_RECORD_OUTPUT_BUFFER = packed RECORD
    FileReferenceNumber: Int64;
    FileRecordLength: DWORD;
    FileRecord: TFileRecord;
  END;
  PNTFS_FILE_RECORD_OUTPUT_BUFFER = ^TNTFS_FILE_RECORD_OUTPUT_BUFFER;

  TNTFS_FILE_RECORD_INPUT_BUFFER = RECORD
    FileReferenceNumber: TLargeInteger;
  end;

  TNTFSFileRecordHeader = packed record
    FileID: array [0 .. 3] of AnsiChar; // 'FILE'
    UsaOffset: Word;
    UsaSize: Word;
    LogFileSeqNum: Int64;
    SeqNum: Word;
    ReferenceCount: Word;
    OffsetToFirstAttr: Word;
    // Flags: DWORD; // set of (fInUse OR 1,fDir OR 2, fAlign=15)
    Flags: TNTFSMFTRecFlags; // set of (fInUse,fDir,fAlign=15)
    RealFileRecSize: Cardinal;
    AllocatedFileRecSize: Cardinal;
    RefToBaseFileRec: Int64;
    NextAttrID: Word;
    Align: Word;
    NumberOfRec: Cardinal;
    UsaNum: Word;
  end;
  PNTFSFileRecordHeader = ^TNTFSFileRecordHeader;

  TNTFSAttrFlags = set of (fCompressed, fEncrypted = 14, fSparse = 15);

  TNTFSStandardAttributeHeader = packed record
    AttributeType: Cardinal;
    Size: Cardinal; // headers + attribute
    NonResident: boolean;
    NameLength: Byte;
    NameOffset: Word;
    Flags: TNTFSAttrFlags; // $0001 compressed
    AttributeID: Word;
    // Data stream continues with resident or non-resident header according to NonResident... (records below)
  end;
  PNTFSStandardAttributeHeader = ^TNTFSStandardAttributeHeader;

  TNTFSStandardAttributeHeader_Resident = packed record
    AttributeLength: Cardinal;
    AttributeOffset: Word;
    IndexedFlag: Byte;
    Padding: Byte;
    // If attribute is named, unicode name comes here... (length=NameLength*2)
  end;
  PNTFSStandardAttributeHeader_Resident = ^TNTFSStandardAttributeHeader_Resident;

  TNTFSStandardAttributeHeader_NonResident = packed record
    StartVCN: Int64;
    LastVCN: Int64; // ULONGLONG
    DataRunsOffset: Word; // USHORT
    ComprtessionUnit: Word;
    Padding: Cardinal; // ULONG
    AllocatedSize: Int64;
    RealSize: Int64;
    StreamInitializedSize: Int64;
    // If attribute is named, unicode name comes here (length=NameLength*2) before run lists (pointed by DataRunsOffset)
  end;
  PNTFSStandardAttributeHeader_NonResident = ^TNTFSStandardAttributeHeader_NonResident;

  TNTFSPermissions = set of (aReadOnly, aHidden, aSystem, aNull1, aNull2,
    aArchive, aDevice, aNormal, aTemporary, aSparseFile, aReparsePoint,
    aCompressed, aOffline, aNotIndexed, aEncrypted, fNull = 31);

  TNTFSStandardInfoAttribute = packed record
    TimeCreated         : Int64;
    TimeModified        : Int64;
    TimeModifiedFileRec : Int64;
    TimeAccessedFileRec : Int64;
    FileAttributes: TNTFSPermissions;
    // AlignmentOrReservedOrUnknown[3]: cardinal;
    // QuotaId: cardinal; // NTFS 3.0 only
    // SecurityId: cardinal; // NTFS 3.0 only
    // QuotaCharge: int64; // NTFS 3.0 only
    // Usn; // NTFS 3.0 only
  end;
  PNTFSStandardInfoAttribute = ^TNTFSStandardInfoAttribute;

  TNTFSNameSpace = (npPosix, npWin32, npDos, npDosAndWin32);

  TNTFSFileNameAttribute = packed record
    RefToParentDir: Int64;
    TimeCreated: Int64;
    TimeLastModified: Int64;
    TimeMTFAltr: Int64;
    TimeLastAccess: Int64;
    AllocatedSize: Int64;
    RealSize: Int64;
    Flags: TNTFSPermissions;
    ForEAsReparse: Cardinal;
    FileNameLength: Byte;
    FileNamespace: TNTFSNameSpace;
    // FileNameStart: PWideChar;
    // unicode name comes here... (length=NameLength*2)
  end;
  PNTFSFileNameAttribute = ^TNTFSFileNameAttribute;

  TNTFSIndexRecord = packed record
    MFTReference: Int64;
    Size: Word;
    FilenameOffset: Word;
    IndexFlags: Word;
    Align: Word;
    ParentMFTReference: Int64;
    TimeCreated: Int64;
    TimeLastModified: Int64;
    TimeMTFAltr: Int64;
    TimeLastAccess: Int64;
    AllocatedSize: Int64;
    RealSize: Int64;
    Flags: Int64;
    FileNameLength: Byte;
    FileNamespace: TNTFSNameSpace;
  end;
  PNTFSIndexRecord = ^TNTFSIndexRecord;

  TNTFSIndexRoot = packed record
    AttributeType: Cardinal;
    CollationRule: Cardinal;
    SizeOfIndexAllocEntry: Cardinal;
    ClustPerIndexRec: Byte;
    Align: array [1 .. 3] of Byte;
    OffsetOfFirstEntry: Cardinal;
    SizeOfIndexEntries: Cardinal;
    AllocatedSizeForEntries: Cardinal;
    Flags: Byte;
    Align1: array [1 .. 3] of Byte;
  end;
  PNTFSIndexRoot = ^TNTFSIndexRoot;

  TNTFSIndexHeader = packed record
    IndxID: array [0 .. 3] of AnsiChar; // 'INDX'
    UpdateSeqOffset: Word;
    SizeOfUpdateSeq: Word;
    LogFileSeqNum: Int64;
    VCN: Int64;
    BytesToIndexEntries: Cardinal;
    SizeOfIndexEntries: Cardinal;
    AllocatedSizeForEntries: Cardinal;
    NotLeafNode: boolean;
    Align: array [1 .. 3] of Byte;
    UpdateSeq: Word;
  end;
  PNTFSIndexHeader = ^TNTFSIndexHeader;

  TNTFSAttributeListRecord = packed record
    AttributeType: Cardinal;
    Size: Word;
    NameLength: Byte;
    NameOffset: Byte;
    StartVCN: Int64;
    FileRef: Int64;
    AttributeID: Word;
    // Name (Size: NameLength * 2)
  end;
  PNTFSAttributeListRecord = ^TNTFSAttributeListRecord;

  TGUID = packed record
    Data1: ULONG;
    Data2: Word; // USHORT;
    Data3: Word; // USHORT;
    Data4: array [0 .. 7] of UCHAR;
  end;

  TNTFSAttributeOBJECTID = packed record
    ObjectId: TGUID;
    BirthVolumeId: TGUID;
    BirthObjectId: TGUID;
    DomainId: TGUID;
    ExtendedInfo: array [0 .. 47] of UCHAR;
  end;

  TAttribute = record
    Hdr: TNTFSStandardAttributeHeader;
    Data: TMemoryStream;
    case boolean of
      True:
        (HdrResident: TNTFSStandardAttributeHeader_Resident);
      False:
        (HdrNonResident: TNTFSStandardAttributeHeader_NonResident);
  end;
  PAttribute = ^TAttribute;

  (*
    typedef struct {
    ULONG ReparseTag;
    USHORT ReparseDataLength;
    USHORT Reserved;
    UCHAR ReparseData[1];
    } REPARSE_POINT, *PREPARSE_POINT;
  *)

  TReparsePoint = packed record
    ReparseTag: ULONG;
    ReparseDataLength: Word;
    Reserved: Word;
    // data comes here at
    ReparseData: PWideChar; //
  end;

  // internal array type to hold information about fragments
  TRunList = array of record
    desc: string;
    Lcn: Cardinal;
    Len: Cardinal;
    VCN: Cardinal;
    Sparse: boolean;
    MFTid: Int64;
  end;

  TMFTEntry = class;

  TNTFS = class
  public
    Disk: TDisk;
    PartInfo: TPartInfo;

    VolHandle: Cardinal;
    // extra handle to drive by letter \\.\c: for FSCTL_GET_NTFS_FILE_RECORD
    VolumeInfo: TNTFS_VOLUME_DATA_BUFFER;
    BootSec: TNTFSBootRecord;
    MFTStartSec: Int64;
    MFTRunList: TRunList; // contains clusters for MFT
    MaxMFT: Cardinal;
    BytesPerMFTRecord: Int64;

    bCloseHandleOnDestroy: boolean;
    PartName: string;

    constructor Create(aPartInfo: TPartInfo);
    destructor Destroy; override;

    procedure OpenVolHandle;

    function GetMFTRecordCluster(var MFTid: Cardinal): Int64;
    //procedure GetMFTRecordCluster(MFTid: Cardinal; var MFTStart, MFTneeded:Int64);
    function GetMFTRecordSector(MFTid: Cardinal): Integer;

    function ReadClusters(Start, Count: Cardinal): TMemoryStream;
    function ReadMFTData(MFTid: Cardinal): TMemoryStream;
    function ReadMFTDataAPI(MFTid: Cardinal): TMemoryStream;
    function ReadMFTEntry(MFTid: Cardinal): TMFTEntry;

    function FixupUpdateSequenceArray(Data: array of Byte; NTFSRecHdr: TNTFSFileRecordHeader; MFTid: Int64): boolean;

    function GetNTFSInfo: boolean;
    procedure DumpMetaFiles(const DriveId: AnsiChar; Results: TStrings);

    function ReadFile(Attr: TAttribute; InfoStdHdr: TNTFSStandardInfoAttribute): TMemoryStream;
    procedure SaveFileWithProgress(RunList: TRunList; FileName: String; filesize: Int64; InfoStdHdr: TNTFSStandardInfoAttribute);
    // function SaveFileAs(Attr: TAttribute; fname: string): Boolean;
    function GetReferencedDir(MFTid: Cardinal; dirname: String; var bDir: boolean): String;

    function AttributeProc(MFTentry1: TMFTEntry; I: integer; MFTid, parent: Cardinal; wantedVCN: Cardinal; aLevel: integer; dirname: widestring; var RunList: TRunList): boolean;
    function AddToRunList(var MFTid, parent: Cardinal; MFTentry1: TMFTEntry; attrid: integer; wantedVCN: Cardinal; var RunList: TRunList; dirname, attrname: String): boolean;

    // reads clusters runlist's data
    function ReadRunListData(Attr: TAttribute; FileName: String; InfoStdHdr: TNTFSStandardInfoAttribute): TMemoryStream;

    // reads runlist from attribute into dynamic array (TRunList)
    function ReadRunList(DataAttr: TAttribute): TRunList;
    procedure ReadRunList2(MFTid, parent: Cardinal; wantedVCN: Cardinal; aLevel: integer; dirname: widestring; var RunList: TRunList; runlistflag: boolean);
    procedure ReadRunListWrap(MFTid: Cardinal; var dirname: widestring; var RunList: TRunList);
  end;

  TMFTEntry = class
    constructor Create(Src: TMemoryStream; fNTFS: TNTFS; fMFTid: Cardinal);
    destructor Destroy; override;
    function GetFileSize: Int64;
    function IndexOfAttrib(AttributeType: Word): integer;
  public
    MFTid: Cardinal;
    NTFS: TNTFS;
    FileRecHdr: TNTFSFileRecordHeader;
    InfoStdHdr: TNTFSStandardInfoAttribute;
    AttribCount: integer;
    Attrib: array of TAttribute;
    function GetFileName(var parent: Cardinal): string;
    function GetFileNameAttr: TNTFSFileNameAttribute;
  end;

  // helper functions
function NTFSMFTRecFlagsToStr(Flags: TNTFSMFTRecFlags): string;
// function NTFSMFTRecFlagsToStr(Flags: WORD): string;

function NTFSAttrFlagsToStr(Flags: TNTFSAttrFlags): string;

function NTFSTimeToStr(Time: Int64): string;
function NTFSTimeToIntAge(Time: Int64): integer;

function NTFSAttribName(Attrib: Cardinal): string;
function NTFSAttribFileName(MFTentry: TMFTEntry; attrid: Cardinal; MFTid: Cardinal; aLevel: integer; aDir: boolean): string;
function NTFSPermissionsToStr(Perm: TNTFSPermissions): string;

// converts relative runlist cluster number into signed integer
function ToSigned(Value: integer; Size: Byte): Cardinal;
function ConvertUnicodeName(const name: PWideChar; Len: integer): AnsiString;
function WideStringToString(const ws: widestring; Len: integer): AnsiString;

// return code type
type
  NTSTATUS = UINT;

const
  // Error codes returned by NtFsControlFile (see NTSTATUS.H)
  STATUS_SUCCESS = NTSTATUS($00000000);
  STATUS_BUFFER_OVERFLOW = NTSTATUS($80000005);
  STATUS_INVALID_PARAMETER = NTSTATUS($C000000D);
  STATUS_BUFFER_TOO_SMALL = NTSTATUS($C0000023);
  STATUS_ALREADY_COMMITTED = NTSTATUS($C0000021);
  STATUS_INVALID_DEVICE_REQUEST = NTSTATUS($C0000010);

  // File System Control command for getting NTFS information
  FILE_ANY_ACCESS = $00;
  METHOD_BUFFERED = 0;
  FILE_DEVICE_FILE_SYSTEM = $00000009;
  // FSCTL_GET_VOLUME_INFORMATION = $90064;
  FSCTL_GET_NTFS_VOLUME_DATA = $00090064;
  // CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 25, METHOD_BUFFERED, FILE_ANY_ACCESS);
  // NTFS_VOLUME_DATA_BUFFER
  FSCTL_GET_NTFS_FILE_RECORD = DWORD(((9 shl 16) or (26 shl 2)));

  // Metadata file names
  MetaFileNames: array [0 .. 8] of string = ('$MFT', '$MFTMirr', '$Logfile',
    '$Volume', '$AttrDef', '$Bitmap', '$Boot', '$BadClus', '$Upcase');

  NameSpaces: array [npPosix .. npDosAndWin32] of string = ('POSIX', 'Win32',
    'DOS', 'DOS and Win32');

  AttribNames: array [$01 .. $10] of string =
  // atrribute id must be divided with 10 so the name array will match!
    ('Standard Info', 'Attribute List', 'File Name', 'Object ID',
    'Security Descriptor', 'Volume Name', 'Volume Information', 'Data',
    'Index Root', 'Index Allocation', 'Bitmap', 'Reparse Point',
    'Extended Attribute Information', 'Extended Attribute', 'Property Set',
    'Logged Utility Stream');
  { ('STANDARD INFORMATION',
    'ATTRIBUTE LIST',
    'FILE NAME',
    'OBJECT ID',
    'SECURITY DESCRIPTOR',
    'VOLUME NAME',
    'VOLUME INFORMATION',
    'DATA',
    'INDEX ROOT',
    'INDEX ALLOCATION',
    'BITMAP',
    'REPARSE POINT',
    'EXTENDED ATTRIBUTE INFORMATION',
    'EXTENDED ATTRIBUTE',
    'PROPERTY SET',
    'LOGGED UTILITY STREAM'); }

  (*
    :$DATA
    :$EA_INFORMATION
    :$INDEX_ALLOCATION
    :$INDEX_ROOT
    :$LOGGED_UTILITY_STREAM
    :$VOLUME_NAME
    :$VOLUME_INFORMATION
    ::$ATTRIBUTE_LIST
    ::$BITMAP
    ::$EA
    ::$PropertySet
    ::$Reparse
    ::$SECURITY_DESCRIPTOR
  *)

  AttribFilenames: array [$01 .. $10] of string = ('::$STANDARD_INFORMATION',
    '::$ATTRIBUTE_LIST', ':$FILE_NAME', '::$OBJECT_ID',
    '::$SECURITY_DESCRIPTOR', ':$VOLUME_NAME', ':$VOLUME_INFORMATION', '',
    // ':$DATA',       only needed for sub-attributes with non-resident runs
    ':$INDEX_ROOT', ':$INDEX_ALLOCATION', '::$BITMAP', '::$REPARSE_POINT',
    ':$EA_INFORMATION', '::$EA', '::$PROPERTY_SET', ':$LOGGED_UTILITY_STREAM');

  NTFS_AttribID_Standard_Information = $10;
  NTFS_AttribID_Attribute_List = $20;
  NTFS_AttribID_File_Name = $30;
  NTFS_AttribID_Object_ID = $40;
  NTFS_AttribID_Security_Descriptor = $50;
  NTFS_AttribID_Volume_Name = $60;
  NTFS_AttribID_Volume_Information = $70;
  NTFS_AttribID_Data = $80;
  NTFS_AttribID_Index_Root = $90;
  NTFS_AttribID_Index_Allocation = $A0;
  NTFS_AttribID_Bitmap = $B0;
  NTFS_AttribID_Reparse_Point = $C0;
  NTFS_AttribID_Extended_Attribute_Information = $D0;
  NTFS_AttribID_Extended_Attribute = $E0;
  NTFS_AttribID_Property_Set = $F0;
  NTFS_AttribID_Logged_Utility_Stream = $100;

  /// /////////////////////////////////////////////////////////////////////////////
  /// /////////////////////////////////////////////////////////////////////////////
  /// /////////////////////////////////////////////////////////////////////////////
  /// /////////////////////////////////////////////////////////////////////////////
  /// /////////////////////////////////////////////////////////////////////////////
  /// /////////////////////////////////////////////////////////////////////////////

implementation

type
  PVOID = Pointer;
  HANDLE = Cardinal;

  // Io Status block (see NTDDK.H)
  PIO_STATUS_BLOCK = ^_IO_STATUS_BLOCK;

  _IO_STATUS_BLOCK = record
    Status: NTSTATUS;
    Information: ULONG;
  end;

  // Apc Routine (see NTDDK.H)
  PIO_APC_ROUTINE = procedure(ApcContext: PVOID;
    IoStatusBlock: PIO_STATUS_BLOCK; Reserved: ULONG);

  // The undocumented NtFsControlFile
  // This function is used to send File System Control (FSCTL)
  // commands into file system drivers. Its definition is
  // in ntdll.dll (ntdll.lib), a file shipped with the NTDDK.
  TNtFsControlFile = function(FileHandle: HANDLE; Event: HANDLE; // optional
    ApcRoutine: PIO_APC_ROUTINE; // optional
    ApcContext: PVOID; // optional
    IoStatusBlock: PIO_STATUS_BLOCK; FsControlCode: ULONG; InputBuffer: PVOID;
    // optional
    InputBufferLength: ULONG; OutputBuffer: PVOID; // optional
    OutputBufferLength: ULONG): NTSTATUS; stdcall;

var
  NtFsControlFile: TNtFsControlFile = nil;

function ConvertUnicodeName(const name: PWideChar; Len: integer): AnsiString;
begin
  // Convert the Unicode full name to ANSI.
  SetLength(Result, Len);
  WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK or WC_DISCARDNS or
    WC_SEPCHARS or WC_DEFAULTCHAR, name, -1, @Result, Len, nil, nil);
end;

function WideStringToString(const ws: widestring; Len: integer): AnsiString;
var
  l: integer;
begin
  if ws = '' then
    Result := ''
  else
  begin
    l := WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK or WC_DISCARDNS or
      WC_SEPCHARS or WC_DEFAULTCHAR, @ws[1], -1, nil, 0, nil, nil);
    SetLength(Result, l - 1);
    if l > 1 then
      WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK or WC_DISCARDNS or
        WC_SEPCHARS or WC_DEFAULTCHAR, @ws[1], -1, @Result[1], l - 1, nil, nil);
  end;
end;

function NTFSMFTRecFlagsToStr(Flags: TNTFSMFTRecFlags): string;
// function NTFSMFTRecFlagsToStr(Flags: WORD): string;
var
  fl: Word;
begin
  // if ((flags and 1)=1) then
  if fInUse in Flags then
    Result := Result + 'in use, '
  else
    Result := Result + 'deleted, ';

  // if ((flags and 2)=2) then
  if fDir in Flags then
    Result := Result + 'dir, ';

  if Result <> '' then
    Delete(Result, Length(Result) - 1, 2)
  else
    Result := '(none)';

  fl := Word(Flags);
  Result := Result + ' (0x' + IntToHex(fl, 2) + ' - ' + IntToStr(fl) + ')';
end;

function NTFSAttrFlagsToStr(Flags: TNTFSAttrFlags): string;
begin
  if fCompressed in Flags then
    Result := Result + 'compressed, ';
  if fEncrypted in Flags then
    Result := Result + 'encrypted, ';
  if fSparse in Flags then
    Result := Result + 'sparse, ';
  if Result <> '' then
    Delete(Result, Length(Result) - 1, 2)
  else
    Result := '(none)';
end;

function NTFSPermissionsToStr(Perm: TNTFSPermissions): string;
begin
  if aReadOnly in Perm then
    Result := 'read only, ';
  if aHidden in Perm then
    Result := Result + 'hidden, ';
  if aSystem in Perm then
    Result := Result + 'system, ';
  if aArchive in Perm then
    Result := Result + 'archive, ';
  if aDevice in Perm then
    Result := Result + 'device, ';
  if aNormal in Perm then
    Result := Result + 'normal, ';
  if aTemporary in Perm then
    Result := Result + 'temporary, ';
  if aSparseFile in Perm then
    Result := Result + 'sparse file, ';
  if aReparsePoint in Perm then
    Result := Result + 'reparse point, ';
  if aCompressed in Perm then
    Result := Result + 'compressed, ';
  if aOffline in Perm then
    Result := Result + 'offline, ';
  if aNotIndexed in Perm then
    Result := Result + 'not indexed, ';
  if aEncrypted in Perm then
    Result := Result + 'encrypted, ';
  if Result <> '' then
    Delete(Result, Length(Result) - 1, 2)
  else
    Result := '(none)';
end;

function NTFSTimeToIntAge(Time: Int64): integer;
begin
  result:=FileTime(Time).dwLowDateTime*MAXWORD + FileTime(Time).dwHighDateTime;
end;

function NTFSTimeToStr(Time: Int64): string;
var SysTime: TSystemTime;
    ft: tfiletime;
begin
  FileTimeToSystemTime(FileTime(Time), SysTime);
  Result :=
    IntToStr(SysTime.wYear) + '-' +
    CountStr(SysTime.wMonth,2) + '-'+
    CountStr(SysTime.wDay,2) + ' ' +
    //format('%.2d:%.2d:%.2d',[SysTime.wHour, SysTime.wMinute, SysTime.wSecond]);
    format('%.2d:%.2d',[SysTime.wHour, SysTime.wMinute]);
end;

function NTFSAttribName(Attrib: Cardinal): string;
begin
  if (Attrib mod $10 <> 0) or (Attrib > $100) then
    Result := 'Attribute $' + IntToHex(Attrib, 2)
  else
    Result := AttribNames[Attrib div $10];
end;

function NTFSAttribFileName(MFTentry: TMFTEntry; attrid: Cardinal;
  MFTid: Cardinal; aLevel: integer; aDir: boolean): string;
var
  Win32Name: widestring;
  aname, attrname: AnsiString;
begin
  Result := '';
  if (MFTentry.Attrib[attrid].Hdr.NameLength > 0) and
  // (not (fDir in MFTEntry.FileRecHdr.Flags)) then
  // ((fDir and 2) <> 2) then
    (not aDir) then
  begin
    SetLength(Win32Name, MFTentry.Attrib[attrid].Hdr.NameLength);
    MFTentry.Attrib[attrid].Data.Position := MFTentry.Attrib[attrid]
      .Hdr.NameOffset;
    MFTentry.Attrib[attrid].Data.
      Read(Win32Name[1], MFTentry.Attrib[attrid].Hdr.NameLength * 2);
    aname := WideStringToString(Win32Name,
      MFTentry.Attrib[attrid].Hdr.NameLength);
    if (aLevel > 0) and
      (MFTentry.Attrib[attrid].Hdr.AttributeType = NTFS_AttribID_Data) then
      Result := ':' + aname + ':$DATA'
    else
      Result := ':' + aname;
  end;

  if (MFTentry.Attrib[attrid].Hdr.AttributeType mod $10 <> 0) or
    (MFTentry.Attrib[attrid].Hdr.AttributeType > $100) then
    Result := Result + 'Attribute $' +
      IntToHex(MFTentry.Attrib[attrid].Hdr.AttributeType, 2)
  else
  begin
    if (MFTentry.Attrib[attrid].Hdr.AttributeType =
      NTFS_AttribID_Index_Allocation) and
    // (attrname<>AttribFileNames[MFTentry.Attrib[AttrID].Hdr.AttributeType div $10]) and
    // (fDir in MFTentry.FileRecHdr.Flags)
      ((aDir)) then
      attrname := ''
    else
    begin
      attrname := AttribFilenames
        [MFTentry.Attrib[attrid].Hdr.AttributeType div $10];
      if (MFTentry.Attrib[attrid].Hdr.AttributeType =
        NTFS_AttribID_Index_Allocation) and (attrname <> '') and
        (pos(':$', Result) = 0) then
        attrname := ':' + attrname;
    end;

    if MFTid = 5 then
    begin
      if (aLevel = 0) and
        (MFTentry.Attrib[attrid].Hdr.AttributeType = NTFS_AttribID_Index_Allocation)
      then
      begin
        if attrname <> '' then
        begin
          if (pos(':$', attrname) = 0) then
          begin
            Result := Result + ':' + attrname
          end
          else
          begin
            Result := Result + attrname;
          end;
        end;
      end
      else
      begin
        Result := Result + '$Root' + attrname
      end;
    end
    else
    begin
      Result := Result + attrname;
    end;
  end;

  {
    if (MFTentry.Attrib[AttrID].Hdr.AttributeType=NTFS_AttribID_Index_Allocation) and
    (attrname<>AttribFileNames[MFTentry.Attrib[AttrID].Hdr.AttributeType div $10]) then
    attrname:='';

    if (pos('::$',Result)=0) and (pos(':$',Result)=0) then
    begin
    if (pos('::$',result)=0) and (pos(':$',result)=0) and (result<>'') then
    Result := result+':'+attrname;
    (*
    if aDir then
    Result := ':'+attrname
    else
    Result :=     attrname
    *)
    end else
    Result := result+attrname;
  }

end;

{ TNTFS }
destructor TNTFS.Destroy;
begin
  if bCloseHandleOnDestroy then
    Disk.CloseVolume
  else if (VolHandle <> INVALID_HANDLE_VALUE) and (VolHandle <> 0) then
    CloseHandle(VolHandle);
  VolHandle := 0;
  inherited Destroy;
end;

function TNTFS.ReadClusters(Start, Count: Cardinal): TMemoryStream;
begin
  if Start > BootSec.SectorSize div BootSec.SecPerClust then
    raise EInOutError.Create('Tried to read cluster out of range.');
  Result := TMemoryStream.Create;

  //Disk.Position := Int64(PartInfo.StartSec + BootSec.SecPerClust * Start) * VolumeInfo.BytesPerSector;
  Disk.Position := Int64(PartInfo.StartSec + BootSec.SecPerClust * Start) * VolumeInfo.BytesPerSector;
  Result.CopyFrom(Disk, Int64(BootSec.SecPerClust * Count) * VolumeInfo.BytesPerSector); // 512);

  //Disk.Position      := PartInfo.StartSec + Int64(Start * VolumeInfo.BytesPerCluster); //  cardinal( Word(BootSec.SecPerClust) * BootSec.BytesPerSec )   ); // 512;
  //Result.CopyFrom(Disk, Int64(Count * VolumeInfo.BytesPerCluster));
  Result.Position := 0;
end;

procedure TNTFS.SaveFileWithProgress(RunList: TRunList; FileName: String; filesize: Int64; InfoStdHdr: TNTFSStandardInfoAttribute);
var
  Run: TMemoryStream;
  Stream: TStream;
  buffsize: integer;
  buff: array [0 .. 65535] of Byte;
  TotalClusters, currClusters, Clusters: Int64;
  currbytesread, fragbytesread, bytesread, bytestoread: Int64;
  lastPercent: integer;
  i, clustersatonce: integer;
begin
  // if filename is given then copy with progressbar
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    lastPercent := 0;
    currclusters := 0;
    currbytesread := 0;
    TotalClusters := 0;
    // count totalsize
    for I := 0 to Length(RunList) - 1 do
    begin
      TotalClusters := TotalClusters + RunList[I].Len;
    end;

    clustersatonce := 4;
    // buff:=TMemoryStream.Create;
    // buff.Size:=clustersatonce*512; //BootSec.SecPerClust *

    for I := 0 to Length(RunList) - 1 do
    begin
      if RunList[I].Sparse then
      begin
        // if sparse, no need to read anything, just extend the stream with SetSize
        // Result.SetSize(Result.Position + (RunList[I].Len * BootSec.SecPerClust * 512));
        // Result.Position := Result.Position + (RunList[I].Len * BootSec.SecPerClust * 512);
      end else begin

        // Run := ReadClusters(RunList[I].Lcn, RunList[I].Len);
        // Result.CopyFrom(Run, Run.Size);
        clusters := 0;
        fragbytesread := 0;
                                      // 512 *
        Disk.Position := BootSec.BytesPerSec * (PartInfo.StartSec + BootSec.SecPerClust * RunList[I].Lcn);
        while (clusters < RunList[I].Len) do
        begin
          //
          // buff.Memory;
          bytestoread := BootSec.BytesPerSec * BootSec.SecPerClust *
            clustersatonce;
          bytesread := Disk.Read(buff, bytestoread);

          if clusters + clustersatonce > RunList[I].Len then
          begin
            bytesread := (RunList[I].Len - clusters) * BootSec.BytesPerSec *
              BootSec.SecPerClust;
            Inc(currclusters, RunList[I].Len - clusters);
            Inc(clusters, RunList[I].Len - clusters);
          end
          else
          begin
            Inc(currclusters, clustersatonce);
            Inc(clusters, clustersatonce);
          end;

          if currbytesread + bytesread > filesize then
          begin
            bytesread := filesize - currbytesread;
          end;

          Inc(fragbytesread, bytesread);
          Inc(currbytesread, bytesread);

          Stream.WriteBuffer(buff, bytesread);

          // buff := ReadClusters(pos, rest);
          // Stream.Write(buff, buff.Size);

          if Trunc(currclusters / TotalClusters * 100) > lastPercent then
          begin
            lastPercent := Trunc(currclusters / TotalClusters * 100);
            ProgressWin.ProgBar.Position := lastPercent;
            ProgressWin.lbProgress.Caption:=IntToDottetInt(Currclusters*BootSec.BytesPerSec * BootSec.SecPerClust)+' byte';
            ProgressWin.Refresh;
            Application.processmessages;
            if not ProgressWin.Visible then
              break
          end;
        end;
      end;
    end;
    // buff.Free;
  finally
    Stream.Free;
  end;
  //FileSetDate(FileName, DateTimeToFileDate());
  //NTFSTimeToIntAge(InfoStdHdr.TimeCreated));
  //NTFSTimeToIntAge(
end;

// --------------------------------------------------------------------
// Does directory lookups of NTFS metadata files.
// --------------------------------------------------------------------
procedure TNTFS.DumpMetaFiles(const DriveId: AnsiChar; Results: TStrings);
var
  FileName: String;
  FindData: WIN32_FIND_DATA;
  FindHandle: HANDLE;
  I: integer;
  Line: String;
begin
  // Loop through array of files
  for I := 0 to Length(MetaFileNames) - 1 do
  begin
    FileName := format('%s:\%s', [DriveId, MetaFileNames[I]]);
    FindHandle := FindFirstFile(PChar(FileName), FindData);
    if (FindHandle <> INVALID_HANDLE_VALUE) then
    begin
      Line := format('%-11s    %d bytes',
        [FindData.cFileName, FindData.nFileSizeLow]);
      Windows.FindClose(FindHandle);
      Results.Add(Line);
    end;
  end;
end;

// Open the volume and query its data.
function TNTFS.GetNTFSInfo: boolean;
var
  VolumeName: String;
  VolumeHandle: HANDLE;
  lpBytesReturned: DWORD;
begin
  Result := False;
  if PartInfo.sLetter = '' then
    exit;
  if (@NtFsControlFile = nil) then
  begin
    SysErrMsg(GetLastError(), 'Unsupported version of Windows NT.');
    exit;
  end;

  // open the volume
  VolumeName := format('\\.\%s:', [PartInfo.sLetter]);
  VolumeHandle := CreateFile(PChar(VolumeName), GENERIC_READ, FILE_SHARE_READ or
    FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
  if (DeviceIoControl(VolumeHandle, FSCTL_GET_NTFS_VOLUME_DATA, nil, 0,
    @VolumeInfo, sizeof(TNTFS_VOLUME_DATA_BUFFER), lpBytesReturned, nil) <> True)
  then
  begin
    // SysErrMsg(GetLastError(), 'DiskRead: ');
    Result := False;
  end
  else
  begin
    Result := True;
  end;
  CloseHandle(VolumeHandle);
end;

function TNTFS.ReadFile(Attr: TAttribute; InfoStdHdr: TNTFSStandardInfoAttribute): TMemoryStream;
begin
  if Attr.Hdr.NonResident then
  begin
    Result := ReadRunListData(Attr, '', InfoStdHdr);
    // HdrNonRes.RealSize in sometimes way too big here too... initializedsize seems to have sensible values
    Result.SetSize(Attr.HdrNonResident.RealSize); // StreamInitializedSize);
    Result.Position := 0;
  end
  else
  begin
    Attr.Data.Position := Attr.HdrResident.AttributeOffset;
    Result := TMemoryStream.Create;
    Result.CopyFrom(Attr.Data, Attr.HdrResident.AttributeLength);
    Result.Position := 0;
  end;
end;

function TNTFS.FixupUpdateSequenceArray(Data: array of Byte; NTFSRecHdr: TNTFSFileRecordHeader; MFTid: Int64): boolean;
var
  I, idx: ULONG;
  Inc: ULONG;
  // data: array of byte;
  // usa: PUSHORT;
  sector: array of Word;
  usa: array of Word;
begin
  Result := False;

  if (NTFSRecHdr.UsaOffset + NTFSRecHdr.UsaSize * 2 > 510) then
  begin
    // something wrong in this record
    // DrawText1(2,"MFTRecord #%I64u inconsistent - use CHKDSK!", MFTid);
    exit;
  end;

  SetLength(sector, Length(Data) div 2);
  SetLength(usa, Length(Data) div 2);

  sector := @Data;
  usa := @(Data[NTFSRecHdr.UsaOffset]);
  Inc := Trunc(Disk.DiskGeometry.BytesPerSector / sizeof(SHORT));
  idx := Inc - 1;

  for I := 1 to NTFSRecHdr.UsaSize - 1 do
  begin
    // Check if the last 2 bytes of the sector contain the Update Sequence Number.
    // If not then return FALSE.
    if (sector[idx] <> usa[0]) then
    begin
      // DrawText1(2,"error: MFTRecord #%I64u inconsistent - use CHKDSK!", MFTid);
      Result := False;
      break;
    end;

    // Replace the last 2 bytes in the sector with the value from the Usa array.
    sector[idx] := usa[I];
    idx := idx + Inc;
  end;
end;

{
procedure TNTFS.GetMFTRecordCluster(MFTid: Cardinal; var MFTStart, MFTneeded:Int64);
var I: integer;
begin
  MFTStart:=VolumeInfo.MFTStart.QuadPart;
  MFTneeded:=MFTid * trunc(VolumeInfo.BytesPerMFTRecord/VolumeInfo.BytesPerCluster);
  for I := 0 to Length(MFTRunList) - 1 do
  begin
    if MFTRunList[I].Lcn = 0 then
      continue;
    if (MFTneeded < MFTRunList[I].Len) then
    begin
        MFTstart := MFTRunList[I].Lcn;
        //MFTstartLcn   = MFTstartLcn+MFTneededLcn;
        break;
    end else  begin
        MFTstart := MFTRunList[I].Lcn;
        MFTneeded := MFTneeded - MFTRunList[I].Len;
    end;
    (*
    // if index is NOT in fragment's range
    if (MFTid*2 >= cLcn  * BootSec.SecPerClust) then
    begin
      // if this is last fragment... then there's error
      if I = Length(MFTRunList) - 1 then
        // raise EInOutError.Create('MFT index #' + IntToStr(Index) + ' is beyond MFT file.')
        MFTid := MFTid - ((MFTRunList[I].Len * BootSec.SecPerClust) div 2)
        // otherwise let's decrement the index amount of MFT blocks (in fragment I)
      else
        MFTid := MFTid - ((MFTRunList[I].Len * BootSec.SecPerClust) div 2)
    end else begin
      // ok index is in fragment's range
      FragmentStartClust := MFTRunList[I].Lcn;
      break;
    end;
    *)
  end;
  //Result := FragmentStartClust;
end;

// reads data for MFT entry, afaik, it's always 1 kb eq 2 sectors
function TNTFS.ReadMFTData(MFTid: Cardinal): TMemoryStream;
var
  I: integer;
  DiskPos: Int64;
  ClusterOffset: integer;
  //clusteroffset: Int64;
  MFTStart, MFTneeded: Int64;
begin
  Result := TMemoryStream.Create;
  GetMFTRecordCluster(MFTid, MFTStart, MFTneeded);
  clusteroffset := trunc(VolumeInfo.BytesPerCluster/VolumeInfo.BytesPerMFTRecord);
  clusteroffset := MFTid mod clusteroffset;

  // calc cluster-position
  //(MFTstartLcn + MFTneededLcn) * gNTFSDATA.BytesPerCluster +
  //           clusteroffset * gNTFSDATA.BytesPerFileRecordSegment;
  Disk.Position := (MFTStart + MFTneeded) * VolumeInfo.BytesPerCluster +
                   clusteroffset * VolumeInfo.BytesPerMFTRecord;

  //clusteroffset := VolumeInfo.BytesPerCluster div VolumeInfo.BytesPerMFTRecord;
  //clusteroffset := trunc(VolumeInfo.BytesPerCluster/VolumeInfo.BytesPerMFTRecord);
  //clusteroffset := MFTid mod clusteroffset;

  //10410483712

  //DiskPos := (VolumeInfo.MFTStart.QuadPart + FragmentStartClust) * VolumeInfo.BytesPerCluster;
  //Disk.Position := (DiskPos + MFTid * VolumeInfo.BytesPerMFTRecord);

  //BytePos = (gNTFSDATA.MftStartLcn.QuadPart + FragmentStartClust)*gNTFSDATA.BytesPerCluster+
  //          clusteroffset * gNTFSDATA.BytesPerFileRecordSegment;

  Result.CopyFrom(Disk, Int64(BytesPerMFTRecord));
  // Result.SaveToFile('C:\4.bin');
  // Result.CopyFrom(Disk,Int64(2 * 512));
  Result.Position := 0;
end;
}

function TNTFS.GetMFTRecordCluster(var MFTid: Cardinal): Int64;
var
  FragmentStartClust: Cardinal;
  I: integer;
begin
  FragmentStartClust := 0;
  // run through all runlist-entries
  for I := 0 to Length(MFTRunList) - 1 do
  begin
    if MFTRunList[I].Lcn = 0 then continue;
    // if index is NOT in fragment's range
    if (MFTid * 2 >= MFTRunList[I].Len * BootSec.SecPerClust) then
    begin
      // if this is last fragment... then there's error
      if I = Length(MFTRunList) - 1 then
        // raise EInOutError.Create('MFT index #' + IntToStr(Index) + ' is beyond MFT file.')
        MFTid := MFTid - ((MFTRunList[I].Len * BootSec.SecPerClust) div 2)
      else
        // otherwise let's decrement the index amount of MFT blocks (in fragment I)
        MFTid := MFTid - ((MFTRunList[I].Len * BootSec.SecPerClust) div 2)
    end else begin
      // ok index is in fragment's range
      if i>0 then
        FragmentStartClust := MFTRunList[I].Lcn // - MFTid * 2
      else
        FragmentStartClust := MFTRunList[I].Lcn;
      break;
    end;
  end;
  Result := FragmentStartClust;
end;

function TNTFS.GetMFTRecordSector(MFTid: Cardinal): Integer;
begin
  result := Trunc( GetMFTRecordCluster(MFTid) / BootSec.SecPerClust);
end;

// reads data for MFT entry, afaik, it's always 1 kb eq 2 sectors
// blume: its actually  -  VolumeInfo.BytesPerMFTRecord
function TNTFS.ReadMFTData(MFTid: Cardinal): TMemoryStream;
var I, iDiv: integer;
    FragmentStartClust: Int64;
begin
  Result := TMemoryStream.Create;
  //Result := TMemoryStream.Create;
  // MFTid may change too
  iDiv:=VolumeInfo.BytesPerCluster div VolumeInfo.BytesPerMFTRecord div 2;
  FragmentStartClust:=GetMFTRecordCluster(MFTid);
  Disk.Position := VolumeInfo.BytesPerSector *  // 512 *
                   (PartInfo.StartSec +
                   (Int64(FragmentStartClust * BootSec.SecPerClust) +
                   (MFTid * iDiv)));

  Result.CopyFrom(Disk,Int64(2 * BootSec.BytesPerSec));
  //Result.CopyFrom(Disk,Int64(2 * 512));
  Result.Position := 0;
end;

function TNTFS.ReadMFTDataAPI(MFTid: Cardinal): TMemoryStream;
var
  NTFSRECORDi: TNTFS_FILE_RECORD_INPUT_BUFFER;
  NTFSRECORDo: TNTFS_FILE_RECORD_OUTPUT_BUFFER;
  lpBytesReturned: Cardinal;
begin
  if (VolHandle = INVALID_HANDLE_VALUE) or (VolHandle = 0) then
    OpenVolHandle;

  // SetLength(NTFSRECORDo.FileRecord,VolumeInfo.BytesPerMFTRecord);
  // NTFSRECORDo.FileRecordLength:=VolumeInfo.BytesPerMFTRecord;
  NTFSRECORDi.FileReferenceNumber := MFTid;

  SetLastError(0);
  if (DeviceIoControl(VolHandle, FSCTL_GET_NTFS_FILE_RECORD, @NTFSRECORDi,
      sizeof(TNTFS_FILE_RECORD_INPUT_BUFFER), @NTFSRECORDo,
      sizeof(TNTFS_FILE_RECORD_OUTPUT_BUFFER), // +VolumeInfo.BytesPerMFTRecord-1,
      lpBytesReturned, Nil) = True) then
  begin
    if (NTFSRECORDo.FileReferenceNumber <> MFTid) then
    begin
      SysErrMsg(GetLastError, 'Get_NTFS_FileRecord: ');
      Result := nil;
    end
    else
    begin
      Result := TMemoryStream.Create;
      // NTFSRECORDo.FileRecord
      // Result.CopyFrom(@NTFSRECORDo.FileRecord,VolumeInfo.BytesPerMFTRecord);
      Result.Write(NTFSRECORDo.FileRecord, VolumeInfo.BytesPerMFTRecord);
      // Result.WriteBuffer(NTFSRECORDo.FileRecord,VolumeInfo.BytesPerMFTRecord);
      Result.Position := 0;
    end;
  end else begin
    SysErrMsg(GetLastError, 'GET_NTFS_FILE_RECORD: ');
    Result := nil;
  end;
end;

// creates mft entry object for specified mft entry index
function TNTFS.ReadMFTEntry(MFTid: Cardinal): TMFTEntry;
var MFTdata: TMemorystream;
begin
  Result := nil;
  if (MFTid < MaxMFT) and (MFTid>=0) then
  begin
    //MFTdata := ReadMFTDataAPI(MFTid);
    MFTdata := ReadMFTData(MFTid);
    Result := TMFTEntry.Create(MFTdata, self, MFTid);
    MFTdata.Free;
  end;
end;

procedure TNTFS.OpenVolHandle;
begin
  if (Disk.DiskHandle <> 0) and (Disk.DiskHandle + 1 <> 0) then
  begin
    VolHandle := Disk.DiskHandle;
  end else begin
    // if PartInfo.sLetter<>'' then
    VolHandle := CreateFile(PChar('\\.\' + PartInfo.sLetter + ':'),
      GENERIC_READ, FILE_SHARE_READ OR FILE_SHARE_WRITE, nil, OPEN_EXISTING,
      0, 0);
      //FILE_FLAG_SEQUENTIAL_SCAN OR FILE_FLAG_NO_BUFFERING
  end;
end;


constructor TNTFS.Create(aPartInfo: TPartInfo);
var
  Data: TMemoryStream;
  MFTentry: TMFTEntry;
  I: integer;
begin
  PartInfo := aPartInfo;
  Disk := PartInfo.Disk;
  VolHandle := 0;
  Disk.CloseVolume;

  bCloseHandleOnDestroy := False;
  if (Disk.DiskHandle = 0) or (Disk.DiskHandle = INVALID_HANDLE_VALUE) then
  begin
    if Disk.Physical then
      Disk.OpenVolumePhysical
    else
      Disk.OpenVolumeByLetter(PartInfo.sLetter);
    bCloseHandleOnDestroy := True;
  end;

  // read bootsector
  I := sizeof(BootSec);
  Disk.Position := PartInfo.StartSec * Int64(Disk.DiskGeometry.BytesPerSector);
  // Int64(512);
  // Disk.Position := (PartInfo.StartSec) * Int64(512);
  if Disk.Read(BootSec, 512) = 512 then
  begin
    if (BootSec.SectorSize>0) and (BootSec.SectorSize<Disk.TotalSectors) then
    begin
      if not GetNTFSInfo then
      begin
      //VolumeInfo.BytesPerCluster
      VolumeInfo.TotalClusters.QuadPart:=Bootsec.SecPerClust*aPartInfo.SizeSec; // aPartInfo.SizeSec
      aPartInfo.SizeSec := aPartInfo.SizeSec;
      VolumeInfo.FreeClusters.QuadPart:=0;
      VolumeInfo.BytesPerMFTRecord:=BootSec.BytesPerSec*BootSec.ClustersPerFileRec;
      VolumeInfo.BytesPerCluster:=BootSec.BytesPerSec*BootSec.SecPerClust;
      VolumeInfo.BytesPerSector:=BootSec.BytesPerSec;
      VolumeInfo.ClustersPerMFTRecord:=Bootsec.ClustersPerFileRec;
      VolumeInfo.MFTStart.QuadPart:=Bootsec.MFTClust;
      end; // then exit;

      //VolumeInfo
      //  exit;

      // if Disk.DiskHandle+1 = 0 then
      OpenVolHandle;

      // read MFT's fragments
      Data := TMemoryStream.Create;

      (*
        data.SetSize(512*2);

        //MFTStartSec := (PartInfo.StartSec +
        //               (BootSec.MTFClust * BootSec.SecPerClust));
        //Disk.Position := MFTStartSec * BootSec.BytesPerSec; //Int64(512);
        //Disk.Position := Int64(PartInfo.StartSec + BootSec.SecPerClust + MFTStartSec) * 512;
        //Disk.Position := Int64(PartInfo.StartSec + BootSec.SecPerClust + BootSec.MFTClust) * 512;
        //Disk.Position := Int64(BootSec.MFTClust * Disk.DiskGeometry.BytesPerSector * BootSec.SecPerClust);
        Disk.Position := PartInfo.StartSec + (BootSec.MFTClust * Int64(Disk.DiskGeometry.BytesPerSector) * Int64(BootSec.SecPerClust));
        Disk.Read(data, 512*2);
        //Data.CopyFrom(Disk,512 * 2);
        Data.Position := 0;
      *)

      if (BootSec.ClustersPerFileRec >= 128) then
        BytesPerMFTRecord := Int64(1) shl (256 - BootSec.ClustersPerFileRec)
        // ((ULONG64)1 << (256 - ClustersPerMftRecord))
      else
        BytesPerMFTRecord := BootSec.ClustersPerFileRec *
          Disk.DiskGeometry.BytesPerSector * BootSec.SecPerClust;

      VolumeInfo.BytesPerMFTRecord:=BytesPerMFTRecord;
      MFTStartSec :=
        (PartInfo.StartSec + (BootSec.MFTClust * Int64(BootSec.SecPerClust))) *
        Int64(BootSec.BytesPerSec);
      VolumeInfo.MFTStart.QuadPart:=MFTStartSec;
      // read MFT's fragments
      Data := TMemoryStream.Create;
      Disk.Position := MFTStartSec;
      Data.CopyFrom(Disk, BytesPerMFTRecord);
      Data.Position := 0;
      // Data.SaveToFile('c:\1.bin');
      MFTentry := TMFTEntry.Create(Data, self, 0);
      if MFTentry.FileRecHdr.FileID = 'FILE' then
      begin
        MFTRunList:=ReadRunList(MFTentry.Attrib[MFTentry.IndexOfAttrib(NTFS_AttribID_Data)]);
        // MaxMFT     := (GetFileSize div 512) div 2;
        // MaxMFT     := (MFTEntry.GetFileSize div BootSec.BytesPerSec) shr 1;
        MaxMFT := (MFTentry.GetFileSize div BootSec.BytesPerSec) div 2;
        VolumeInfo.MFTLength.QuadPart:=MFTentry.GetFileSize;
      end;
      MFTentry.Free;

    end else begin
      MessageDlg('Could not read BOOT-sector.', mtInformation, [mbOK], 0);
    end;
  end;

end;

// --------------------------------------------------------------------
// -------------------------------------------------------
{ TMFTEntry }
// -------------------------------------------------------
// --------------------------------------------------------------------

constructor TMFTEntry.Create(Src: TMemoryStream; fNTFS: TNTFS; fMFTid: Cardinal);
var AttribHdr: TNTFSStandardAttributeHeader;
    BaseAddr: Word;
    wstr: String;
    // sstr:string;
begin
  NTFS := fNTFS;
  MFTid := fMFTid;
  Src.Position := 0;

  // MFTdata.Position:=0;
  // MFTData.ReadBuffer(FileRecHdr,SizeOf(FileRecHdr));
  // SetLength(wstr,4);
  // Pointer(@FileRecHdr):=; //@MFTData[0];
  // src.Position:=0;
  // Src.ReadBuffer(FileRecHdr,SizeOf(FileRecHdr));
  // FileRecHdr := TNTFSFileRecordHeader(MFTData.Memory^);
  Src.Position := 0;
  Src.ReadBuffer(FileRecHdr, sizeof(FileRecHdr));
  Src.Position := 0;

  // Move(src.Memory, PTNTFSFileRecordHeader(@FileRecHdr[0])^, SizeOf(FileRecHdr));
  // PChar(@Buffer[0])^

  wstr := Copy(FileRecHdr.FileID, 0, 4);
  // sstr:=ConvertUnicodeName(PWideChar(wstr),4);

  if (FileRecHdr.FileID <> 'FILE') and (FileRecHdr.FileID <> 'BAAD') then
  begin
    // MessageBox(0,'MFT record ID is invalid.','Error',MB_ICONERROR);
    // Abort;
    exit;
  end
  else if (FileRecHdr.FileID <> 'BAAD') then
  begin
    AttribCount := 0;
    Src.Position := FileRecHdr.OffsetToFirstAttr;
    if (Src.Position < NTFS.BytesPerMFTRecord) then
    begin
        BaseAddr := Src.Position;
        Src.ReadBuffer(AttribHdr, sizeof(AttribHdr));
        FillChar(InfoStdHdr,sizeof(InfoStdHdr),0);
        while (AttribHdr.AttributeType <> $FFFFFFFF) do
        begin
          SetLength(Attrib, AttribCount + 1);
          Attrib[AttribCount].Hdr := AttribHdr;
          if AttribHdr.NonResident then
            Src.ReadBuffer(Attrib[AttribCount].HdrNonResident,
              sizeof(Attrib[AttribCount].HdrNonResident))
          else
            Src.ReadBuffer(Attrib[AttribCount].HdrResident,
              sizeof(Attrib[AttribCount].HdrResident));

          if AttribHdr.AttributeType = NTFS_AttribID_Standard_Information then
            Src.ReadBuffer(InfoStdHdr,SizeOf(InfoStdHdr));

          Src.Position := BaseAddr;
          Attrib[AttribCount].Data := TMemoryStream.Create;
          Attrib[AttribCount].Data.CopyFrom(Src, AttribHdr.Size);
          Inc(AttribCount);
          BaseAddr := Src.Position;
          if Src.Size - Src.Position < sizeof(AttribHdr) then
            break;
          Src.ReadBuffer(AttribHdr, sizeof(AttribHdr));
          AttribHdr.Size := AttribHdr.Size and $0000FFFF;
        end;
    end;
  end;
end;

destructor TMFTEntry.Destroy;
var
  I: integer;
begin
  for I := 0 to AttribCount - 1 do
  begin
    Attrib[I].Data.Free;
  end;
  SetLength(Attrib, 0);
  inherited Destroy;
end;

function TMFTEntry.GetFileNameAttr: TNTFSFileNameAttribute;
var
  I: integer;
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
begin
  for I := 0 to AttribCount - 1 do
  begin
    if Attrib[I].Hdr.AttributeType = NTFS_AttribID_File_Name then
    begin
      Attrib[I].Data.Position := 0;
      Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
      Attrib[I].Data.ReadBuffer(Result, sizeof(Result));
      break;
    end;
  end;
end;

function TMFTEntry.IndexOfAttrib(AttributeType: Word): integer;
var
  I: integer;
begin
  Result := -1;
  for I := 0 to AttribCount - 1 do
    if Attrib[I].Hdr.AttributeType = AttributeType then
    begin
      Result := I;
      break;
    end;
end;

/// /////////////////////////////////////////////////////////////////////////////
/// /////////////////////////////////////////////////////////////////////////////

function TMFTEntry.GetFileSize: Int64;
var
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
  AttrNonResident: TNTFSStandardAttributeHeader_NonResident;
  NameAttr: TNTFSFileNameAttribute;
  ListRec: TNTFSAttributeListRecord;

  MFTentry2: TMFTEntry;

  dsize, fsize, newparent: Int64;
  addr, mftno: Cardinal;
  I, l: integer;
  aDir2: boolean;
  newname, Win32Name, dirname: string;
  M: TMemoryStream;

begin
  dsize := 0;
  fsize := 0;
  for I := 0 to AttribCount - 1 do
  begin
    if Attrib[I].Hdr.AttributeType = NTFS_AttribID_File_Name then
    begin
      Attrib[I].Data.Position := 0;
      Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
      Attrib[I].Data.ReadBuffer(NameAttr, sizeof(NameAttr));
      if fsize = 0 then
        fsize := NameAttr.RealSize;
    end;

    if Attrib[I].Hdr.AttributeType = NTFS_AttribID_Data then
    begin
      Attrib[I].Data.Position := 0;
      Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      if Attrib[I].Hdr.NonResident then
        dsize := Attrib[I].HdrNonResident.RealSize
      else
        dsize := Attrib[I].HdrResident.AttributeLength;
    end;

    if Attrib[I].Hdr.AttributeType = NTFS_AttribID_Attribute_List then
    begin
      Attrib[I].Data.Position := 0;
      Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));

      if not Attrib[I].Hdr.NonResident then
      begin
        // resident AttributeList
        Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
        Attrib[I].Data.ReadBuffer(ListRec, sizeof(ListRec));

        addr := AttrResident.AttributeOffset;
        // Attrib[I].Data.Position;

        // run trough attrlist-entries and get runlist
        while (Attrib[I].Data.Size - Attrib[I].Data.Position >
          sizeof(ListRec)) do
        begin
          Attrib[I].Data.Position := addr + ListRec.Size;
          addr := Attrib[I].Data.Position;
          Attrib[I].Data.Read(ListRec, sizeof(ListRec));

          mftno := (ListRec.FileRef and $FFFFFFFFFF);
          if (mftno > 0) and (mftno <> 0) and (mftno <> 5) and (mftno <> MFTid)
            and (mftno < NTFS.MaxMFT) then
          begin
            MFTentry2 := nil; // just in case line below causes exception
            MFTentry2 := NTFS.ReadMFTEntry(mftno);
            // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
            if MFTentry2 <> nil then
            begin
              fsize := MFTentry2.GetFileSize;
              if (fsize <> $FFFFFFFF) and (fsize > 0) then
                break;
              (*
                if mftid=118045 then
                mftid:=118045;
                aDir2:=fDir in FileRecHdr.Flags;
                //newparent:=MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;
                newparent:=MFTentry2.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

                Win32Name := NTFS.GetReferencedDir(newparent, '',aDir2); //dirname, aDir2);
                dirname:=win32name;
                //if aDir2 then dirname:=dirname+'\';
                //ReadRunList2(mftno,dirname,Runlist,aLevel+1,false); //runlistflag
              *)

              (* *-)
                L:=Length(RunList);
                SetLength(RunList,L+1);
                RunList[L].desc:=dirname + NTFSAttribFileName(MFTentry1, i);
                (* *)
              MFTentry2.Free;
            end;

          end;

        end; // while

      end else begin

        // Non-Resident AttributeList with RUN-Lists NTFS_AttribID_Attribute_List

        if dirname = '' then
          continue;

        (* *)
        M := NTFS.ReadRunListData(Attrib[I], '', MFTentry2.InfoStdHdr);
        addr := M.Position;
        FillChar(ListRec, sizeof(ListRec), 0);
        //
        if (Attrib[I].HdrNonResident.RealSize<NTFS.PartInfo.SizeSec*NTFS.BootSec.BytesPerSec) then
        while (Attrib[I].HdrNonResident.RealSize - M.Position > sizeof(ListRec)) do
        begin
          M.Position := addr + ListRec.Size;
          addr := M.Position;
          M.Read(ListRec, sizeof(ListRec));
          if (ListRec.AttributeType = 0) then
            break;

          mftno := (ListRec.FileRef and $FFFFFFFFFF);
          if (mftno > 0) and (mftno <> 0) and (mftno <> 5) // and
          // (mftno<>MFTid) and (mftno<maxmft) and
          then
          begin
            MFTentry2 := nil; // just in case line below causes exception
            MFTentry2 := NTFS.ReadMFTEntry(mftno);
            // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
            if MFTentry2 <> nil then
            begin
              fsize := MFTentry2.GetFileSize;
              if (fsize <> $FFFFFFFF) and (fsize > 0) then
                break;

              (*
                aDir2:=fDir in FileRecHdr.Flags;
                newparent:=MFTentry2.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;
                //Win32Name := GetReferencedDir(newparent, dirname, aDir2);
                //dirname:=win32name;
                if aDir2 then dirname:=dirname+'\';
                MFTentry2.GetFileSize;
                //NTFS.ReadRunList2(mftno,dirname,Runlist,aLevel+1,false); //runlistflag
              *)
              MFTentry2.Free;
            end;

          end;
        end; // while not end of ListRec
        M.free;

      end;
    end; // NTFS_AttribID_Attribut_List

  end;

  if dsize <> 0 then
    Result := dsize
  else
    Result := fsize;
end;

/// /////////////////////////////////////////////////////////////////////////////
/// /////////////////////////////////////////////////////////////////////////////

function TMFTEntry.GetFileName(var parent: Cardinal): string;
var
  Attr: TNTFSStandardAttributeHeader;
  AttrHdrResident: TNTFSStandardAttributeHeader_Resident;
  NameAttr: TNTFSFileNameAttribute;
  OtherName, Win32Name: widestring;
  AOtherName, AWin32Name: AnsiString;
  I: integer;
  MFTentry2: TMFTEntry;
begin
  for I := 0 to AttribCount - 1 do
  begin
    if Attrib[I].Hdr.AttributeType = NTFS_AttribID_File_Name then
    begin
      Attrib[I].Data.Position := 0;
      Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      Attrib[I].Data.ReadBuffer(AttrHdrResident, sizeof(AttrHdrResident));
      Attrib[I].Data.ReadBuffer(NameAttr, sizeof(NameAttr));
      if NameAttr.FileNamespace = npWin32 then
      begin
        SetLength(Win32Name, NameAttr.FileNameLength);
        Attrib[I].Data.Read(Win32Name[1], NameAttr.FileNameLength * 2);
        AWin32Name := WideStringToString(Win32Name, NameAttr.FileNameLength);
        parent := NameAttr.RefToParentDir;
        break;
      end else begin
        SetLength(OtherName, NameAttr.FileNameLength);
        Attrib[I].Data.Read(OtherName[1], NameAttr.FileNameLength * 2);
        AOtherName := WideStringToString(OtherName, NameAttr.FileNameLength);
        parent := NameAttr.RefToParentDir;
      end;
    end;
  end;

  if AWin32Name <> '' then
    Result := AWin32Name
  else if AOtherName <> '' then
    Result := AOtherName
  else
    Result := '';

  (**)
  if (Result = '') and (MFTid <> 0) then
  begin
    MFTentry2 := nil; // just in case line below causes exception
    MFTentry2 := NTFS.ReadMFTEntry(FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF);
    // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
    if MFTentry2 <> nil then
    begin
      // Win32Name := GetReferencedDir(parent, Win32Name);
      // recursiv call - rekursiver aufruf
      Result := MFTentry2.GetFileName(parent);
      // ReadRunList2(mftno,dirname,Runlist,aLevel+1,false); //runlistflag
      MFTentry2.Free;
    end;
  end;
  (* *)

end;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

function TNTFS.GetReferencedDir(MFTid: Cardinal; dirname: String; var bDir: boolean): String;
var
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
  AttrNonResident: TNTFSStandardAttributeHeader_NonResident;
  NameAttr: TNTFSFileNameAttribute;
  ListRec: TNTFSAttributeListRecord;
  MFTentry1, MFTentry2: TMFTEntry;
  I: integer;
  Win32Name: widestring;
  addr, mftno: Cardinal;
  parent: Cardinal;
  M: TMemoryStream;
  aDir: boolean;
  dw: DWORD;
begin
  Result := '';

  MFTentry1 := nil;
  MFTentry1 := ReadMFTEntry(MFTid);
  if MFTentry1 = nil then
    exit;
  aDir := fDir in MFTentry1.FileRecHdr.Flags;
  // aDir:=((MFTentry1.FileRecHdr.Flags and 2)=2);
  // I := MFT.IndexOfAttrib(NTFS_AttribID_File_Name);

  for I := 0 to MFTentry1.AttribCount - 1 do
  begin
    if MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_File_Name then
    begin
      // dw:=WORD(MFTentry1.FileRecHdr.Flags);
      if (fDir in MFTentry1.FileRecHdr.Flags) // or
      // ( dw >= 2 )
      then
        Win32Name := '\' + MFTentry1.GetFileName(parent) + dirname;

      MFTentry1.Attrib[I].Data.Position := 0;
      MFTentry1.Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      MFTentry1.Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
      MFTentry1.Attrib[I].Data.ReadBuffer(NameAttr, sizeof(NameAttr));
      // if NameAttr.RefToParentDir = npWin32 then

      mftno:=(NameAttr.RefToParentDir and $FFFFFFFFFF);
      if (mftno <> 5) and (NameAttr.RefToParentDir and $FFFFFFFFFF >= 0) and (mftno < MaxMFT) and (mftno <> MFTid) and
         (System.pos(widestring(':\'), Win32Name) < 1) // PWideChar(
      then
        Win32Name := GetReferencedDir(mftno, Win32Name, aDir);

      if (mftno = 5) then
        if Win32Name <> PartInfo.sLetter + ':' then
          Win32Name := PartInfo.sLetter + ':' + Win32Name;
    end;
  end;

  if (Win32Name = '') then
  begin
    for I := 0 to MFTentry1.AttribCount - 1 do
    begin
      if MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Attribute_List
      then
      begin
        if MFTentry1.Attrib[I].Hdr.NonResident then
        begin
            MFTentry1.Attrib[I].Data.Position := 0;
            MFTentry1.Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
            MFTentry1.Attrib[I].Data.ReadBuffer(AttrNonResident, sizeof(AttrNonResident));

            M := ReadRunListData(MFTentry1.Attrib[I], '',MFTentry1.InfoStdHdr);
            addr := M.Position;
            FillChar(ListRec, sizeof(ListRec), 0);
            if (MFTentry1.Attrib[I].HdrNonResident.RealSize<PartInfo.SizeSec*BootSec.BytesPerSec) then
            while (MFTentry1.Attrib[I].HdrNonResident.RealSize - M.Position > sizeof(ListRec)) do
            begin
              M.Position := addr + ListRec.Size;
              addr := M.Position;
              M.Read(ListRec, sizeof(ListRec));
              if (ListRec.AttributeType = 0) then
                break;

              mftno := (ListRec.FileRef and $FFFFFFFFFF);
              if (mftno > 0) and (mftno <> 0) and (mftno <> 5) and
                 (mftno <> MFTid) and (mftno < MaxMFT) then
              begin
                MFTentry2 := nil; // just in case line below causes exception
                MFTentry2 := ReadMFTEntry(mftno);
                // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
                if MFTentry2 <> nil then
                begin
                  // Win32Name := GetReferencedDir(parent, Win32Name);
                  MFTentry2.GetFileName(mftno);
                  // ReadRunList2(mftno,dirname,Runlist,aLevel+1,false); //runlistflag
                  MFTentry2.Free;
                end;
              end;
            end;
            M.Free;
            M:=nil;

        end else begin

          MFTentry1.Attrib[I].Data.Position := 0;
          MFTentry1.Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
          MFTentry1.Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
          MFTentry1.Attrib[I].Data.ReadBuffer(ListRec, sizeof(ListRec));

          addr := AttrResident.AttributeOffset;
          // Attrib[I].Data.Position;
          while (MFTentry1.Attrib[I].Data.Size - MFTentry1.Attrib[I].Data.Position > sizeof(ListRec)) and
                (Result = '') do
          begin
            MFTentry1.Attrib[I].Data.Position := addr + ListRec.Size;
            addr := MFTentry1.Attrib[I].Data.Position;
            MFTentry1.Attrib[I].Data.Read(ListRec, sizeof(ListRec));

            mftno := (ListRec.FileRef and $FFFFFFFFFF);
            if (mftno > 0) then // and (mftno<ntfs.maxmft) then
            begin
              MFTentry2 := nil; // just in case line below causes exception
              MFTentry2 := ReadMFTEntry(mftno);
              // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
              if MFTentry2 <> nil then
              begin
                // Result:=MFTentry2.GetFileName;
                Win32Name := MFTentry2.GetFileName(parent);
                // Result := MFTentry2.GetFileNameWithParent(parent);
                if Win32Name <> '' then
                begin
                  Win32Name := '\' + MFTentry2.GetFileName(parent) + dirname;
                  // if (parent<>0) and (parent<>5) then
                  // Win32Name := GetReferencedDir(parent, Win32Name);
                  if parent = 5 then
                    Win32Name := PartInfo.sLetter + ':' + Win32Name;

                  MFTentry2.Free;
                  MFTentry2:=nil;
                  break;
                end;
                MFTentry2.Free;
                MFTentry2:=nil;
              end;
            end;

          end; // while
        end;
      end;
    end; // for
  end;

  mftno := (MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF);
  if (mftno<>mftid) and (mftno > 0) and (mftno <> 0) and (mftno <> 5) and (mftno < MaxMFT) and
     (System.pos(widestring(':\'), Win32Name) < 1)
  then
  begin
       if (Win32Name = '') then
       begin
         //Win32Name := GetReferencedDir(mftno, Win32Name, aDir);
       end;
       if (Win32Name = '') and (MFTentry1.AttribCount = 1) and (MFTentry1.Attrib[0].Hdr.AttributeType = NTFS_AttribID_Data) then
       begin
         Win32Name := GetReferencedDir(mftno, Win32Name, aDir);
       end;
       if (Win32Name = '') and (MFTentry1.AttribCount = 1) and (MFTentry1.Attrib[0].Hdr.AttributeType = NTFS_AttribID_Index_Allocation) then
       begin
         Win32Name := GetReferencedDir(mftno, Win32Name, aDir);
       end;
  end;
  MFTentry1.Free;
  MFTentry1:=nil;

  // if Win32Name = '\.' then Win32Name:='';
  Result := Win32Name;
  if (Length(Result) = 4) and (Result[4] = '.') then
    Result := copy(Result, 1, 3);
  if (Length(Result) = 2) then
    Result := Result + '\';
end;

/// /////////////////////////////////////////////////////////////////////////////
/// /////////////////////////////////////////////////////////////////////////////

function ToSigned(Value: integer; Size: Byte): Cardinal;
// makes bitmask for xor...
  function MakeMask(Size: Byte): integer;
  var
    I: integer;
  begin
    Result := 0;
    for I := 0 to (Size * 8) - 1 do
      Result := Result or (1 shl I);
  end;

begin
  if boolean((Value shr ((Size * 8) - 1)) and 1) then
    Result := -1 * ((Value xor MakeMask(Size)) + 1)
  else
    Result := Value;
end;

// reads a RunList entries into dynamic array (TRunList)
function TNTFS.ReadRunList(DataAttr: TAttribute): TRunList;
var
  B: Byte;
  BytesNeeded: Byte;
  ByteOffset: Byte;
  OldByteOffset: Cardinal;
  RunOffset: Cardinal;
  RunLength: Cardinal;
  RunVcn: Cardinal;
  M: TMemoryStream;
  l: integer;
  Data: array [0 .. 1023] of Byte;
begin
  if not DataAttr.Hdr.NonResident then
    raise EInOutError.Create('Tried to read runlist of resident attribute.');
  M := DataAttr.Data;

  // M.Position := DataAttr.HdrNonResident.DataRunsOffset;
  // SetLength(data, VolumeInfo.BytesPerMFTRecord);
  // if PartInfo.sLetter<>'' then
  // disk.OpenVolumePhysical;
  // disk.Handle:=disk.Handle;
  // M.Read(data, VolumeInfo.BytesPerMFTRecord);

  RunVcn := DataAttr.HdrNonResident.StartVCN;
  M.Position := DataAttr.HdrNonResident.DataRunsOffset;

  SetLength(Result, 0);
  RunOffset := 0;
  repeat
    // if RunOffset=244675 then
    // RunOffset:=RunOffset;

    M.ReadBuffer(B, 1);
    if (B <> 0) and (DataAttr.Hdr.Size - M.Position > 2) then
    begin
      // new RunOffset will be added relatively to old one...
      OldByteOffset := RunOffset;
      BytesNeeded := B and $0F;
      if BytesNeeded >= 5 then
        BytesNeeded := 1;
      ByteOffset := (B and $F0) shr 4;

      if (DataAttr.Hdr.Size < M.Position + BytesNeeded + 2) then
        break;

      RunOffset := 0;
      RunLength := 0;

      M.ReadBuffer(RunLength, BytesNeeded);
      M.ReadBuffer(RunOffset, ByteOffset);
      RunOffset := OldByteOffset + ToSigned(RunOffset, ByteOffset);

      l := Length(Result);
      SetLength(Result, l + 1);
      if ByteOffset = 0 then
      begin
        Result[l].Sparse := True
      end
      else
      begin
        Result[l].Sparse := False;
        Result[l].Lcn := RunOffset;
      end;
      Result[l].Len := RunLength;
      Result[l].VCN := RunVcn;
      // Result[L].MFTid:=
      Inc(RunVcn, RunLength);

    end;
  until (B = 0) or (DataAttr.Hdr.Size - M.Position <= 2);
end;

// ReadRunListData into memory stream
function TNTFS.ReadRunListData(Attr: TAttribute; FileName: String; InfoStdHdr: TNTFSStandardInfoAttribute): TMemoryStream;
var
  RunList: TRunList;
  I, j: integer;
  Run: TMemoryStream;

  Stream: TStream;
  buffsize: integer;
  buff: TMemoryStream;
  pos, gpos: Int64;
  Size, last, clust: Int64;
  rest, hlpI: Int64;
  clusters: integer;
begin
  RunList := ReadRunList(Attr);
  if (FileName <> '') then
  begin
    SaveFileWithProgress(RunList, FileName, Attr.HdrNonResident.RealSize, InfoStdHdr);
    Result := nil;
  end else begin
    Result := TMemoryStream.Create;
    for I := 0 to Length(RunList) - 1 do
    begin
      if RunList[I].Sparse then
      begin
        // if sparse, no need to read anything, just extend the stream with SetSize
        Result.SetSize(Result.Position +(RunList[I].Len * BootSec.SecPerClust * 512));
        Result.Position := Result.Position +(RunList[I].Len * BootSec.SecPerClust * 512);
      end else begin
        if (RunList[I].Len > 0) and (int64(RunList[I].Lcn)<VolumeInfo.TotalClusters.QuadPart) and
           (RunList[I].Lcn>=0)
        then
        begin
          (* *)
          Run := ReadClusters(RunList[I].Lcn, RunList[I].Len);
          // if Run.Size<10000000 then
          Result.SetSize(Result.Size + Run.Size);
          // Result.Position := Result.Size;
          Result.CopyFrom(Run, Run.Size);
          Run.Free;
          (* *)
        end;
      end;
    end;
    Result.Position := 0;
  end;
end;

function AlreadyAdded(var zwRunlist: TRunList; RunList: TRunList): boolean;
var
  j, l: integer;
begin
  // check if this Runlist was already added
  Result := False;
  l := Length(RunList);
  if l < 1 then
    exit;
  for j := 0 to l - 1 do
  begin
    if RunList[j].Lcn = zwRunlist[0].Lcn then
    begin
      Result := True;
      break;
    end;
  end;
end;

function TNTFS.AddToRunList(var MFTid, parent: Cardinal; MFTentry1: TMFTEntry;
  attrid: integer; wantedVCN: Cardinal; var RunList: TRunList;
  dirname, attrname: String): boolean;
var
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
  NameAttr: TNTFSFileNameAttribute;
  ListRec: TNTFSAttributeListRecord;

  M: TMemoryStream;
  // MFTentry1, MFTentry2: TMFTEntry;
  zwRunlist: TRunList;
  addr, mftno, newMFTid: Cardinal;
  newparent: Cardinal;
  j, K, l, zwL, idx: integer;
  docontinue, aDir, aDir2: boolean;
  Win32Name, newpath: string;
begin
  Result := False;
  (*
    l:=Length(zwRunList);
    if attrname<>'' then
    zwRunList[0].desc:=attrname
    else if zwRunList[0].desc='' then
    zwRunList[0].desc:='\\\\\\\\';

    if parent=INVALID_HANDLE_VALUE then
    parent:=MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

    if AlreadyAdded(zwRunList,Runlist) then continue;
    L:=Length(RunList);
    if L=0 then continue;

    zwRunList:=ReadRunList(MFTentry1.Attrib[I]);
    zwL:=Length(zwRunList);
    if zwL=0 then continue;

    L  :=Length(  RunList);
    zwL:=Length(zwRunList);
    SetLength(RunList,L+zwL); //+1
    for J:=0 to zwL-1 do
    begin
    //if (dirname<>zwRunList[J].desc) then
    //if (aLevel<>0) and (dirname<>zwRunList[J].desc) then
    if (zwRunList[J].Vcn=0) then
    RunList[L+J].desc   :=zwRunList[J].desc;
    RunList[L+J].Lcn    :=zwRunList[J].Lcn;
    RunList[L+J].Len    :=zwRunList[J].Len;
    RunList[L+J].Vcn    :=zwRunList[J].Vcn;
    RunList[L+J].Sparse :=zwRunList[J].Sparse;
    end;
  *)
  (*
    docontinue:=false;
    for J:=0 to zwL-1 do
    begin
    if RunList[J].desc=dirname then
    begin
    docontinue:=true;
    break;
    end;
    end;
    if docontinue then
    continue;
  *)

  if parent = INVALID_HANDLE_VALUE then
    parent := MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

  // L:=Length(RunList);
  // if L=0 then exit;

  zwRunlist := ReadRunList(MFTentry1.Attrib[attrid]);
  zwL := Length(zwRunlist);
  if zwL = 0 then
    exit;

  (*
    if ((L>0) and
    ((zwRunList[0].Vcn=RunList[L-1].Vcn+RunList[L-1].Len+1) or
    (zwRunList[0].Vcn=RunList[L-1].Vcn)                    or
    (Length(zwRunList)>0) and (zwRunList[0].Vcn=0))
    )
    then
    continue;
  *)

  if AlreadyAdded(zwRunlist, RunList) then
    exit;
  if pos('::$', dirname) > 0 then
    zwRunlist[0].desc := dirname
  else
    zwRunlist[0].desc := dirname + attrname;

  if False and (RunList[l - 1].desc = '') then
  begin
    RunList[l - 1].desc := '::noname::';
    (* *-)
      if MFTentry1.Attrib[I].HdrNonResident.StartVCN<>0 then
      RunList[L].desc:='<continued runlist>'
      else
      RunList[L].desc:='<initial runlist>';
      (* *)
  end;

  l := Length(RunList);
  zwL := Length(zwRunlist);
  SetLength(RunList, l + zwL);
  for j := 0 to zwL - 1 do
  begin
    // if (dirname<>zwRunList[J].desc) then
    // if (aLevel<>0) and (dirname<>zwRunList[J].desc) then
    if (zwRunlist[j].VCN = 0) then
      RunList[l + j].desc := zwRunlist[j].desc;
    // RunList[L+J].desc   :=zwRunList[J].desc;
    RunList[l + j].Lcn := zwRunlist[j].Lcn;
    RunList[l + j].Len := zwRunlist[j].Len;
    RunList[l + j].VCN := zwRunlist[j].VCN;
    RunList[l + j].Sparse := zwRunlist[j].Sparse;
  end;
  Result := True;
end;

function TNTFS.AttributeProc(MFTentry1: TMFTEntry; I: integer; MFTid, parent: Cardinal;
  wantedVCN: Cardinal; aLevel: integer; dirname: widestring; var RunList: TRunList): boolean;
var
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
  NameAttr: TNTFSFileNameAttribute;
  ListRec: TNTFSAttributeListRecord;
  MFTentry2: TMFTEntry;
  zwRunlist: TRunList;
  M: TMemoryStream;
  addr, mftno: Cardinal;
  newparent, newMFTid: Cardinal;
  docontinue: boolean;
  aDir, aDir2: boolean;
  j, l, zwL: integer;
  Win32Name, newname, attrname: string;
begin
  // aDir:=(MFTentry1.FileRecHdr.Flags and 2) = 2;
  aDir := fDir in MFTentry1.FileRecHdr.Flags;
  attrname := NTFSAttribFileName(MFTentry1, I, MFTid, aLevel, aDir);

  Result := True;
  if not MFTentry1.Attrib[I].Hdr.NonResident then
  begin
    // resident AttributeAttributeList
    MFTentry1.Attrib[I].Data.Position := 0;
    MFTentry1.Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
    MFTentry1.Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
    MFTentry1.Attrib[I].Data.ReadBuffer(ListRec, sizeof(ListRec));

    addr := AttrResident.AttributeOffset;
    // Attrib[I].Data.Position;

    // run trough attrlist-entries and get runlist
    while (MFTentry1.Attrib[I].Data.Size - MFTentry1.Attrib[I].Data.Position >
      sizeof(ListRec)) do
    begin
      MFTentry1.Attrib[I].Data.Position := addr + ListRec.Size;
      addr := MFTentry1.Attrib[I].Data.Position;
      MFTentry1.Attrib[I].Data.Read(ListRec, sizeof(ListRec));

      mftno := (ListRec.FileRef and $FFFFFFFFFF);
      if (mftno > 0) and (mftno <> 0) and (mftno <> 5) and (mftno <> MFTid) and
        (mftno < MaxMFT) then
      begin
        MFTentry2 := nil; // just in case line below causes exception
        MFTentry2 := ReadMFTEntry(mftno);
        // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
        if MFTentry2 <> nil then
        begin

          // aDir2:=(MFTentry2.FileRecHdr.Flags and 2) = 2 ;//in MFTentry2.FileRecHdr.Flags);
          aDir2 := fDir in MFTentry2.FileRecHdr.Flags;
          // newparent:=MFTentry2.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;
          newparent := MFTentry2.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

          if newparent<>mftid then
          Win32Name := GetReferencedDir(newparent, '', aDir2);
          // dirname, aDir2);
          // dirname:=win32name + attrname;
          // if aDir2 then dirname:=dirname+'\';
          // if dirname='' then
          // wantedVCN:=ListRec.StartVCN;
          // attrname:=NTFSAttribFileName(MFTentry1,I,MFTid,aLevel,aDir);
          Win32Name := ExtractFileNameCutStream(dirname);
          ReadRunList2(mftno, MFTid, wantedVCN, aLevel + 1,
            Win32Name + attrname, RunList, True); // runlistflag

          MFTentry2.Free;
        end;
      end;
    end; // while

  end else begin
    // Non-Resident AttributeList with RUN-Lists NTFS_AttribID_Attribute_List

    if dirname = '' then
      Result := False;

    (* *)
    if (mftid<>8) and (MFTentry1.GetFileSize<VolumeInfo.MFTLength.QuadPart) then
    begin
        M := ReadRunListData(MFTentry1.Attrib[I],'',MFTentry1.InfoStdHdr);
        AddToRunList(MFTid, parent, MFTentry1, I, wantedVCN, RunList, dirname, attrname);
        addr := M.Position;
        FillChar(ListRec, sizeof(ListRec), 0);
        if (MFTentry1.Attrib[I].HdrNonResident.RealSize<PartInfo.SizeSec*BootSec.BytesPerSec) then
        while MFTentry1.Attrib[I].HdrNonResident.RealSize - M.Position > sizeof(ListRec) do
        begin
          M.Position := addr + ListRec.Size;
          addr := M.Position;
          M.Read(ListRec, sizeof(ListRec));
          if (ListRec.AttributeType = 0) or (ListRec.Size = 0) then
            break;

          mftno := (ListRec.FileRef and $FFFFFFFFFF);
          if (mftno > 0) and (mftno <> 0) and (mftno <> 5) and (mftno <> MFTid) and
            (mftno < MaxMFT) then
          begin
            MFTentry2 := nil; // just in case line below causes exception
            MFTentry2 := ReadMFTEntry(mftno);
            // MFTentry2.ntfs:=TNTFS.Create(NTFS.PartInfo);;
            if MFTentry2 <> nil then
            begin
              // aDir2:=(MFTentry2.FileRecHdr.Flags and 2) =2;
              aDir2 := fDir in MFTentry2.FileRecHdr.Flags;
              newparent := MFTentry2.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;
              if newparent<>mftid then
              Win32Name := GetReferencedDir(newparent, dirname, aDir2);
              // dirname:=win32name;
              // dirname:=win32name + attrname;
              Win32Name := ExtractFileNameCutStream(dirname);
              // if aDir2 then dirname:=dirname+'\';
              ReadRunList2(mftno, parent, wantedVCN, aLevel + 1,
                Win32Name + attrname, RunList, True); // runlistflag
              MFTentry2.Free;
            end;

          end;
        end;
        M.Free;
    end;
  end;
end;

/// /////////////////////////////////////////////////////////////////////////////
/// /////////////////////////////////////////////////////////////////////////////

// reads runlist from attribute into dynamic array (TRunList)
procedure TNTFS.ReadRunList2(MFTid, parent: Cardinal; wantedVCN: Cardinal;
  aLevel: integer; dirname: widestring; var RunList: TRunList;
  runlistflag: boolean);
var
  Attr: TNTFSStandardAttributeHeader;
  AttrResident: TNTFSStandardAttributeHeader_Resident;
  NameAttr: TNTFSFileNameAttribute;
  ListRec: TNTFSAttributeListRecord;
  MFTentry1: TMFTEntry;
  addr, mftno, newMFTid: Cardinal;
  newparent: Cardinal;
  I, j, K, l, zwL, idx: integer;
  docontinue, aDir, aDir2: boolean;
  Win32Name, newname, newpath: string;
  attrname: string;
begin
  MFTentry1 := ReadMFTEntry(MFTid);
  if MFTentry1 = nil then
    exit;

  if MFTid = 9 then
    MFTid := 9;
  if MFTid = 83859 then
    MFTid := 83859;
  if MFTid = 17836 then
    MFTid := 17836;
  if MFTid = 15 then
    MFTid := 15;

  // aDir:=(MFTentry1.FileRecHdr.Flags and 2) = 2;
  aDir := fDir in MFTentry1.FileRecHdr.Flags;
  parent := INVALID_HANDLE_VALUE;
  newpath := '';

  // if alevel=0 then wantedVCN:=0;
  if (aLevel = 0) and (not runlistflag) and (NameAttr.RefToParentDir and $FFFFFFFFFF <> mftid) then
  begin
    NameAttr := MFTentry1.GetFileNameAttr();
    dirname := GetReferencedDir(NameAttr.RefToParentDir and $FFFFFFFFFF, '',
      aDir) + MFTentry1.GetFileName(parent);
  end;

  // ---------------------------------------------------------------------------
  // first loop scan for name and other attributes than attrlist and nonres data
  // first loop - scan for name and parent
  for I := 0 to MFTentry1.AttribCount - 1 do
  begin
    attrname := '';

    if MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_File_Name then
    begin
      if parent = INVALID_HANDLE_VALUE then
        parent := MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

      MFTentry1.Attrib[I].Data.Position := 0;
      MFTentry1.Attrib[I].Data.ReadBuffer(Attr, sizeof(Attr));
      MFTentry1.Attrib[I].Data.ReadBuffer(AttrResident, sizeof(AttrResident));
      MFTentry1.Attrib[I].Data.ReadBuffer(NameAttr, sizeof(NameAttr));

      if (aLevel = 0) or (dirname = '') then
        newname := '\' + MFTentry1.GetFileName(parent);

      if newname <> '' then
      begin
        if (newname[1] = '\') then
        begin
          if (newname = '\.') or (newname = '.') then
            newname := ''
          else if (parent = 5) then
            newname := copy(newname, 2, Length(newname) - 1);
          if (NameAttr.RefToParentDir and $FFFFFFFFFF <> mftid) then
          newname := GetReferencedDir(NameAttr.RefToParentDir and $FFFFFFFFFF,'', aDir) + newname;
        end
        else
          newname := PartInfo.sLetter + ':' + newname
      end;

      (*
        if length(runlist)>0 then
        begin
        if (runlist[0].desc<>'') and (runlist[0].desc<>dirname) and
        (Pos(dirname,runlist[0].desc)>0) and (length(dirname) > length(runlist[0].desc))
        then
        runlist[0].desc:=newname+runlist[0].desc
        else if (length(dirname) > length(runlist[0].desc)) then
        runlist[0].desc:=newname
        else if (not FileExists(ExtractFileNameCutStream(runlist[0].desc))) then
        runlist[0].desc:=newname+runlist[0].desc;
        end;
      *)

      l := Length(RunList);
      for K := 0 to l - 1 do
      begin
        if RunList[K].desc = '\\\\\\\\' then
        begin
          RunList[K].desc := newname;
        end
        else if (RunList[K].desc <> '') then
        begin
          if not FileExists(ExtractFileNameCutStream(RunList[K].desc)) then
            RunList[K].desc := newname + RunList[K].desc;
        end
        else if RunList[K].VCN = 0 then
        begin
          // if (k>0) and (runlist[k-1].Vcn<>0) then
          // if (runlist[k].desc<>'') then
          // if FileExists(ExtractFileNameCutStream(runlist[k].desc)) then begin
          if (RunList[K].desc = '\\\\\\\\') then
            RunList[K].desc := newname
        end;
      end;

      if (dirname <> newname) and (newname <> '') then
        // dirname := newname + dirname;
        dirname := newname;
    end;
  end; // first for

  // ----------------------------------------------------------------------------
  // second loop scan for NTFS_AttribID_Attribute_Data in main-entry
  // second loop - only scan for ATTRIBUTE_LIST
  // if (aLevel=0) and (not runlistflag) then
  for I := 0 to MFTentry1.AttribCount - 1 do
  begin
    if (MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Data) then
    begin
      if (MFTentry1.Attrib[I].Hdr.NonResident) then
      begin
        // AttributeProc(MFTentry1,I,MFTid,parent,wantedVCN,aLevel,dirname,RunList);

        (* *)
        attrname := NTFSAttribFileName(MFTentry1, I, MFTid, aLevel, aDir);
        if AddToRunList(MFTid, parent, MFTentry1, I, wantedVCN, RunList, dirname, attrname) then
          wantedVCN := RunList[Length(RunList) - 1].VCN +RunList[Length(RunList) - 1].Len + 1
        else
          continue;
        (* *)

        (*
          attrname:=NTFSAttribFileName(MFTentry1,I,MFTid,aLevel,aDir);
          if attrname<>'' then
          RunList[L].desc:=dirname+attrname+':$DATA'
          else
          RunList[L].desc:=dirname+attrname;
        *)
      end;
    end;
  end; // second for

  if MFTid = 15 then
    MFTid := 15;
  if MFTid = 0 then
    MFTid := 0;

  // ----------------------------------------------------------------------------
  // third loop scan AttributeList for dataruns
  // third loop - scan for NonResident/NTFS_AttribID_Data only in this record
  // if (aLevel=0) then
  // if (not runlistflag) then
  for I := 0 to MFTentry1.AttribCount - 1 do
  begin
    if (MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Attribute_List)
    then
    begin
      AttributeProc(MFTentry1, I, MFTid, parent, wantedVCN, aLevel, dirname, RunList);
      // attrname:=NTFSAttribFileName(MFTentry1,I,MFTid,aLevel,aDir);
      // if AddToRunList(MFTid,parent,MFTentry1,I,wantedVCN,RunList,dirname,attrname) then
      // wantedVCN:=RunList[length(RunList)-1].Vcn+RunList[length(RunList)-1].Len+1
    end;
  end; // third for

  // ----------------------------------------------------------------------------
  // fourth loop scan for NTFS_AttribID_Attribute_Data in main-entry
  // fourth loop - scan AttributeLists for data with a certain Vcn
  // if (aLevel<2) and (not runlistflag) then
  for I := 0 to MFTentry1.AttribCount - 1 do
  begin
    if (MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Attribute_List) or
      (MFTentry1.Attrib[I].Hdr.NonResident) then
    begin
      // if length(RunList)>0 then
      // if AddToRunList(MFTid,parent,MFTentry1,I,wantedVCN,RunList,newname,attrname) then
      // wantedVCN:=RunList[length(RunList)-1].Vcn+RunList[length(RunList)-1].Len+1;
      AttributeProc(MFTentry1, I, MFTid, parent, wantedVCN, aLevel, dirname, RunList)
    end;
    {
      if (MFTentry1.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Data) then
      begin
      if (MFTentry1.Attrib[I].Hdr.NonResident) then
      begin
      newname:='';
      newMFTid:=MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF;

      // todo read parentid and get path from it!!!!!
      // 33398 - M:\WebPlayer\Dollhouse\Dollhouse.S02e01.flv

      //MFTentry1.Attrib[I].
      //if dirname='' then
      //  newname:=GetReferencedDir(MFTid,dirname,aDir);
      //win32name:=MFTentry1.GetFileName(newMFTid);
      newname:=newname+win32name;

      attrname:=NTFSAttribFileName(MFTentry1,I,MFTid,aLevel,aDir);
      if attrname<>'' then
      newname:=dirname+attrname+':$DATA'
      else
      newname:=dirname+attrname;

      if AddToRunList(MFTid,parent,MFTentry1,I,wantedVCN,RunList,newname,attrname) then
      wantedVCN:=RunList[length(RunList)-1].Vcn+RunList[length(RunList)-1].Len+1
      else
      continue;

      //AddToRunList(MFTid,parent,MFTentry1,I,wantedVCN,RunList,newname,attrname);

      (*
      if Length(RunList)>0 then
      begin
      RunList[0].desc:=newname;
      wantedVCN:=RunList[length(RunList)-1].Vcn+RunList[length(RunList)-1].Len+1;
      AddToRunList(MFTid,parent,MFTentry1,I,wantedVCN,RunList,newname,attrname);
      end;
      *)
      end;
      end;
    }
  end; // fourth for

  (*
    if (MFTentry1.AttribCount=1) and (MFTentry1.Attrib[0].Hdr.AttributeType = NTFS_AttribID_Data) then
    dirname := GetReferencedDir(MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF, dirname);
    if (MFTentry1.AttribCount=1) and (MFTentry1.Attrib[0].Hdr.AttributeType = NTFS_AttribID_Index_Allocation) then
    dirname := GetReferencedDir(MFTentry1.FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF, dirname);
  *)
  MFTentry1.Free;
  if dirname = '\.' then
    dirname := '';
end;

procedure TNTFS.ReadRunListWrap(MFTid: Cardinal; var dirname: widestring;
  var RunList: TRunList);
begin
  ReadRunList2(MFTid, 0, 0, 0, dirname, RunList, False); // runlistflag
end;

initialization

// Loads function entry points in NTDLL
@NtFsControlFile := GetProcAddress(GetModuleHandle('ntdll.dll'), 'NtFsControlFile');

finalization

end.


