unit NTFSBrowse;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, StdCtrls, ComCtrls, ToolWin, Menus, ExtCtrls, Math,
  NTFS, Util;

type
  TNTFSBrowseForm = class(TForm)
    View: TListView;
    ImageList1: TImageList;
    ImageList2: TImageList;
    CoolBar1: TCoolBar;
    ToolBar1: TToolBar;
    ToolButton3: TToolButton;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    PopupMenu1: TPopupMenu;
    Info1: TMenuItem;
    ViewFirstClusterInDiskEditor1: TMenuItem;
    ViewMFTEntryInDiskEditor1: TMenuItem;
    Open1: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    SaveAs1: TMenuItem;
    SaveDlg: TSaveDialog;
    Status: TStatusBar;
    Panel1: TPanel;
    btnreload: TButton;
    Edit1: TEdit;
    btnstop: TButton;
    tmpLV: TListView;

    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);

    procedure ViewDblClick(Sender: TObject);
    procedure ViewMFTEntry1Click(Sender: TObject);
    procedure ViewMFTEntryInDiskEditor1Click(Sender: TObject);
    procedure ViewFirstClusterInDiskEditor1Click(Sender: TObject);
    procedure ViewKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure ViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure ViewCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
    function CompareNumbers(s1: string; s2: string): Integer;

    procedure Info1Click(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure OpeninNewWindow1Click(Sender: TObject);
    procedure StatusDblClick(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
    procedure btnreloadClick(Sender: TObject);
    procedure btnstopClick(Sender: TObject);


    procedure OneDirUp;
    procedure ShowDir(M: TMemoryStream);

    procedure ProcessDirectory(MFTid: cardinal; recurse: Boolean; logfile: TFileStream);
    procedure ProcessIndexRoot(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream;InfoStdHdr: TNTFSStandardInfoAttribute);
    procedure ProcessIndexAlloc(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream;InfoStdHdr: TNTFSStandardInfoAttribute);
    procedure ProcessAttributeList(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream;InfoStdHdr: TNTFSStandardInfoAttribute);

    procedure SaveAs1Click(Sender: TObject);
    procedure SaveFileAs(MFTid: Cardinal; Attr: TAttribute; fname: string; fsize: Int64; InfoStdHdr: TNTFSStandardInfoAttribute);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
  public
    NTFS: TNTFS;
    CurMFT: cardinal;
    Parent: cardinal;
    BackHistory: TList;
    NextHistory: TList;
    PartName: string;
    destdir: string;
    ColumnToSortFound: Integer;
    reverseorder: Boolean;
    processing: Boolean;
    stop: Boolean;
    closing: Boolean;
    procedure AddInfo(Name: string; size: int64; typ: string; MFTid: cardinal; FileAttrRec: TNTFSFileNameAttribute);
  end;

var
  NTFSBrowseForm: TNTFSBrowseForm;

implementation

uses Main, NTFSMFT, DiskEdit, progressbar;

{$R *.dfm}

procedure TNTFSBrowseForm.FormCreate(Sender: TObject);
begin
  BackHistory := TList.Create;
  NextHistory := TList.Create;
  ColumnToSortFound := 0;
  reverseorder := false;
  stop := false;
  closing := false;
  processing:=false;
end;

procedure TNTFSBrowseForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  stop := true;
  closing := true;
  sleep(10);
  Application.ProcessMessages;
end;

procedure TNTFSBrowseForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  stop := true;
  closing := true;
  sleep(10);
  Application.ProcessMessages;
  NTFS.Free;
end;

procedure TNTFSBrowseForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #27 then
    stop := true;
end;

procedure TNTFSBrowseForm.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Key = VK_F5 then
    btnreloadClick(Sender);
  if (Key = 88) or (Key = 120) then
    if ((ssAlt in Shift) OR (ssCtrl in Shift)) then // strg/alt+x
      Close;
end;

procedure TNTFSBrowseForm.FormResize(Sender: TObject);
begin
  Edit1.Width := Width - 200;
end;

procedure TNTFSBrowseForm.FormShow(Sender: TObject);
begin
  Application.ProcessMessages;
  ClearMessageBuffers;
end;

procedure TNTFSBrowseForm.ViewKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key = VK_RETURN) then
    ViewDblClick(Sender);
  if Key = VK_Back then
    ToolButton3Click(Sender);
end;

procedure TNTFSBrowseForm.StatusDblClick(Sender: TObject);
begin
  InputBox('Filename', '', Status.Panels[1].Text);
end;


procedure TNTFSBrowseForm.SaveFileAs(MFTid: Cardinal; Attr: TAttribute; fname: string; fsize: Int64; InfoStdHdr: TNTFSStandardInfoAttribute);
var RunList2: TRunList;
    fnamew: WideString;
    res: TMemoryStream;
begin
  // ProgressWin.ProgBar.Max:=totalclusters;
  ProgressWin.Label1.Caption := FName;
  ProgressWin.ProgBar.Position := 0;
  ProgressWin.lbProgress.Caption:='0';
  ProgressWin.lbMax.Caption:=IntToDottetInt(fsize)+' byte';
  ProgressWin.Refresh;
  RunList2 := nil;
  fnamew:=WideString(fname);
  if Attr.Hdr.NonResident then
  begin
    ProgressWin.Caption:='Get RunList ...';
    SetLength(Runlist2, 0);
    NTFS.ReadRunListWrap(MFTid, fnamew, Runlist2);
    ProgressWin.Caption:='Save';
    NTFS.SaveFileWithProgress(Runlist2, fname, fsize, InfoStdHdr);
  end else begin
    res := TMemoryStream.Create;
    res := NTFS.ReadFile(Attr,InfoStdHdr);
    res.SaveToFile(fname);
    res.Free;
  end;
end;

procedure TNTFSBrowseForm.SaveAs1Click(Sender: TObject);
var
  Runlist2: TRunList;
  MFT: TMFTEntry;
  MFTid: cardinal;
  I, J: Integer;
  //f: TMemoryStream;
  fname, fdir: WideString;
  fsize: int64;
  Item: TListItem;
begin
  // mftlist:=TStringList.Create;
  // namelist:=TStringList.Create;
  if (View.Items.Count>0) and (View.SelCount>0) then
  begin

    SaveDlg.Options := [ofEnableSizing];
    SaveDlg.FileName := View.Selected.Caption;
    SaveDlg.DefaultExt := '';
    if SaveDlg.Execute then
    begin
      processing:=true;
      ProgressWin.Show;
      Application.ProcessMessages;
      // have to temp save them
      // cause when going into dir the selection is gone
      tmpLV.Clear;
      for J := 0 to View.Items.Count - 1 do
      begin
        if View.Items[J].Selected then
        begin
          //tmpLV.AddItem(View.Items[J].Caption,nil);
          Item:=tmpLV.Items.Add;
          Item.Caption:=View.Items[J].Caption;
          for i:=0 to View.Items[J].SubItems.Count-1 do
          begin
            Item.SubItems.Add(View.Items[J].SubItems[i]);
            Item.Data:=View.Items[J].Data;
          end;
        end;
      end;
      tmpLV.Refresh;

      // go through tmp list
      for J := 0 to tmpLV.Items.Count-1 do
      begin
        MFTid := cardinal(tmpLV.Items[J].Data);
        MFT := NTFS.ReadMFTEntry(MFTid);
        if (MFT <> nil) then
        begin
          if (tmpLV.Items[J].ImageIndex = 0) then
          begin
            // mftlist.Add(IntToStr(Cardinal(tmpLV.Items[J].Data)));
            // namelist.Add(tmpLV.Items[J].Caption);
            destdir := ExtractFileDir(SaveDlg.FileName) + '\' + tmpLV.Items[J].Caption + '\';
            {$I-}
            mkdir(destdir);
            {$I+}
            ProcessDirectory(MFTid, true, nil);

          end else if (tmpLV.Items[J].ImageIndex = 1) then begin

            fname:=tmpLV.Items[J].Caption;
            fdir:=ExtractFileDir(SaveDlg.FileName);
            fname := fdir + '\' + fname;
            fname := ReplaceAllStringsInString(fname, WideString('\\'),WideString('\'));
            fsize := MFT.GetFileSize;

            // ProgressWin.ProgBar.Max:=totalclusters;
            ProgressWin.Label1.Caption := FName;
            ProgressWin.ProgBar.Position := 0;
            ProgressWin.lbProgress.Caption:='0';
            ProgressWin.lbMax.Caption:=IntToDottetInt(fsize)+' byte';
            ProgressWin.Caption:=inttostr(J)+' / '+inttostr(tmpLV.Items.Count-1) + 'Get RunList ...';
            ProgressWin.Refresh;

            SetLength(Runlist2, 0);
            NTFS.ReadRunListWrap(MFTid, fname, Runlist2);
            ProgressWin.Caption:=inttostr(J)+' / '+inttostr(tmpLV.Items.Count-1) + ' Save';
            NTFS.SaveFileWithProgress(Runlist2, fname, fsize, mft.InfoStdHdr);
          end;
          MFT.Free;
          if not ProgressWin.Visible then break;
        end;
      end; // for
      processing:=false;
      ProgressWin.Hide;
    end;
  end;
end;

procedure TNTFSBrowseForm.ViewDblClick(Sender: TObject);
var MFTid: cardinal;
begin
  if processing then exit;
  if View.ItemIndex > -1 then
    if View.Selected.ImageIndex = 0 then
    begin
      if (View.Selected.Caption = '..') and (View.Selected.SubItems[1] = '') then
      begin
        stop := true;
        OneDirUp;
      end else begin
        MFTid := cardinal(View.Selected.Data);
        BackHistory.Add(Pointer(CurMFT));
        ProcessDirectory(MFTid, false, nil);
      end;
    end
    else
    begin
      ViewMFTEntry1Click(Sender);
    end;
end;

procedure TNTFSBrowseForm.ToolButton1Click(Sender: TObject);
var
  MFTid: cardinal;
begin
  if BackHistory.Count > 0 then
  begin
    MFTid := cardinal(BackHistory[BackHistory.Count - 1]);
    BackHistory.Delete(BackHistory.Count - 1);
    NextHistory.Add(Pointer(CurMFT));
    ProcessDirectory(MFTid, false, nil);
  end
  else
    Beep;
end;

procedure TNTFSBrowseForm.ToolButton2Click(Sender: TObject);
var
  MFTid: cardinal;
begin
  if NextHistory.Count > 0 then
  begin
    MFTid := cardinal(NextHistory[NextHistory.Count - 1]);
    NextHistory.Delete(NextHistory.Count - 1);
    BackHistory.Add(Pointer(CurMFT));
    ProcessDirectory(MFTid, false, nil);
  end
  else
    Beep;
end;

procedure TNTFSBrowseForm.OneDirUp;
var
  I: Integer;
  lastone: string;
begin
  lastone := IntToStr(CurMFT);
  BackHistory.Add(Pointer(CurMFT));
  ProcessDirectory(Parent, false, nil);
  Application.ProcessMessages;
  for I := 0 to View.Items.Count - 1 do
  begin
    if View.Items[I].SubItems[6] = lastone then
    begin
      View.Selected := nil;
      Application.ProcessMessages;
      View.Items[0].Selected := false;
      View.Items[0].Focused := false;
      View.Items[I].Selected := true;
      View.Items[I].Focused := true;
      View.Items[I].MakeVisible(false);
      break;
    end;
  end;
end;

procedure TNTFSBrowseForm.btnreloadClick(Sender: TObject);
var
  I: Integer;
  lastone: string;
begin
  stop := true;
  if View.Selected <> nil then
    lastone := View.Selected.SubItems[6];
  ProcessDirectory(CurMFT, false, nil);
  for I := 0 to View.Items.Count - 1 do
  begin
    if View.Items[I].SubItems[6] = lastone then
    begin
      View.Selected := nil;
      Application.ProcessMessages;
      View.Items[0].Selected := false;
      View.Items[0].Focused := false;
      View.Items[I].Selected := true;
      View.Items[I].Focused := true;
      View.Items[I].MakeVisible(false);
      break;
    end;
  end;
end;

procedure TNTFSBrowseForm.btnstopClick(Sender: TObject);
begin
  stop := true;
end;

procedure TNTFSBrowseForm.ToolButton3Click(Sender: TObject);
begin
  stop := true;
  OneDirUp;
end;

procedure TNTFSBrowseForm.ViewMFTEntry1Click(Sender: TObject);
begin
  with TNTFSMFTForm.Create(Application) do
  begin
    NTFS := TNTFS.Create(Self.NTFS.PartInfo);
    PartName := Self.PartName;
    CurMFT := Integer(View.Selected.Data);
    UpdateNew;
  end;
end;

procedure TNTFSBrowseForm.Info1Click(Sender: TObject);
var
  MFT: TMFTEntry;
  I: Integer;
begin
  MFT := NTFS.ReadMFTEntry(Integer(View.Selected.Data));
  I := MFT.IndexOfAttrib(NTFS_AttribID_File_Name);
  MFT.Free;
  if I > 0 then
  begin
    with TNTFSMFTForm.Create(Application) do
    begin
      NTFS := TNTFS.Create(Self.NTFS.PartInfo);
      CurMFT := Integer(View.Selected.Data);
      UpdateNew;
      Tree.Selected := Tree.Items[0].Item[I].GetLastChild;
    end;
  end else begin
    MessageDlg('not found!',mtInformation,[mbOK],0)
  end;
end;

procedure TNTFSBrowseForm.ViewMFTEntryInDiskEditor1Click(Sender: TObject);
var MFTStart, MFTid: cardinal;
begin
  MFTid:=Cardinal(View.Selected.Data);
  with TDiskEditor.Create(Application) do
  begin
    partinfo:=NTFS.PartInfo;
    if (partinfo.partno=0) then
    begin
      // for mounted disks
      partition:=false;
      Disk:=partinfo.Disk;
      if Disk.DiskHandle=0 then
        Disk.OpenVolumePhysical;
      MinSec := 0;
      MaxSec := PartInfo.SizeSec;
      //NTFS.GetMFTRecordCluster(CurMFT, MFTStart, MFTneeded);
      //CurSec:=NTFS.GetMFTRecordCluster(MFTid);
      MFTStart:=NTFS.GetMFTRecordCluster(MFTid);
      CurSec :=Int64(MFTStart * ntfs.BootSec.SecPerClust) +
               MFTid*2; //Disk.DiskGeometry.BytesPerSector;
      BytesPerSector := Disk.DiskGeometry.BytesPerSector;
      BytesPerCluster := Disk.DiskGeometry.BytesPerSector*4;

    end else begin
      Partition := True;
      Disk := partinfo.Disk;
      if Disk.DiskHandle=0 then
        Disk.OpenVolumePhysical;

      MinSec := NTFS.PartInfo.StartSec; //  NTFS.BootSec.HiddenSectors;
      MaxSec := NTFS.PartInfo.StartSec+NTFS.PartInfo.SizeSec; //NTFS.BootSec.HiddenSectors + NTFS.BootSec.SectorSize;
      MFTStart:=NTFS.GetMFTRecordCluster(MFTid);
      CurSec := (NTFS.PartInfo.StartSec +
                Int64(MFTStart * ntfs.BootSec.SecPerClust) +
                MFTid*2); //Disk.DiskGeometry.BytesPerSector;
      //Position := (PartInfo.StartSec +
      //               (Int64(FragmentStartClust * BootSec.SecPerClust) +
      //               (MFTid * iDiv))) * VolumeInfo.BytesPerSector;
      //BytesPerSector := NTFS.BootSec.BytesPerSec;
      //BytesPerCluster := NTFS.BootSec.SecPerClust * NTFS.BootSec.BytesPerSec;
      BytesPerSector := NTFS.Disk.DiskGeometry.BytesPerSector; //BootSec.BytesPerSec;
      BytesPerCluster := NTFS.Disk.DiskGeometry.BytesPerSector*4;
    end;
    UpdateSec;
  end;
end;

procedure TNTFSBrowseForm.ViewFirstClusterInDiskEditor1Click(Sender: TObject);
var
  MFT: TMFTEntry;
  I: Integer;
  M: TMemoryStream;
  Hdr: TNTFSStandardAttributeHeader;
  HdrNonRes: TNTFSStandardAttributeHeader_NonResident;
  B: Byte;
  LengthBytes: Byte;
  OffsetBytes: Byte;
  RunLength: cardinal;
  RunOffset: cardinal;
begin
  MFT := NTFS.ReadMFTEntry(Integer(View.Selected.Data));
  if fDir in MFT.FileRecHdr.Flags then
    // if MFT.FileRecHdr.Flags=$0003 then
    I := MFT.IndexOfAttrib(NTFS_AttribID_Index_Allocation)
  else
    I := MFT.IndexOfAttrib(NTFS_AttribID_Data);
  if I > -1 then
  begin
    if MFT.Attrib[I].Hdr.NonResident then
    begin
      M := MFT.Attrib[I].Data;
      M.Position := 0;
      M.ReadBuffer(Hdr, SizeOf(Hdr));
      M.ReadBuffer(HdrNonRes, SizeOf(HdrNonRes));
      M.Position := HdrNonRes.DataRunsOffset;
      M.ReadBuffer(B, 1);
      LengthBytes := B and $0F;
      OffsetBytes := (B and $F0) shr 4;
      M.ReadBuffer(RunLength, LengthBytes);
      M.ReadBuffer(RunOffset, OffsetBytes);
      with TDiskEditor.Create(Application) do
      begin
        // Partition := True;
        Disk := NTFS.PartInfo.Disk;
        //MinSec := NTFS.BootSec.HiddenSectors;
        //MaxSec := NTFS.BootSec.HiddenSectors + NTFS.BootSec.SectorSize;
        MinSec := NTFS.PartInfo.StartSec; // NTFS.BootSec.HiddenSectors;
        MaxSec := NTFS.PartInfo.StartSec + NTFS.BootSec.SectorSize - 1;
        CurSec := NTFS.PartInfo.StartSec + RunOffset*NTFS.BootSec.SecPerClust;
        BytesPerSector := Disk.DiskGeometry.BytesPerSector;
        BytesPerCluster := Disk.DiskGeometry.BytesPerSector*4;
        UpdateSec;
      end;
    end else begin
      with TDiskEditor.Create(Application) do
      begin
        Partition := true;
        Disk := NTFS.PartInfo.Disk;
        //MinSec := NTFS.BootSec.HiddenSectors;
        //MaxSec := NTFS.BootSec.HiddenSectors + NTFS.BootSec.SectorSize;
        MinSec := NTFS.PartInfo.StartSec; //  NTFS.BootSec.HiddenSectors;
        MaxSec := NTFS.PartInfo.StartSec+NTFS.PartInfo.SizeSec; //NTFS.BootSec.HiddenSectors + NTFS.BootSec.SectorSize;
        CurSec := NTFS.PartInfo.StartSec + cardinal(View.Selected.Data) * 2 * NTFS.BootSec.SecPerClust;
        BytesPerSector := Disk.DiskGeometry.BytesPerSector;
        BytesPerCluster := Disk.DiskGeometry.BytesPerSector*4;
        UpdateSec;
      end;
    end;
  end else
    MessageDlg('MFT entry does not contain data or index allocation attribute. File or directory data is resident inside MFT record. To be able to view it use "View MFT Entry in Disk Editor..." instead.',
      mtInformation, [mbOK], 0);
  MFT.Free;
end;

procedure TNTFSBrowseForm.PopupMenu1Popup(Sender: TObject);
var
  I: Integer;
begin
  if View.ItemIndex = -1 then
  begin
    for I := 0 to PopupMenu1.Items.Count - 1 do
    begin
      PopupMenu1.Items[I].Visible := true;
      PopupMenu1.Items[I].Enabled := false;
      Info1.default := false;
      Open1.default := true;
    end
  end
  else
  begin
    for I := 0 to PopupMenu1.Items.Count - 1 do
      PopupMenu1.Items[I].Enabled := true;
    Open1.Visible := View.Selected.ImageIndex = 0;
    Info1.Visible := true;
    Open1.default := View.Selected.ImageIndex = 0;
    Info1.default := View.Selected.ImageIndex <> 0;
    // ViewMFTEntry1.Visible                    := True;
    ViewFirstClusterInDiskEditor1.Visible := true;
    ViewMFTEntryInDiskEditor1.Visible := true;
    // SaveAs1.Visible          := View.Selected.ImageIndex = 1;
  end;
end;

procedure TNTFSBrowseForm.Open1Click(Sender: TObject);
var
  MFT: cardinal;
begin
  MFT := Integer(View.Selected.Data);
  BackHistory.Add(Pointer(CurMFT));
  ProcessDirectory(MFT, false, nil);
end;

procedure TNTFSBrowseForm.OpeninNewWindow1Click(Sender: TObject);
var
  localmftno: cardinal;
  zw: TNTFSBrowseForm;
begin
  localmftno := cardinal(View.Selected.Data);
  zw := TNTFSBrowseForm.Create(nil);
  zw.NTFS := TNTFS.Create(Self.NTFS.PartInfo);
  zw.NTFS.PartName := Self.NTFS.PartName;
  zw.PartName := Self.NTFS.PartName;
  zw.Caption := 'NTFS Browser - ' + PartName;
  zw.Show;
  Application.ProcessMessages;
  zw.ProcessDirectory(localmftno, false, nil);
end;


procedure TNTFSBrowseForm.AddInfo(Name: string; size: int64; typ: string; MFTid: cardinal; FileAttrRec: TNTFSFileNameAttribute);
var
  added: Boolean;
  idx: Integer;
  Item: TListItem;
begin
  added := false;
  if (name <> '..') then
    for idx := 0 to View.Items.Count - 1 do
    begin
      if (int64(View.Items[idx].Data) = MFTid) then
      begin
        added := true;
        if (Length(View.Items[idx].Caption) < Length(name)) then
          View.Items[idx].Caption := name;
        break;
      end;
    end;

  if not added then
  begin
    Item := View.Items.Add;
    Item.Caption := name;
    Item.Data := Pointer(MFTid);
    // Item.Data := Pointer(IdxRec.MFTReference and $FFFFFFFFFF );
    if (typ = 'reparse (link)') then
    begin
      Item.ImageIndex := 2;
      Item.SubItems.Add(typ);
      Item.SubItems.Add('');
    end
    else if (typ = 'dir') then
    begin
      Item.ImageIndex := 0;
      Item.SubItems.Add(typ);
      Item.SubItems.Add('');
    end
    else
    begin
      Item.ImageIndex := 1;
      Item.SubItems.Add(typ);
      Item.SubItems.Add(IntToStr(size));
    end;
    //if (InfoAttr<>) then
    begin
    Item.SubItems.Add(NTFSTimeToStr(FileAttrRec.TimeLastModified));
    Item.SubItems.Add(NTFSTimeToStr(FileAttrRec.TimeLastAccess));
    Item.SubItems.Add(NTFSTimeToStr(FileAttrRec.TimeMTFAltr));
    Item.SubItems.Add(NTFSTimeToStr(FileAttrRec.TimeCreated));
    end;
    Item.SubItems.Add(IntToStr(MFTid));
    tag := tag + 1;
  end;
end;

/// ////////////////////////////////////////////////////////////////////////
// ***********************************************************************//
/// ////////////////////////////////////////////////////////////////////////

procedure TNTFSBrowseForm.ProcessAttributeList(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream;InfoStdHdr: TNTFSStandardInfoAttribute);
var
  M: TMemoryStream;
  ListRec: TNTFSAttributeListRecord;
  Addr: cardinal;
  MFT: TMFTEntry;
  I: Integer;
  RunList: TRunList;
  fnamew: WideString;
begin
  Status.Panels[0].Text := 'Searching, please wait...';
  if Attrib.Hdr.NonResident then
  begin
    fnamew:=WideString(IntToStr(i));
    (*
    // ProgressWin.ProgBar.Max:=totalclusters;
    ProgressWin.Label1.Caption := FName;
    ProgressWin.ProgBar.Position := 0;
    ProgressWin.lbProgress.Caption:='0';
    ProgressWin.lbMax.Caption:=IntToDottetInt(fsize)+' byte';
    ProgressWin.Refresh;
    RunList2 := nil;
    *)


  end else begin
    M := Attrib.Data;
    M.Position := Attrib.HdrResident.AttributeOffset;

    Addr := M.Position;
    FillChar(ListRec, SizeOf(ListRec), 0);
    while M.size - M.Position > SizeOf(ListRec) do
    begin
      if stop then
        break;
      M.Position := Addr + ListRec.size;
      Addr := M.Position;
      // M.ReadBuffer(ListRec,SizeOf(ListRec));
      M.Read(ListRec, SizeOf(ListRec));
      if (ListRec.AttributeType = NTFS_AttribID_Index_Allocation) and
        (CurMFT <> ListRec.FileRef and $FFFFFFFFFF) then
      begin
        MFT := NTFS.ReadMFTEntry(ListRec.FileRef and $FFFFFFFFFF);
        for I := 0 to MFT.AttribCount - 1 do
        begin
          if stop then
            break;
          if MFT.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Index_Allocation
          then
            ProcessIndexAlloc(MFT.Attrib[I], pos, recurse, logfile,InfoStdHdr);
        end;
        MFT.Free;
      end;
      if (ListRec.AttributeType = NTFS_AttribID_Index_Root) and
        (CurMFT <> ListRec.FileRef and $FFFFFFFFFF) then
      begin
        MFT := NTFS.ReadMFTEntry(ListRec.FileRef and $FFFFFFFFFF);
        for I := 0 to MFT.AttribCount - 1 do
        begin
          if stop then
            break;
          if MFT.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Index_Allocation
          then
          begin
            ProcessIndexRoot(MFT.Attrib[I], pos, recurse, logfile,MFT.InfoStdHdr);
          end;
        end;
        MFT.Free;
      end;
    end;
    Status.Panels[0].Text := '';
  end;
  Application.ProcessMessages;
end;

// ***********************************************************************//
procedure TNTFSBrowseForm.ProcessIndexRoot(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream; InfoStdHdr: TNTFSStandardInfoAttribute);
var
  M: TMemoryStream;
  RootIdx: TNTFSIndexRoot;
  IdxRec: TNTFSIndexRecord;
  Item: TListItem;
  MFTEnt: TMFTEntry;
  NextPos: Word;
  aNameStr: String;
  NameStr: WideString;
  // localparent: Cardinal;
  aDir: Boolean;
begin
  Status.Panels[0].Text := 'Searching, please wait...';
  M := Attrib.Data;

  if Attrib.Hdr.NonResident then
  begin
      // indizes in nonresident runlist fuck
      ShowDir(NTFS.ReadRunListData(Attrib, '',InfoStdHdr));
  end else begin
      M.Position := Attrib.HdrResident.AttributeOffset;
      M.ReadBuffer(RootIdx, SizeOf(RootIdx));

      if M.size - M.Position < SizeOf(IdxRec) then
        exit;

      View.Items.BeginUpdate;

      NextPos := M.Position;
      M.ReadBuffer(IdxRec, SizeOf(IdxRec));
      while (IdxRec.MFTReference <> 0) and
            (IdxRec.MFTReference and $FFFFFFFFFF < NTFS.MaxMFT) do
      begin
        if stop then break;
        if View.Items.Count mod 200 = 0 then
        begin
          View.Items.EndUpdate;
          View.Refresh;
          Application.ProcessMessages;
          View.Items.BeginUpdate;
        end;

        if (IdxRec.MFTReference and $FFFFFFFFFF) <> 0 then
        begin
          MFTEnt := NTFS.ReadMFTEntry(IdxRec.MFTReference and $FFFFFFFFFF);
          SetLength(NameStr, IdxRec.FileNameLength);
          M.ReadBuffer(NameStr[1], IdxRec.FileNameLength * 2);
          aNameStr := WideStringToString(NameStr, IdxRec.FileNameLength);
          if (IdxRec.Flags and NTFS_AttribID_Reparse_Point) > 0 then
            AddInfo(aNameStr, 0, 'reparse (link)', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr)
          else if (IdxRec.Flags and $10000000) > 0 then
            AddInfo(aNameStr, 0, 'dir', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr)
          else
            AddInfo(aNameStr, IdxRec.RealSize, 'file', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr);
          MFTEnt.Free;
        end;

        M.Position := NextPos + IdxRec.size;
        NextPos := M.Position;
        if M.size - M.Position < SizeOf(IdxRec) then
          exit;
        M.ReadBuffer(IdxRec, SizeOf(IdxRec));
      end;

      View.Items.EndUpdate;
  end;
  Status.Panels[0].Text := IntToStr(View.Items.Count) + ' files';
  Status.Panels[1].Text := NTFS.GetReferencedDir(CurMFT, '', aDir);
  Application.ProcessMessages;
end;

// ***********************************************************************//
procedure TNTFSBrowseForm.ProcessIndexAlloc(Attrib: TAttribute; pos: Integer; recurse: Boolean; logfile: TFileStream; InfoStdHdr: TNTFSStandardInfoAttribute);
var
  M: TMemoryStream;
  MFTEntry: TMFTEntry;
  IdxHdr: TNTFSIndexHeader;
  IdxRec: TNTFSIndexRecord;
  Item: TListItem;
  Addr, MFTid: cardinal;
  I, J, K: Integer;
  localparent: cardinal;
  fsize:int64;
  aDir: Boolean;
  NameStr: String;
begin
  Status.Panels[0].Text := 'Searching, please wait...';
  Status.Refresh;
  View.SortType:=stNone;
  View.Items.BeginUpdate;
  M := NTFS.ReadRunListData(Attrib, '',InfoStdHdr);
  for I := 0 to M.size div (512 * 8) - 1 do
  begin
    if stop then
    begin
      Status.Panels[0].Text := 'stop';
      Status.Refresh;
      break;
    end;
    M.Position := I * (512 * 8); // 512
    if (M.Position >= Attrib.HdrNonResident.RealSize) then
      break;

    M.ReadBuffer(IdxHdr, SizeOf(IdxHdr));
    if IdxHdr.IndxID <> 'INDX' then
    // raise EInOutError.Create('Index Header signature missing while processing directory index allocation.');
    begin
      Continue;
      // ShowMessage('Error: Invalid index header signature!');
      // Abort;
    end;

    M.Position := int64(I * 512 * 8) + int64(IdxHdr.BytesToIndexEntries + $18);
    if (M.Position >= Attrib.HdrNonResident.RealSize) then
      break;
    Addr := M.Position;
    M.ReadBuffer(IdxRec, SizeOf(IdxRec));

    while (IdxRec.MFTReference <> 0) and
      ((M.Position - I * (512 * 8)) < (IdxHdr.SizeOfIndexEntries + $18)) do
    begin
      if stop then
        break;
      if (IdxRec.FileNamespace <> npDos) then
      begin
        if ((View.Items.Count mod 200) = 0) then
        begin
          View.Items.EndUpdate;
          Application.ProcessMessages;
          if stop then break;
          Status.Panels[1].Text := IntToStr(View.Items.Count) + ' Items...';
          View.Refresh;
          Status.Refresh;
          View.Items.BeginUpdate;
        end;

        MFTid := IdxRec.MFTReference and $FFFFFFFFFF;
        if (MFTid >= 0) and (MFTid < NTFS.MaxMFT) then
        begin
          MFTEntry := NTFS.ReadMFTEntry(MFTid);
          if MFTEntry <> nil then
          begin
            if MFTentry.FileRecHdr.FileID = 'FILE' then
            begin
              fsize:=MFTEntry.GetFileSize;
              NameStr := MFTEntry.GetFileName(localparent);
              if NameStr = '' then
                NameStr := ExtractFilename(NTFS.GetReferencedDir(MFTid,'',aDir));

              if (IdxRec.Flags = -6917529014487735290) then
                AddInfo(NameStr, 0, 'reparse (link)', IdxRec.MFTReference and $FFFFFFFFFF, MFTentry.GetFileNameAttr)
              else if ((IdxRec.Flags and $10000000) > 0) then
              begin
                // dir ????????????????????????????
                if logfile = nil then
                begin
                  AddInfo(NameStr, 0, 'dir', IdxRec.MFTReference and $FFFFFFFFFF,MFTentry.GetFileNameAttr)
                end else begin
                  NameStr := destdir + '\' + MFTEntry.GetFileName(localparent);
                  logfile.Write(PChar(NameStr + Chr(13) + Chr(10))^, Length(NameStr) + 2);
                  Status.Panels[1].Text := NameStr;
                end;

                // only recurse when save dir, so make directory and process it
                if recurse then
                begin
                  if (MFTid <> 0) and (MFTid <> 5) then
                  begin
                    destdir := destdir + MFTEntry.GetFileName(localparent) + '\';
                    // destdir:=destdir+Item.Caption+'\';
                    {$I-}
                     mkdir(destdir);
                    {$I+}
                    ProcessDirectory(IdxRec.MFTReference and $FFFFFFFFFF,recurse, logfile);
                    K := Length(destdir) - 1;
                    while destdir[K] <> '\' do Dec(K);
                    destdir := Copy(destdir, 1, K);
                  end;
                end;

              end else begin
                // file ????????????????????????????
                if logfile = nil then
                begin
                  // namestr:=MFTEntry.GetFileName(localparent);
                  if (fDir in MFTEntry.FileRecHdr.Flags) // or
                  // (WORD(MFTEntry.FileRecHdr.Flags)=9)
                  then
                    AddInfo(NameStr, fsize, 'dir', IdxRec.MFTReference and $FFFFFFFFFF,MFTentry.GetFileNameAttr)
                  else
                    AddInfo(NameStr, fsize, 'file', IdxRec.MFTReference and $FFFFFFFFFF,MFTEntry.GetFileNameAttr);

                  // only recurse when save dir, so save
                  if recurse then
                  begin
                    if (MFTid <> 0) and (MFTid <> 5) then
                    begin
                      J := MFTEntry.IndexOfAttrib(NTFS_AttribID_Data);
                      if J > -1 then
                      begin
                        //Status.Panels[1].Text := NameStr;
                        SaveFileAs(MFTid,MFTEntry.Attrib[J], destdir + NameStr,fsize,InfoStdHdr);
                        Application.ProcessMessages;
                      end;
                    end;
                  end;
                end else begin
                  NameStr := destdir + '\' + MFTEntry.GetFileName(localparent);
                  logfile.Write(PChar(NameStr + Chr(13) + Chr(10))^, Length(NameStr) + 2);
                end;

              end;

            end; // valid 'FILE'

            MFTEntry.Free;
          end;

        end;

      end;

      M.Position := Addr + IdxRec.size;
      Addr := M.Position;
      M.ReadBuffer(IdxRec, SizeOf(IdxRec));
    end; // while
  end; // for
  View.Items.EndUpdate;
  M.Free;
  Status.Panels[0].Text := IntToStr(View.Items.Count) + ' Items...';
  Application.ProcessMessages;
end; // proc

procedure TNTFSBrowseForm.ShowDir(M: TMemoryStream);
var
  IdxHdr: TNTFSIndexHeader;
  IdxRec: TNTFSIndexRecord;
  MFTEnt:TMFTEntry;
  Addr: cardinal;
  I: Integer;
  NameStr: WideString;
  FileName: AnsiString;
  aDir: Boolean;
begin
  tag := 0;
  View.AllocBy := M.size div 2;
  View.Items.BeginUpdate;
  View.SortType:=stNone;

  for I := 0 to M.size div (512 * 8) - 1 do
  begin
    if stop then
      break;
    M.Position := I * (512 * 8);
    M.ReadBuffer(IdxHdr, SizeOf(IdxHdr));
    if IdxHdr.IndxID <> 'INDX' then
    begin
      Continue;
      // ShowMessage('Error: Invalid index header signature!');
    end;

    M.Position := int64(I * 512 * 8) + int64(IdxHdr.BytesToIndexEntries + $18);
    Addr := M.Position;
    M.ReadBuffer(IdxRec, SizeOf(IdxRec));

    while (M.Position - I * (512 * 8)) < IdxHdr.SizeOfIndexEntries + $18 do
    begin
      if stop then
        break;
      if ((tag mod 200) = 0) then
      begin
        View.Items.EndUpdate;
        Application.ProcessMessages;
        View.Refresh;
        View.Items.BeginUpdate;
      end;

      if (IdxRec.MFTReference and $FFFFFFFFFF) < NTFS.MaxMFT then
      begin
        MFTEnt := NTFS.ReadMFTEntry(IdxRec.MFTReference and $FFFFFFFFFF);
        SetLength(NameStr, IdxRec.FileNameLength);
        M.ReadBuffer(NameStr[1], IdxRec.FileNameLength * 2);
        FileName := WideStringToString(NameStr, IdxRec.FileNameLength);

        if (IdxRec.Flags and NTFS_AttribID_Reparse_Point) > 0 then
          AddInfo(NameStr, 0, 'reparse (link)', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr)
        else if (IdxRec.Flags and $10000000) > 0 then
          AddInfo(FileName, 0, 'dir', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr)
        else
          AddInfo(FileName, IdxRec.RealSize, 'file', IdxRec.MFTReference and $FFFFFFFFFF,MFTEnt.GetFileNameAttr);
        MFTEnt.Free;
      end;

      M.Position := Addr + IdxRec.size;
      Addr := M.Position;
      M.ReadBuffer(IdxRec, SizeOf(IdxRec));
    end; // while
  end; // for

  View.Items.EndUpdate;
  View.AllocBy := 0;
  Status.Panels[0].Text := IntToStr(View.Items.Count) + ' Items...';
  Status.Panels[1].Text := NTFS.GetReferencedDir(CurMFT, '', aDir);
  // Label2.Caption:=' '+IntToStr(realfiles) + ' real files ('+ IntToStr(List.Items.Count) +')';
end;

// ***********************************************************************//
procedure TNTFSBrowseForm.ProcessDirectory(MFTid: cardinal; recurse: Boolean; logfile: TFileStream);
var
  MFTEntry: TMFTEntry;
  RunList: TRunList;
  I, J, K: Integer;
  lastsort: Integer;
  ipos: Integer;
  Name: WideString;
  aDir: Boolean;
  fsize: Int64;

begin
  stop := false;
  Status.Panels[0].Text := 'Searching, please wait...';

  CurMFT := MFTid;
  MFTEntry := NTFS.ReadMFTEntry(MFTid);
  if MFTEntry = nil then
  begin
    MessageDlg('No data found on Volume ' + PartName, mtError, [mbOK], 0);
    Close;
    exit;
  end;
  if not processing then
  begin
    View.Items.BeginUpdate;
    View.ShowColumnHeaders := false;
    View.Clear;
    View.SortType := stnone;
    ToolButton1.Enabled := false;
    ToolButton2.Enabled := false;
    ToolButton3.Enabled := false;
  end;
  name := NTFS.GetReferencedDir(CurMFT, '', aDir);
  if Length(name) > 3 then
    name := name + '\';
  Edit1.Text:=name;
  fsize:=MFTEntry.GetFileSize;
  ipos := MFTEntry.FileRecHdr.OffsetToFirstAttr;
  for I := 0 to MFTEntry.AttribCount - 1 do
  begin

    case MFTEntry.Attrib[I].Hdr.AttributeType of
      NTFS_AttribID_Index_Root:
        if not stop then
          ProcessIndexRoot(MFTEntry.Attrib[I], ipos, recurse, logfile,MFTEntry.InfoStdHdr);
      NTFS_AttribID_Index_Allocation:
        if not stop then
          ProcessIndexAlloc(MFTEntry.Attrib[I], ipos, recurse, logfile,MFTEntry.InfoStdHdr);
      NTFS_AttribID_Attribute_List:
        if not stop then
          if MFTEntry.Attrib[I].Hdr.NonResident then
          begin
            name := '';
            NTFS.ReadRunListWrap(MFTid, name, RunList);
            for J := 0 to Length(RunList) - 1 do
            begin
              if pos('::$INDEX_ALLOCATION', RunList[J].desc) > 0 then
              begin
                K := J + 1;
                while (RunList[K].desc = '') and (K < Length(RunList)) do
                begin
                  ShowDir(NTFS.ReadClusters(RunList[K].Lcn, RunList[K].Len));
                  Inc(K);
                end;
              end;
            end;
          end else begin
            if recurse and (not stop) then
            begin
              ProgressWin.Caption:='Get RunList ...';
              SetLength(Runlist, 0);
              NTFS.ReadRunListWrap(MFTid, name, Runlist);
              ProgressWin.Caption:='Save';
              NTFS.SaveFileWithProgress(Runlist, name, fsize, MFTEntry.InfoStdHdr);
            end else if not stop then
              ProcessAttributeList(MFTEntry.Attrib[I], ipos, recurse, logfile, MFTEntry.InfoStdHdr);
          end;
      NTFS_AttribID_File_Name:
        Parent := MFTEntry.GetFileNameAttr.RefToParentDir and $FFFFFFFFFF;
    end; // case
    Inc(ipos, MFTEntry.Attrib[I].Hdr.size);
  end;

  if closing then
  begin
    MFTEntry.Free;
    exit;
  end;
  if MFTid <> 5 then
      AddInfo('..', 0, 'dir', MFTid,MFTEntry.GetFileNameAttr);
  MFTEntry.Free;

  if (not processing) then
  begin
    processing:=true;
    Status.Panels[0].Text := IntToStr(View.Items.Count - 1) + ' Items.';
    Status.Panels[1].Text := NTFS.GetReferencedDir(CurMFT, '', aDir);
    if Length(Status.Panels[1].Text) > 3 then
      Status.Panels[1].Text := Status.Panels[1].Text + '\';

    View.SortType := stText;
    View.Items.EndUpdate;
    View.ShowColumnHeaders := true;
    View.SetFocus;

    ToolButton1.Enabled := true;
    ToolButton2.Enabled := true;
    ToolButton3.Enabled := true;

    processing:=false;
    lastsort:=ColumnToSortFound;
    if lastsort<>0 then begin
      ColumnToSortFound:=0;
      View.AlphaSort;
    end else begin
      View.AlphaSort;
      ColumnToSortFound:=1;
      View.AlphaSort;
      ColumnToSortFound:=lastsort;
    end;
    View.Refresh;
    if View.Items.Count > 0 then
    begin
      View.Items[0].Selected := true;
      View.Items[0].Focused := true;
    end;
  end;

end;

function TNTFSBrowseForm.CompareNumbers(s1: string; s2: string): Integer;
var N1, N2: int64;
begin
	result:=0;
	n1:=StrToInt64Def(s1,-1); n2:=StrToInt64Def(s2,-1);
	if n1<n2 then result:=1 else if n1>n2 then result:=-1;
end;

procedure TNTFSBrowseForm.ViewCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
var ix: Integer;
begin
  if not View.ShowColumnHeaders then exit;
  if processing then exit;
  if View.SortType = stnone then exit;
  if Status.Panels[0].Text = 'Searching, please wait...' then exit;
  if Item1.Caption = '..' then begin
    Compare := -1;
    exit;
  end;
  if Item2.Caption = '..' then begin
    Compare := 1;
    exit;
  end;
  ix:=ColumnToSortFound-1;
  if (ix < 0) then
    if not reverseorder then
      Compare :=  CompareStr(Item1.Caption, Item2.Caption)
    else
      Compare := -CompareStr(Item1.Caption, Item2.Caption)
  else if (ix = 1) or (ix = 6) then
    if not reverseorder then
      Compare :=  CompareNumbers(Item1.SubItems[ix], Item2.SubItems[ix])
    else
      Compare := -CompareNumbers(Item1.SubItems[ix], Item2.SubItems[ix])
  else if not reverseorder then
      Compare :=  CompareStr(Item1.SubItems[ix], Item2.SubItems[ix])
  else
      Compare := -CompareStr(Item1.SubItems[ix], Item2.SubItems[ix])
end;

procedure TNTFSBrowseForm.ViewColumnClick(Sender: TObject; Column: TListColumn);
begin
  if Status.Panels[0].Text = 'Searching, please wait...' then
    exit;
  if (ColumnToSortFound = Column.Index) then
    reverseorder := not reverseorder;
  ColumnToSortFound := Column.Index;
  View.SortType := stDAta;
  (Sender as TCustomListView).AlphaSort;
end;

end.
