unit NTFSMFT;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, StrUtils,
  Dialogs, ComCtrls, StdCtrls, ImgList, ToolWin, Menus, ExtCtrls, Grids,
  ShellAPI,
  NTFS, Util;

type
  TNTFSMFTForm = class(TForm)
    Tree: TTreeView;
    List: TListView;
    CoolBar1: TCoolBar;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ImageList1: TImageList;
    ImageList2: TImageList;
    StatusBar1: TStatusBar;
    SaveDlg: TSaveDialog;
    Button2: TButton;
    Edit1: TEdit;
    Label1: TLabel;
    Edit2: TEdit;
    Label2: TLabel;
    Splitter1: TSplitter;
    ToolButton8: TToolButton;
    ToolButton10: TToolButton;
    Button1: TButton;
    ToolButton17: TToolButton;
    Button3: TButton;
    Panel1: TPanel;
    CoolBar2: TCoolBar;
    ToolBar2: TToolBar;
    RunlistView: TListView;
    Panel2: TPanel;
    Splitter2: TSplitter;
    Label3: TLabel;
    withrunlist: TCheckBox;
    attrsearch: TComboBox;
    Label4: TLabel;
    Label5: TLabel;
    Button5: TButton;
    doshort: TCheckBox;
    docount: TCheckBox;
    ToolButton6: TToolButton;
    Label6: TLabel;

    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);

    procedure StatusBar1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

    procedure ToolButton10Click(Sender: TObject);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure ToolButton4Click(Sender: TObject);
    procedure ToolButton5Click(Sender: TObject);
    procedure ToolButton17Click(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);

    procedure Edit1KeyPress(Sender: TObject; var Key: Char);
    procedure Edit2KeyPress(Sender: TObject; var Key: Char);

    procedure StatusBar1DblClick(Sender: TObject);
    procedure ListDblClick(Sender: TObject);
    procedure WithrunListClick(Sender: TObject);

    procedure AddInfo(Name,Value: string; ps:integer; lastValue:string='');
    procedure AddRunInfo(const info: string; Lcn,Len,Vcn:int64; ps:integer);
    procedure TreeChange(Sender: TObject; Node: TTreeNode);

    procedure ShowFileRecHeader;
    procedure ShowStandardHdr(Index: Word);
    procedure ShowStandardHdrExt(Index: Word);
    procedure ShowDir(M: TMemoryStream);
    procedure ShowAttribute(Index: Cardinal);

    procedure ShowRunList(Index: Word);
    procedure ShowRunList2(sname:string);

    procedure FindAttribute(Sender: TObject; attrtype: longint);
    procedure doshortClick(Sender: TObject);
    procedure ToolButton6Click(Sender: TObject);

  public
    NTFS: TNTFS;
    MFTEntry: TMFTEntry;
    CurMFT: Cardinal;
    Buffer: TMemoryStream;
    partname: string;
    bDBL: Boolean;
    stop: boolean;
    closing:boolean;
    bUpdating:Boolean;
    procedure RefreshAttributes;
    procedure UpdateNew;
  end;

var NTFSMFTForm: TNTFSMFTForm;

implementation

uses Main, Disks, DiskEdit;

{$R *.dfm}

procedure TNTFSMFTForm.FormCreate(Sender: TObject);
begin
  Tag:=0;
  Label1.Caption:='';
  Label2.Caption:='';
  stop:=false;
  closing:=false;
  bUpdating:=FALSE;
  Show;
end;

procedure TNTFSMFTForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  stop:=true;
  closing:=true;
  sleep(10);
  Application.ProcessMessages;
end;

procedure TNTFSMFTForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  stop:=true;
  closing:=true;
  sleep(10);
  Application.ProcessMessages;
  //ntfs.Free;
  Buffer.Free;
end;

procedure TNTFSMFTForm.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key=27 then stop:=true;
  if key=VK_F5 then Button3Click(sender);
    if (Key=88) or (key=120) then
      if ((ssAlt in shift) OR (ssCtrl in shift)) then // strg/alt+x
        Close;
end;

procedure TNTFSMFTForm.FormShow(Sender: TObject);
begin
  if tag<>1111 then begin
  tag:=1111;
  UpdateNew;
  tag:=0;
  end;
end;

procedure TNTFSMFTForm.ShowFileRecHeader;
var FileRecHdr: TNTFSFileRecordHeader;
begin
  FileRecHdr := MFTEntry.FileRecHdr;
  AddInfo('FileID',FileRecHdr.FileID,-1);
  AddInfo('Update sequence offset',IntTostr(FileRecHdr.UsaOffset) + ' (' + IntToHex(FileRecHdr.UsaOffset,4)+')',-1);
  AddInfo('Update sequence size',IntToStr(FileRecHdr.UsaSize),-1);
  AddInfo('Log file sequence number',IntToStr(FileRecHdr.LogFileSeqNum),-1);
  AddInfo('Sequence number',IntToStr(FileRecHdr.SeqNum),-1);
  AddInfo('Reference count (hard links)',IntToStr(FileRecHdr.ReferenceCount),-1);
  AddInfo('First attribute''s offset',IntTostr(FileRecHdr.OffsetToFirstAttr) + ' (' + IntToHex(FileRecHdr.OffsetToFirstAttr,4)+')',-1);
  AddInfo('Flags', NTFSMFTRecFlagsToStr(FileRecHdr.Flags),-1);
  AddInfo('Real size of file record',IntToStr(FileRecHdr.RealFileRecSize),-1);
  AddInfo('Allocated size of file record',IntToStr(FileRecHdr.AllocatedFileRecSize),-1);
  AddInfo('Reference to base file record #',
           IntToStr(FileRecHdr.RefToBaseFileRec and $FFFFFFFFFF),-1);
  AddInfo('First attribute''s ID',IntTostr(FileRecHdr.NextAttrID) + ' ('+IntToHex(FileRecHdr.NextAttrID,2)+')',-1);
  AddInfo('MFT record number',IntToStr(FileRecHdr.NumberOfRec),-1);
  AddInfo('Update sequence number',IntToStr(FileRecHdr.UsaNum),-1);
end;

procedure TNTFSMFTForm.ShowStandardHdr(Index: Word);
const YesNo: array [False..True] of string = ('No','Yes');
var AttrHdr: TNTFSStandardAttributeHeader;
begin
  if MFTEntry.AttribCount=0 then exit;
  AttrHdr := MFTEntry.Attrib[Index].Hdr;
  AddInfo('Attribute type',IntTostr(AttrHdr.AttributeType) + ' ('+IntToHex(AttrHdr.AttributeType,4) + ')',-1);
  AddInfo('Size',IntToStr(AttrHdr.Size),-1);
  AddInfo('Resident',YesNo[not AttrHdr.NonResident],-1);
  AddInfo('Name length',IntToStr(AttrHdr.NameLength)  + ' characters',-1);
  AddInfo('Name offset',IntTostr(AttrHdr.NameOffset) + ' ('+IntToHex(AttrHdr.NameOffset,4) + ')',-1);
  AddInfo('Flags',NTFSAttrFlagsToStr(AttrHdr.Flags),-1);
  AddInfo('Attribute ID',IntToStr(AttrHdr.AttributeID),-1);
end;

procedure TNTFSMFTForm.ShowStandardHdrExt(Index: Word);
var HdrResident: TNTFSStandardAttributeHeader_Resident;
    HdrNonResident: TNTFSStandardAttributeHeader_NonResident;
    Name: WideString;
begin
  if MFTEntry.Attrib[Index].Hdr.NonResident then
  begin
   HdrNonResident := MFTEntry.Attrib[Index].HdrNonResident;
   AddInfo('First virtual cluster number',IntToStr(HdrNonResident.StartVCN),-1);
   AddInfo('Last virtual cluster number',IntToStr(HdrNonResident.LastVCN),-1);
   AddInfo('Offset of Data Runs',IntTostr(HdrNonResident.DataRunsOffset) + ' ('+IntToHex(HdrNonResident.DataRunsOffset,4) + ')',-1);
   AddInfo('Compression unit size',IntToStr(HdrNonResident.ComprtessionUnit),-1);
   AddInfo('Allocated size',IntToStr(HdrNonResident.AllocatedSize),-1);
   AddInfo('Real size',IntToStr(HdrNonResident.RealSize),-1);
   AddInfo('Initialized size of data stream',IntToStr(HdrNonResident.StreamInitializedSize),-1);
  end else begin
   HdrResident := MFTEntry.Attrib[Index].HdrResident;
   AddInfo('Attribute size',IntToStr(HdrResident.AttributeLength),-1);
   AddInfo('Attribute offset',IntTostr(HdrResident.AttributeOffset) + ' ('+IntToHex(HdrResident.AttributeOffset,4) + ')',-1);
   AddInfo('Indexed flag',IntTostr(HdrResident.IndexedFlag) + ' ('+IntToHex(HdrResident.IndexedFlag,4) + ')',-1);
  end;
  if MFTEntry.Attrib[Index].Hdr.NameLength > 0 then
    with MFTEntry.Attrib[Index].Data do
    begin
      Position := MFTEntry.Attrib[Index].Hdr.NameOffset;
      SetLength(Name,MFTEntry.Attrib[Index].Hdr.NameLength);
      ReadBuffer(Name[1],MFTEntry.Attrib[Index].Hdr.NameLength * 2);
    end else
      Name := '(no name)';
  AddInfo('Name',Name,-1);
end;

procedure TNTFSMFTForm.RefreshAttributes;
begin
  List.Clear;
  if Tree.Selected.Level = 0 then
  begin
   ShowFileRecHeader;
   end
  else
  if Tree.Selected.Level = 1 then
   begin
   ShowStandardHdr(Tree.Selected.Index);
   end
  else
  if Tree.Selected.Level = 2 then
  begin
   if Tree.Selected.Index = 0 then
     ShowStandardHdrExt(Tree.Selected.Parent.Index)
   else
   if Tree.Selected.ImageIndex = 4 then
     ShowRunList(Tree.Selected.Parent.Index)
   else
     ShowAttribute(Tree.Selected.Parent.Index);
  end;
end;

procedure TNTFSMFTForm.StatusBar1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
     StatusBar1.Hint:=StatusBar1.Panels[2].Text + StatusBar1.Panels[1].Text;
     ShowHint:=True;
end;

procedure TNTFSMFTForm.StatusBar1DblClick(Sender: TObject);
begin
  InputBox('Filename', '', StatusBar1.Panels[2].Text+StatusBar1.Panels[1].Text)
end;

procedure TNTFSMFTForm.Button5Click(Sender: TObject);
begin
  ShellExecute(Self.Handle,'open','explorer.exe',
               PWidechar('/e,/select,"'+ StatusBar1.Panels[2].Text+StatusBar1.Panels[1].Text+'"'),
               nil, SW_SHOWNORMAL)
end;

procedure TNTFSMFTForm.Edit1KeyPress(Sender: TObject; var Key: Char);
begin
     if (key=#13) then Button2Click(Sender);
end;

procedure TNTFSMFTForm.Edit2KeyPress(Sender: TObject; var Key: Char);
begin
     if (key=#13) then ToolButton4Click(Sender);
end;

procedure TNTFSMFTForm.TreeChange(Sender: TObject; Node: TTreeNode);
begin
  Tree.Enabled:=false;
  RefreshAttributes;
  Tree.Enabled:=true;
  Tree.SetFocus;
end;

procedure TNTFSMFTForm.Button1Click(Sender: TObject);
begin
  Stop:=True;
end;

procedure TNTFSMFTForm.ToolButton5Click(Sender: TObject);
begin // goto first
  CurMFT := 0;
  OnShow(nil);
end;

procedure TNTFSMFTForm.ToolButton6Click(Sender: TObject);
var //FragmentStartClust:int64;
    MFTStart, MFTneeded: Int64;
begin
  with TDiskEditor.Create(Application) do
  begin
    partinfo:=NTFS.PartInfo;
    if (partinfo.partno=0) then
    begin
      // for mounted disks
      partition:=false;
      Disk:=partinfo.Disk;
      if Disk.DiskHandle=0 then
        Disk.OpenVolumePhysical;
      MinSec := 0;
      MaxSec := PartInfo.SizeSec;
      //NTFS.GetMFTRecordCluster(CurMFT, MFTStart, MFTneeded);
      CurSec:=NTFS.GetMFTRecordCluster(CurMFT);
      BytesPerSector := Disk.DiskGeometry.BytesPerSector;
      BytesPerCluster := Disk.DiskGeometry.BytesPerSector*4;
      Caption := 'Disk Viewer - '+NTFS.PartInfo.sLetter+':\['+NTFS.PartInfo.sFSName+'] "'+NTFS.PartInfo.sVolName+'"';
    end else begin
      Partition := True;
      Disk := partinfo.Disk;
      if Disk.DiskHandle=0 then
        Disk.OpenVolumePhysical;
      MinSec := PartInfo.StartSec;
      MaxSec := PartInfo.StartSec + PartInfo.SizeSec - 1;
      //FragmentStartClust:=ntfs.GetMFTRecordCluster(CurMFT);
      //NTFS.GetMFTRecordCluster(CurMFT, MFTStart, MFTneeded);
      MFTStart:=NTFS.GetMFTRecordCluster(CurMFT);
      CurSec := PartInfo.StartSec +
                Int64(MFTStart * ntfs.BootSec.SecPerClust) +
                CurMFT*2;
      BytesPerSector := Disk.DiskGeometry.BytesPerSector;
      BytesPerCluster := Disk.DiskGeometry.BytesPerSector*4;
      Caption := 'Disk Viewer - '+NTFS.PartInfo.sLetter+':\['+NTFS.PartInfo.sFSName+'] "'+NTFS.PartInfo.sVolName+'"';
    end;
    UpdateSec;
  end;
end;

procedure TNTFSMFTForm.ToolButton1Click(Sender: TObject);
begin // goto previous
  if CurMFT > 0 then
  begin
    Dec(CurMFT);
    OnShow(nil);
  end;
end;

procedure TNTFSMFTForm.ToolButton2Click(Sender: TObject);
begin // goto next
  if CurMFT<NTFS.MaxMFT-1 then
    Inc(CurMFT);
  OnShow(nil);
end;

procedure TNTFSMFTForm.ToolButton4Click(Sender: TObject);
begin
  CurMFT := StrToIntDef(Edit2.Text,CurMFT);
  if CurMFT>=NTFS.MaxMFT then CurMFT:=NTFS.MaxMFT-1;
  if CurMFT<0            then CurMFT:=0;
  OnShow(nil);
end;

procedure TNTFSMFTForm.ToolButton3Click(Sender: TObject);
begin // goto last
  CurMFT := NTFS.MaxMFT-1;
  OnShow(nil);
end;

procedure TNTFSMFTForm.Button3Click(Sender: TObject);
begin
  OnShow(nil);
end;

procedure TNTFSMFTForm.ToolButton17Click(Sender: TObject);
begin
  Application.ProcessMessages;
  if ToolButton17.down then ToolButton17.ImageIndex:=3 else ToolButton17.ImageIndex:=4;
end;

{
NTFS_AttribID_Standard_Information            = $10;
NTFS_AttribID_Attribute_List                  = $20;
NTFS_AttribID_File_Name                       = $30;
NTFS_AttribID_Object_ID                       = $40;
NTFS_AttribID_Security_Descriptor             = $50;
NTFS_AttribID_Volume_Name                     = $60;
NTFS_AttribID_Volume_Information              = $70;
NTFS_AttribID_Data                            = $80;
NTFS_AttribID_Index_Root                      = $90;
NTFS_AttribID_Index_Allocation                = $A0;
NTFS_AttribID_Bitmap                          = $B0;
NTFS_AttribID_Reparse_Point                   = $C0;
NTFS_AttribID_Extended_Attribute_Information  = $D0;
NTFS_AttribID_Extended_Attribute              = $E0;
NTFS_AttribID_Property_Set                    = $F0;
NTFS_AttribID_Logged_Utility_Stream           = $100;
}

procedure TNTFSMFTForm.withrunlistClick(Sender: TObject);
begin
  if (withrunlist.State=cbChecked) then
    withrunlist.Caption:='nonresident'
  else if (withrunlist.State=cbUnChecked) then
    withrunlist.Caption:='resident'
  else
    withrunlist.Caption:='resident or not'
end;

procedure TNTFSMFTForm.ShowAttribute(Index: Cardinal);
var Hdr: TNTFSStandardAttributeHeader;
    HdrResident: TNTFSStandardAttributeHeader_Resident;
    FileNameAttr: TNTFSFileNameAttribute;
    ListRec: TNTFSAttributeListRecord;
    VolInfo: TNTFSVolumeInformation;
    ObjectIDInfo:TNTFSAttributeOBJECTID;
    InfoAttr: TNTFSStandardInfoAttribute;
    RootIdx: TNTFSIndexRoot;
    IdxRec: TNTFSIndexRecord;
    ReparseData: TReparsePoint;
    NameStr: WideString;
    NextPos, parent: Cardinal;
    M: TMemoryStream;
    f: TMemorystream;
    RunList2:TRunList;
    filename:WideString;
    mftno,fsize,idcount:Int64;
begin
  Label2.Caption:='';
  Label1.Caption:='';
  idcount:=0;
  M := MFTEntry.Attrib[Index].Data;
  M.Position := 0;
  M.ReadBuffer(Hdr,SizeOf(Hdr));

  if Hdr.NonResident then
  begin
    // Non-Resident Attributes with RUN-Lists

    // NTFS_AttribID_Attribute_List
    if Hdr.AttributeType = NTFS_AttribID_Attribute_List then
    begin
      M:=NTFS.ReadRunListData(MFTEntry.Attrib[Index], '',MFTEntry.InfoStdHdr);
      NextPos := M.Position;
      FillChar(ListRec,SizeOf(ListRec),0);
      while MFTEntry.Attrib[Index].HdrNonResident.RealSize - M.Position > SizeOf(ListRec) do
      begin
        M.Position := NextPos + ListRec.Size;
        NextPos := M.Position;
        M.Read(ListRec,SizeOf(ListRec));
        if (ListRec.AttributeType=0) then break;
        if not doshort.Checked then begin
        AddInfo('Attribute type',IntTostr(ListRec.AttributeType)+ ' ('+IntToHex(ListRec.AttributeType,2) + ')',-1);
        AddInfo('Record Size',IntToStr(ListRec.Size),-1);
        AddInfo('Name Length / Offset',IntToStr(ListRec.NameLength) + ' / '+IntToHex(ListRec.NameOffset,2) + ' (hex)',-1);
        end;
        AddInfo('StartVcn',IntToStr(ListRec.StartVCN),-1);
        AddInfo('Reference to MFT #', IntToStr(ListRec.FileRef and $FFFFFFFFFF),-1);
        if not doshort.Checked then begin
        AddInfo('Attribute',IntToStr(ListRec.AttributeID),-1);
        AddInfo('','',-1);
        end;
        Application.ProcessMessages;
      end;

    // NTFS_AttribID_Bitmap OR NTFS_AttribID_Data
    end else if (Hdr.AttributeType = NTFS_AttribID_Data) or
                (Hdr.AttributeType = NTFS_AttribID_Bitmap) then
    begin
      SaveDlg.FileName := MFTEntry.GetFileName(parent);
      if SaveDlg.Execute then
      begin

          fsize:=MFTEntry.GetFileSize;
          if fsize=0 then
            fsize:=(MFTEntry.Attrib[Index].HdrNonResident.LastVCN-MFTEntry.Attrib[Index].HdrNonResident.StartVCN)*
                   (ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust);
          //MFTEntry.Attrib[Index].HdrNonResident.RealSize
          SetLength(Runlist2,0);
          NTFS.ReadRunListWrap(CurMFT,filename,RunList2);
          NTFS.SaveFileWithProgress(RunList2,SaveDlg.FileName,fsize,MFTEntry.InfoStdHdr);
                 //MFTEntry.Attrib[Index].HdrNonResident.RealSize)

          (*
          if (fsize>10000000) then
          begin
              SetLength(Runlist2,0);
              NTFS.ReadRunListWrap(MFTNum,filename,RunList2);
              //RunList2:=NTFS.ReadRunList(MFTEntry.Attrib[Index]);
              NTFS.SaveFileWithProgress(RunList2,SaveDlg.FileName,fsize);
                 //MFTEntry.Attrib[Index].HdrNonResident.RealSize)
          end else begin
              //F := TMemoryStream.Create;
              F := NTFS.ReadFile(MFTEntry.Attrib[Index]);
              F.SaveToFile(SaveDlg.FileName);
              F.Free;
          end;
          *)
      end;

    // NTFS_AttribID_Index_Allocation
    end else if Hdr.AttributeType = NTFS_AttribID_Index_Allocation then
    begin
      ShowDir(NTFS.ReadRunListData(MFTEntry.Attrib[Index], '',InfoAttr));

    end else
      AddInfo('(unknown non-resident attribute)','',-1);

  end else begin

    // Resident Attributes without RUN-Lists
    M.ReadBuffer(HdrResident,SizeOf(HdrResident));
    M.Position := HdrResident.AttributeOffset;

    case Hdr.AttributeType of
        NTFS_AttribID_Standard_Information:
        begin
          M.ReadBuffer(InfoAttr,SizeOf(InfoAttr));
          AddInfo('Time created',NTFSTimeToStr(InfoAttr.TimeCreated),-1);
          AddInfo('Time modified',NTFSTimeToStr(InfoAttr.TimeModified),-1);
          AddInfo('MFT record last modified',NTFSTimeToStr(InfoAttr.TimeModifiedFileRec),-1);
          AddInfo('MFT record last accessed',NTFSTimeToStr(InfoAttr.TimeAccessedFileRec),-1);
          AddInfo('File Attributes',NTFSPermissionsToStr(InfoAttr.FileAttributes),-1);
        end;

        NTFS_AttribID_Reparse_Point:
        begin
          M.ReadBuffer(ReparseData,sizeof(ReparseData));
          M.Position := M.Position + 4;
          SetLength(NameStr,ReparseData.ReparseDataLength);
          M.ReadBuffer(NameStr[1],ReparseData.ReparseDataLength div 2);
          AddInfo('link to', NameStr, -1);
        end;

        NTFS_AttribID_File_Name:
        begin
          M.ReadBuffer(FileNameAttr,SizeOf(FileNameAttr));
          AddInfo('Reference to parent MFT #', IntToStr(FileNameAttr.RefToParentDir and $FFFFFFFFFF),-1);
          AddInfo('Time created',NTFSTimeToStr(FileNameAttr.TimeCreated),-1);
          AddInfo('Time modified',NTFSTimeToStr(FileNameAttr.TimeLastModified),-1);
          AddInfo('MFT record last modified',NTFSTimeToStr(FileNameAttr.TimeMTFAltr),-1);
          AddInfo('Time last accessed',NTFSTimeToStr(FileNameAttr.TimeLastAccess),-1);
          AddInfo('Allocated size',IntToStr(FileNameAttr.AllocatedSize),-1);
          AddInfo('Real size',IntToStr(FileNameAttr.RealSize),-1);
          AddInfo('Flags',NTFSPermissionsToStr(FileNameAttr.Flags),-1);
          AddInfo('Extended attribute reparse',IntToStr(FileNameAttr.ForEAsReparse),-1);
          AddInfo('Filename length',IntToStr(FileNameAttr.FileNameLength) + ' chars',-1);
          AddInfo('Namespace',NameSpaces[FileNameAttr.FileNamespace],-1);
          SetLength(NameStr,FileNameAttr.FileNameLength);
          M.ReadBuffer(NameStr[1],FileNameAttr.FileNameLength * 2);
          AddInfo('Filename',NameStr,-1);
        end;

        NTFS_AttribID_Index_Root:
        begin
          M.ReadBuffer(RootIdx,SizeOf(RootIdx));
          AddInfo('Attribute type',IntToHex(RootIdx.AttributeType,2) + ' (hex)',-1);
          AddInfo('Collation rule',IntToHex(RootIdx.CollationRule,8) + ' (hex)',-1);
          AddInfo('Index allocation entry size',IntToStr(RootIdx.SizeOfIndexAllocEntry),-1);
          AddInfo('Clusters per index record',IntToStr(RootIdx.ClustPerIndexRec),-1);
          AddInfo('Offset of index entries',IntToStr(RootIdx.OffsetOfFirstEntry),-1);
          AddInfo('Size of index entries',IntToStr(RootIdx.SizeOfIndexEntries),-1);
          AddInfo('Size allocated for index entries',IntToStr(RootIdx.AllocatedSizeForEntries),-1);
          if RootIdx.Flags <> 0 then
            AddInfo('Flags','index allocation used',-1)
          else
            AddInfo('Flags','entries fit in index root',-1);
          if M.Size - M.Position < SizeOf(IdxRec) then
            Exit;
          AddInfo('','',-1);

          NextPos := M.Position;
          M.ReadBuffer(IdxRec,SizeOf(IdxRec));
          mftno:=IdxRec.MFTReference and $FFFFFFFFFF;
          while (IdxRec.MFTReference <> 0) and (mftno < NTFS.MaxMFT) do
          begin
            AddInfo('Reference to MFT #',IntToStr(mftno),-1);
            if not doshort.Checked then begin
            AddInfo('Size',IntToStr(IdxRec.Size),-1);
            AddInfo('Offset to filename',IntToHex(IdxRec.FilenameOffset,4),-1);
            AddInfo('Index flags',Inttohex(IdxRec.IndexFlags,2) + ' (hex)',-1);
            end;
            AddInfo('Reference to parent directory MFT #', IntToStr(mftno),-1);
            if not doshort.Checked then begin
            AddInfo('Time created',NTFSTimeToStr(IdxRec.TimeCreated),-1);
            AddInfo('Time modified',NTFSTimeToStr(IdxRec.TimeLastModified),-1);
            AddInfo('MFT record last modified',NTFSTimeToStr(IdxRec.TimeMTFAltr),-1);
            AddInfo('Time last accessed',NTFSTimeToStr(IdxRec.TimeLastAccess),-1);
            AddInfo('Allocated size',IntToStr(IdxRec.AllocatedSize),-1);
            AddInfo('Real size',IntToStr(IdxRec.RealSize),-1);
            AddInfo('Flags',IntToHex(IdxRec.Flags,2) + ' (hex)',-1);
            AddInfo('Filename length',IntToStr(IdxRec.FileNameLength) + ' characters',-1);
            AddInfo('Namespace',NameSpaces[IdxRec.FileNamespace],-1);
            end;
            SetLength(NameStr,IdxRec.FileNameLength);
            M.ReadBuffer(NameStr[1],IdxRec.FileNameLength * 2);
            AddInfo('Filename',NameStr,-1);
            AddInfo('','',-1);
            M.Position := NextPos + IdxRec.Size;
            NextPos := M.Position;
            if M.Size - M.Position < SizeOf(IdxRec) then
              Exit;
	    	    M.ReadBuffer(IdxRec,SizeOf(IdxRec));
          end;
        end;

        NTFS_AttribID_Attribute_List:
        begin
          NextPos := M.Position;
          FillChar(ListRec,SizeOf(ListRec),0);
          while M.Size - M.Position > SizeOf(ListRec) do
          begin
            M.Position := NextPos + ListRec.Size;
            NextPos := M.Position;
            M.Read(ListRec,SizeOf(ListRec));
            if not doshort.Checked then begin
            AddInfo('Attribute type',IntToHex(ListRec.AttributeType,2) + ' (hex)',-1);
            AddInfo('Record Size',IntToStr(ListRec.Size),-1);
            AddInfo('Name Length / Offset',IntToStr(ListRec.NameLength)+' / '+IntToHex(ListRec.NameOffset,2) + ' (hex)',-1);
            end;
            AddInfo('Virtual Cluster Number',IntToStr(ListRec.StartVCN),-1);
            AddInfo('Reference to MFT #', IntToStr(ListRec.FileRef and $FFFFFFFFFF),-1);
            if not doshort.Checked then begin
            AddInfo('Attribute',IntToStr(ListRec.AttributeID),-1);
            AddInfo('','',-1);
            end;
            inc(idcount);
          end;
          label1.Caption:=IntToStr(idcount)+' references found';
        end;

        NTFS_AttribID_Volume_Name:
        begin
          SetLength(NameStr,HdrResident.AttributeLength);
          M.ReadBuffer(NameStr[1],HdrResident.AttributeLength);
          AddInfo('Volume Name',NameStr,-1);
        end;

        NTFS_AttribID_Object_ID:
        begin
          M.Position:=HdrResident.AttributeOffset;
          M.ReadBuffer(ObjectIDInfo,SizeOf(TNTFSVolumeInformation));
          AddInfo('ObjectId'     ,IntToStr(ObjectIDInfo.ObjectId.Data1),-1);
          AddInfo('BirthVolumeId',IntToStr(ObjectIDInfo.BirthVolumeId.Data1),-1);
          AddInfo('BirthObjectId',IntToStr(ObjectIDInfo.BirthObjectId.Data1),-1);
          AddInfo('DomainId'     ,IntToStr(ObjectIDInfo.DomainId.Data1),-1);
        end;

        NTFS_AttribID_Volume_Information:
        begin
          M.Position:=HdrResident.AttributeOffset;
          M.ReadBuffer(VolInfo,SizeOf(TNTFSVolumeInformation));
          AddInfo('NTFS-Version',IntToStr(VolInfo.Major_Version)+'.'+IntToStr(VolInfo.Minor_Version),-1);
          AddInfo('Flags','0x'+IntToHex(VolInfo.VolumeFlags, 8)+' ('+IntToStr(VolInfo.VolumeFlags)+')',-1);
        end;

        NTFS_AttribID_Data:
        begin
          SaveDlg.FileName := MFTEntry.GetFileName(parent);
          if SaveDlg.Execute then
          begin
            //NTFS.SaveFileWithProgress(NTFS.ReadRunList(MFTEntry.Attrib[Index]), SaveDlg.FileName, Hdr.Size);
            F := TMemoryStream.Create;
            F.CopyFrom(M,HdrResident.AttributeLength);
            F.SaveToFile(SaveDlg.FileName);
            F.Free;
          end;
        end;

    else
        AddInfo('(unknown resident attribute)','',-1);

    end; // case Hdr.AttributeType of

  end; // Resident Attributes

end;

procedure TNTFSMFTForm.ListDblClick(Sender: TObject);
var MFTid            : Cardinal;
    NTFSMFTFormlocal : TNTFSMFTForm;
begin
  if (Pos(' #', List.Selected.Caption)>0)
  then begin
    MFTid := StrToIntdef(Trim(List.Selected.SubItems.Text),-1);
    if (MFTid>=0) and (MFTid <= NTFS.MaxMFT) then begin
      NTFSMFTFormlocal := TNTFSMFTForm.Create(Application);

      NTFSMFTFormlocal.NTFS := TNTFS.Create(NTFS.PartInfo);
      // todo das oder das
      //NTFSMFTFormlocal.NTFS := NTFS;

      NTFSMFTFormlocal.CurMFT:=MFTid;
      NTFSMFTFormlocal.partname:=Partname;
      NTFSMFTFormlocal.UpdateNew;
      NTFSMFTFormlocal.Left:=NTFSMFTFormlocal.Left+50;
      NTFSMFTFormlocal.top:=NTFSMFTFormlocal.top+50;
    end else
      List.Selected.SubItems[0] := List.Selected.SubItems[0] + 'too big';
  end;
end;

procedure TNTFSMFTForm.Button2Click(Sender: TObject);
var i, spos1, spos2, start, c : integer;
begin
  c:=0;
  start:=0;
  with list do begin
    i:=start;
    while i<Items.Count do begin
    //for i:=start to Items.Count-1 do begin
      spos1 := Pos(LowerCase(Edit1.Text), LowerCase(Items[i].Caption));
      spos2 := Pos(LowerCase(Edit1.Text), LowerCase(Items[i].SubItems[0]));
      if (spos1<>0) or (spos2<>0) then begin
        Items[i].Selected:=True;
        inc(c);
      end;
      (*
      if (Items[i].Caption='Reference to parent directory MFT #') or
      ( (spos1=0) and (spos2=0) )
      then begin
        Items[i].Delete;
        Dec(i);
      end;
      *)
      Inc(i);
    end;
  end;
  Label1.Caption:=inttostr(c)+'x found ';
end;

procedure TNTFSMFTForm.AddInfo(Name,Value: string;ps:integer;lastValue:string='');
var I: TListItem;
    added:booleaN;
    idx:integer;
begin
  if ps>=0 then
    I := List.Items.Insert(ps)
  else
    I := List.Items.Add;
  I.Caption := Name;
  I.SubItems.Add(Value);

  I.ImageIndex:=-1;
  added:=lastValue=Value;
  if (not added) and (Pos(' #', Name)>0) and (value<>inttostr(CurMFT)) then
  begin
    for idx:=0 to List.Items.Count-2 do
      if (List.Items[idx].Caption='Reference to MFT #') and
         (List.Items[idx].SubItems[0]=value) then
      begin
        added:=true;
        break;
      end;
    if not added then
      I.ImageIndex:=0;
  end;
  if (Name='Virtual Cluster Number') then
    I.ImageIndex:=4;
end;

procedure TNTFSMFTForm.AddRunInfo(const info: string; Lcn,Len,Vcn:int64; ps:integer);
var I: TListItem;
begin
  if ps>=0 then
    I := runlistview.Items.Insert(ps)
  else begin
    I:=TListItem.Create(runlistview.Items);
    runlistview.Items.AddItem(I);
  end;
  I.Caption := Info;
  if Lcn<>-1 then
  I.SubItems.Add(IntToDottetInt(Lcn)) else I.SubItems.Add('');
  if Len<>-1 then
  I.SubItems.Add(IntToDottetInt(Len)) else I.SubItems.Add('');
  if Vcn<>-1 then
  I.SubItems.Add(IntToDottetInt(Vcn)) else I.SubItems.Add('');
end;

procedure TNTFSMFTForm.ShowDir(M: TMemoryStream);
var
  IdxHdr: TNTFSIndexHeader;
  IdxRec: TNTFSIndexRecord;
  NameStr: WideString;
  Addr: Cardinal;
  I: Integer;
  realfiles, realfilesInSub: Cardinal;
  cI, cIInSub: Cardinal;
  filename: AnsiString;
begin
  realfiles:=0; // 278528 376832
  cI:=0;
  cIInSub:=0;
  realfilesInSub:=0;

  List.AllocBy:=M.Size div 2;
  List.DisableAlign;
  List.Items.BeginUpdate;
  RunlistView.Items.BeginUpdate;

  for I := 0 to M.Size div (512 * 8) - 1 do
  begin
    if Stop then break;

    M.Position := I * (512 * 8);
    M.ReadBuffer(IdxHdr,SizeOf(IdxHdr));
    if IdxHdr.IndxID <> 'INDX' then
    begin
       Continue;
       ShowMessage('Error: Invalid index header signature!');
       //Abort;
    end;
    AddInfo('Index header signature',Trim(IdxHdr.IndxID),-1);
    //AddInfo('Update sequence offset',IntToHex(IdxHdr.UpdateSeqOffset,4) + ' (hex)');
    //AddInfo('Size of update sequence',IntToStr(IdxHdr.SizeOfUpdateSeq));
    //AddInfo('Log file sequence number',IntToStr(IdxHdr.LogFileSeqNum));
    AddInfo('Virtual cluster number',IntToStr(IdxHdr.VCN),-1);
    AddInfo('Offset of index entries',IntToHex(IdxHdr.BytesToIndexEntries,4) + ' (hex)',-1);
    AddInfo('Size of index entries',IntToStr(IdxHdr.SizeOfIndexEntries),-1);
    AddInfo('Size allocated for index entries',IntToStr(IdxHdr.AllocatedSizeForEntries),-1);
    //if IdxHdr.NotLeafNode then
    //AddInfo('NotLeafNode','True') else AddInfo('NotLeafNode','False');
    //AddInfo('Update sequence',IntToStr(IdxHdr.UpdateSeq));
    AddInfo('','',-1);

    M.Position := int64(I * 512 * 8) + int64(IdxHdr.BytesToIndexEntries + $18);
    Addr := M.Position;
    M.ReadBuffer(IdxRec,SizeOf(IdxRec));
    realfilesInSub:=0;
    cIInSub:=0;

    while (M.Position - I * (512 * 8)) < IdxHdr.SizeOfIndexEntries + $18 do
    begin
      if Stop then break;
      if ((cI mod 200) = 0) then begin
         List.Items.EndUpdate;
         List.Refresh;
         Label2.Caption:=IntToStr(realfiles) + ' real files';
         Application.ProcessMessages;
         List.Items.BeginUpdate;
      end;

      SetLength(NameStr,IdxRec.FileNameLength);
      M.ReadBuffer(NameStr[1],IdxRec.FileNameLength * 2);
      filename:=WideStringToString (NameStr, IdxRec.FileNameLength);

      AddInfo('Reference to MFT #', IntToStr(IdxRec.MFTReference and $FFFFFFFFFF),-1);
      if not doshort.Checked then begin
      AddInfo('Size',IntToStr(IdxRec.Size),-1);
      //AddInfo('Offset to filename',IntToHex(IdxRec.FilenameOffset,4),-1);
      //AddInfo('Index flags',Inttohex(IdxRec.IndexFlags,2) + ' (hex)',-1);
      //AddInfo('Reference to parent directory MFT #', IntToStr(IdxRec.ParentMFTReference and $FFFFFFFFFF),-1);
      //AddInfo('Time created',NTFSTimeToStr(IdxRec.TimeCreated),-1);
      //AddInfo('Time modified',NTFSTimeToStr(IdxRec.TimeAltered),-1);
      //AddInfo('MFT record last modified',NTFSTimeToStr(IdxRec.TimeMTFAltr),-1);
      //AddInfo('Time last accessed',NTFSTimeToStr(IdxRec.TimeLastRead),-1);
      AddInfo('Allocated size',IntToStr(IdxRec.AllocatedSize),-1);
      AddInfo('Real size',IntToStr(IdxRec.RealSize),-1);
      //AddInfo('Flags',IntToHex(IdxRec.Flags,2) + ' (hex)',-1);
      end;
      //AddInfo('Filename length',IntToStr(IdxRec.FileNameLength) + ' chars',-1);
      if (IdxRec.FileNamespace<=npDosAndWin32) then
        AddInfo('Filename ('+NameSpaces[IdxRec.FileNamespace]+')',filename,-1)
      else
        AddInfo('Filename (-)',filename,-1);

      if (IdxRec.FileNamespace<>npDos) and
         ((IdxRec.ParentMFTReference and $FFFFFFFFFF)>0) and
         ((IdxRec.ParentMFTReference and $FFFFFFFFFF)<NTFS.MaxMFT) then
      begin
          inc(realfiles);
          inc(realfilesInSub);
      end;
      inc(cI);
      inc(cIInSub);
      M.Position := Addr + IdxRec.Size;
      Addr := M.Position;
      M.ReadBuffer(IdxRec,sizeof(idxrec));
    end; // while
    AddInfo('','',-1);
    if not doshort.Checked then begin
    AddInfo('sub#-all#-subreal#-real#',IntToStr(cIInSub)+'-'+ IntToStr(cI)+'-'+ IntToStr(realfilesInSub)+'-'+IntToStr(realfiles),-1);
    AddInfo('','',-1);
    end;
    Application.ProcessMessages;
  end; // for
  List.Items.EndUpdate;
  List.EnableAlign;
  List.AllocBy:=0;
  List.Refresh;
  RunlistView.Items.EndUpdate;
  RunlistView.Refresh;
  AddInfo('sub#-all#-subreal#-real#',IntToStr(cIInSub)+'-'+ IntToStr(cI)+'-'+ IntToStr(realfilesInSub)+'-'+IntToStr(realfiles),0);
  Label2.Caption:=' '+IntToStr(realfiles) + ' real files ('+ IntToStr(List.Items.Count) +')';
end;

procedure TNTFSMFTForm.ToolButton10Click(Sender: TObject);
begin
if      attrsearch.ItemIndex=17 then
  FindAttribute(Sender, -1) // used
else if attrsearch.ItemIndex=18 then
  FindAttribute(Sender, -2) // unused
else if attrsearch.ItemIndex=19 then
  FindAttribute(Sender, -3) // string
else if attrsearch.ItemIndex=20 then
  FindAttribute(Sender, -4) // cluster of a file
else if attrsearch.ItemIndex=21 then
  FindAttribute(Sender, -5) // multiple linked file
else
  FindAttribute(Sender, attrsearch.ItemIndex*16);
end;

procedure TNTFSMFTForm.FindAttribute(Sender: TObject; attrtype: longint);

    function MatchResidentORNonResident(Attr:TAttribute):boolean;
    begin
      result:=false;
      if (withrunlist.State=cbChecked) and (Attr.Hdr.NonResident) then
        result:= true
      else if (withrunlist.State=cbUnChecked) and (not Attr.Hdr.NonResident) then
        result:= true
      else if (withrunlist.State=cbGrayed) then
        result:=true;
    end;

var I,cAttrID:Integer;
    j,cluster:int64;
    k:WORD;
    initital, acounter: cardinal;
    aname:string;
    found,adocount:boolean;
    parent,refcount: Cardinal;
    Name: WideString;
    li:TListItem;
    RunList2: TRunList;

begin
    if tag=1111 then exit;
    tag:=1111;

    // only count
    adocount:=docount.Checked;
    found:=false;

    refcount:=StrToIntDef(Edit1.Text,1);
    j:=CurMFT;
    initital:=CurMFT;

    cAttrID:=-1;
    acounter:=0;

    Tree.Items.Clear;
    List.Clear;
    RunlistView.Clear;
    StatusBar1.Panels[0].Text := IntToStr(J);
    StatusBar1.Panels[1].Text := IntToStr(MFTEntry.FileRecHdr.NumberOfRec);
    StatusBar1.Panels[2].Text := '';
    StatusBar1.Refresh;
    Application.ProcessMessages;

    // go through all MFTentries
    //for J:=MFTNum+1 to NTFS.MaxMFT-1 do
    while true do
    begin
        if ToolButton17.down then Dec(j)
                             else Inc(j);

        if j>=NTFS.MaxMFT then break;
        if j<0 then break;

        if (j mod 100 = 0) then
        begin
          StatusBar1.Panels[0].Text := IntToStr(J);
          if MFTEntry.FileRecHdr.NumberOfRec=j then
          StatusBar1.Panels[1].Text := IntToStr(MFTEntry.FileRecHdr.NumberOfRec);
          StatusBar1.Refresh;
          StatusBar1.Panels[2].Text := IntToStr(acounter);
          StatusBar1.Refresh;
          Application.ProcessMessages;
          if adocount then
            label1.Caption:=inttostr(acounter) + ' records found';
        end;

        if Assigned(MFTEntry) then
          MFTEntry.Free;
        MFTEntry := nil;//just in case line below causes exception
        MFTEntry := NTFS.ReadMFTEntry(J);
        cAttrID:=-1;
        //todo !?!?!?!??!?!!?
        //MFTEntry.NTFS:=NTFS;

        if          (attrtype=-1) and (fInUse in MFTEntry.FileRecHdr.Flags) then
        //if          (attrtype=-1) and (MFTEntry.FileRecHdr.Flags>0) then
        begin
            // used
            if adocount then
              inc(acounter)
            else begin
              found:=true;
              break;
            end;

        end else if (attrtype=-2) and (not (fInUse in MFTEntry.FileRecHdr.Flags)) then
        //end else if (attrtype=-2) and (MFTEntry.FileRecHdr.Flags=0) then
        begin
            // unused
            if adocount then
              inc(acounter)
            else begin
              found:=true;
              break;
            end;

        end else if (attrtype=-4) and (MFTEntry<>nil) then
        begin
            // cluster for a file
            SetLength(Runlist2,0);
            NTFS.ReadRunListWrap(J,Name,RunList2);
            for I:=0 to length(RunList2)-1 do
            begin
              cluster:=StrToInt64Def(Edit1.Text,0);
              if (cluster >= Runlist2[I].Lcn) and
                 (cluster <= Runlist2[I].Lcn+Runlist2[I].Len)
              then begin
                  if adocount then
                    inc(acounter)
                  else begin
                    found:=true;
                    break;
                  end;
              end;
            end;
            //RunList2.
        end else if (attrtype=-3) and (MFTEntry<>nil) then
        begin
            // string search
            aname:=MFTEntry.GetFileName(parent);
            if Pos(lowercase(Edit1.Text),lowercase(aname))>0 then
            begin
              if adocount then
                inc(acounter)
              else begin
                found:=true;
                break;
              end;
            end;
            for I := 0 to MFTEntry.AttribCount - 1 do
            begin
              aname:='';
              if MFTEntry.Attrib[i].Hdr.NameLength > 0 then
              with MFTEntry.Attrib[i].Data do
              begin
                Position := MFTEntry.Attrib[I].Hdr.NameOffset;
                SetLength(Name,MFTEntry.Attrib[I].Hdr.NameLength);
                ReadBuffer(Name[1],MFTEntry.Attrib[I].Hdr.NameLength * 2);
                aname:=WideStringToString(Name,MFTEntry.Attrib[I].Hdr.NameLength);
                cAttrID:=i;

                if Pos(lowercase(Edit1.Text),lowercase(aname))>0 then
                begin
                  attrtype:=MFTEntry.Attrib[i].Hdr.AttributeType;
                  if adocount then
                    inc(acounter)
                  else begin
                    found:=true;
                    break;
                  end;
                end;

              end;

              cAttrID:=-1;
            end;
        end else if (attrtype=-5) and  (MFTEntry<>nil) and (MFTEntry.FileRecHdr.ReferenceCount>refcount) then
        begin
            // multiple linked file
            if adocount then
              inc(acounter)
            else begin
              found:=true;
              break;
            end;

        end else if (cardinal(attrtype)>=0) and (MFTEntry<>nil) then
        begin
            (**-)
            k:=WORD(MFTentry.FileRecHdr.Flags);
            if ( k <> 3 ) and ( k <> 9 ) and ( k <> 1 ) and ( k <> 0 ) then
            begin
                if adocount then
                  inc(acounter)
                else begin
                  found:=true;
                  break;
                end;
            end;
            (**)
            (**)
            // attribute search
            for I := 0 to MFTEntry.AttribCount - 1 do
            begin
              if MatchResidentORNonResident(MFTentry.Attrib[I]) and
                 ( (MFTentry.Attrib[I].Hdr.AttributeType=cardinal(attrtype) ) or
                   (attrtype=0) // any
                 )
              then begin
                if adocount then
                  inc(acounter)
                else begin
                  found:=true;
                  break;
                end;
              end;
            end;
            (**)
        end;

        if found then break;
        if Stop then break;
    end; // for j

    if Assigned(MFTEntry) then
      MFTEntry.Free;
    MFTEntry := nil;//just in case line below causes exception

    if (found) or (stop) then
    begin
      CurMFT:=J;
    end else
      CurMFT:=initital;

    stop:=FALSE;
    tag:=0;
    OnShow(Sender);
    Application.ProcessMessages;
    Sleep(0);

    if found then
    begin
        if (attrtype=-4) then begin
          List.SetFocus;
          li:=List.FindCaption(0,'Reference',true,true,true);
          //List.Set Select
          li.MakeVisible(False);
          li.Selected:=true;
          li.Focused:=true;
          List.Refresh;
        end else if cAttrID>=0 then
        begin
          Tree.Items[0].item[cAttrID].Item[0].Selected:=TRUE;
        end else begin
          for i:=0 to Tree.Items.Count-1 do
            if Pos(NTFSAttribName(attrtype),Tree.Items[i].Text)>0 then
            begin
              Tree.Items[i].Selected:=TRUE;
              break;
            end;
        end;
    end else
      label1.Caption:='not found';

    if adocount then
      label1.Caption:=inttostr(acounter) + ' records found';
end;

procedure TNTFSMFTForm.ShowRunList(Index: Word);
var
 Value: string;
 RunList: TRunList;
 I,len: Integer;
 b,c:Int64;
 RunCnt,RunCnt2,FragCnt,ClusterCnt: Cardinal;
 SparseCnt,SparseClusterCnt,
 LastLcn,LastLen,LastVcn:cardinal;
begin
 FragCnt := 0;
 RunCnt := 0;
 RunCnt2 := 0;
 ClusterCnt := 0;
 RunList := NTFS.ReadRunList(MFTEntry.Attrib[Index]);
 len:=Length(RunList);
 b:=NTFS.BootSec.BytesPerSec;
 c:=NTFS.BootSec.SecPerClust;
 LastLcn:=0;
 LastLen:=0;
  if (len>0) then begin
    LastLcn:=Runlist[0].Lcn;
    LastLen:=Runlist[0].Len;
    LastVcn:=Runlist[0].Vcn;
  end;
 for I := 0 to len-1 do
 begin
   Inc(ClusterCnt,RunList[I].Len);
   if LastLcn+LastLen+1 = RunList[i].Lcn then
   begin
     Inc(LastLen,RunList[I].Len);
   end else begin
     Inc(RunCnt);
     if (doshort.Checked) and (LastLcn+LastLen=RunList[i].Lcn) then
     begin
       Inc(ClusterCnt,RunList[I].Len);
       if RunList[I].Sparse then
       begin
         Inc(SparseCnt);
         Inc(SparseClusterCnt,RunList[I].Len);
       end;
       Inc(LastLen,RunList[I].Len);
     end else begin
       if (LastLcn+LastLen<>RunList[i].Lcn) then
         Inc(FragCnt);
       Value := 'Vcn '+IntToDottetInt(LastVcn)+' - '+IntToDottetInt(LastLen*b*c)+' byte';
       if RunList[I].Sparse then Value :=Value + ' - sparse';
       if RunCnt>1 then AddInfo('  Run' + IntToDottetInt(RunCnt2) + ': ' + 
               IntToDottetInt(LastLcn) + '  -  Len: '+
               IntToDottetInt(LastLen),
               Value,-1);
       if (RunCnt>0) then Inc(RunCnt2);
     end;
     LastLcn:=RunList[I].Lcn;
     LastLen:=RunList[I].Len;
     LastVcn:=RunList[I].Vcn;
   end;
 end;

  if Len>0 then begin
    //Inc(ClusterCnt,len-1);
    if RunList[len-1].Sparse then
    begin
      Inc(SparseCnt);
      Inc(SparseClusterCnt,RunList[len-1].Len);
    end;
    Value := 'Vcn '+IntToDottetInt(RunList[len-1].Vcn)+' - '+IntToDottetInt(RunList[len-1].Len*b*c)+' byte';
    if RunList[len-1].Sparse then Value :=Value + ' - sparse';
    AddInfo('  Run' + IntToDottetInt(RunCnt2+1) + ': ' + 
               IntToDottetInt(RunList[len-1].Lcn) + '  -  Len: '+
               IntToDottetInt(RunList[len-1].Len),
               Value,-1);
  end;

  if FragCnt<>RunCnt then
  AddInfo(IntToDottetInt(FragCnt)+' frags - '+
          IntToDottetInt(RunCnt)+' runs - '+
          IntToDottetInt(ClusterCnt)+' clusters',
          IntToDottetInt(ClusterCnt*b*c)+' byte',0)
  else
  AddInfo(IntToDottetInt(FragCnt)+' frags in '+
          IntToDottetInt(ClusterCnt)+' clusters',
          IntToDottetInt(ClusterCnt*b*c)+' byte',0);
end;

////////////////////////////////////////////////////////////////////////////////
procedure TNTFSMFTForm.ShowRunList2(sname:string);
var I,len: Integer;
    s:WideString;
    Runlist:TRunList;
    descridx: Cardinal;
    Value: string;
    b,c:Int64;
    RunCnt,RunCnt2,FragCnt,ClusterCnt: Cardinal;
    SparseCnt, SparseClusterCnt: Cardinal;
    LastLcn,LastLen,LastVcn:cardinal;
begin
  stop:=false;
  RunlistView.items.Clear;

  if CurMFT>=NTFS.MaxMFT then CurMFT:=NTFS.MaxMFT-1;
  if CurMFT<0            then CurMFT:=0;

  SetLength(Runlist,0);
  NTFS.ReadRunListWrap(CurMFT,s,Runlist);

  SparseClusterCnt := 0;
  descridx := 0;
  FragCnt := 0;
  RunCnt := 0;
  RunCnt2 := 0;
  SparseCnt := 0;
  ClusterCnt := 0;

  len:=Length(Runlist);
  if (len>0) and (RunList[0].desc='') then
    AddRunInfo(sname,-1,-1,-1,-1);

  Application.ProcessMessages;

  if (len>0) then begin
    LastLcn:=Runlist[0].Lcn;
    LastLen:=Runlist[0].Len;
    LastVcn:=Runlist[0].Vcn;
  end;
  b:=NTFS.BootSec.BytesPerSec;
  c:=NTFS.BootSec.SecPerClust;
  AddRunInfo('',0,0,0,-1);
  RunlistView.Refresh;
  for I := 0 to len-1 do
  begin
    AddRunInfo(Runlist[i].desc,Runlist[i].Lcn, Runlist[i].Len,Runlist[i].Vcn,-1);
    inc(FragCnt);
  end;

  if len>0 then begin
    Inc(ClusterCnt,len-1);
    if RunList[len-1].Sparse then
    begin
      Inc(SparseCnt);
      Inc(SparseClusterCnt,RunList[len-1].Len);
      Value := ' - '+IntToDottetInt(LastLen*b*c)+' byte - sparse';
      Inc(LastLen,RunList[len-1].Len);
    end else begin
      Value := ' - '+IntToDottetInt(LastLen*b*c)+' byte';
    end;
    AddRunInfo(RunList[len-1].desc + ' ' + IntToDottetInt(RunCnt2) + Value,
               LastLcn,LastLen,LastVcn,-1);
    RunlistView.Refresh;
  end;

  if (RunlistView.Items.Count>0) and (len>0) then
  begin
   if SparseCnt>0 then begin
    RunlistView.Items[descridx].Caption:=IntToDottetInt(FragCnt)+' frags - ' +
                                         IntToDottetInt(RunCnt-SparseCnt)+' runs - ' +
                                         IntToDottetInt((ClusterCnt-SparseCnt)*b*c)+' byte';
    if RunCnt>1 then
    RunlistView.Items[descridx].SubItems[0]:='-'
    else
    RunlistView.Items[descridx].SubItems[0]:='-';
    RunlistView.Items[descridx].SubItems[1]:=IntToDottetInt(ClusterCnt-SparseClusterCnt);
    RunlistView.Items[descridx].SubItems[2]:='-';
    RunlistView.Refresh;
    (*
    AddRunInfo(IntToStr(RunCnt-SparseCnt)+' frags - '+
               IntToDottetInt((ClusterCnt-SparseCnt)*ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust)+' byte',
               -1,ClusterCnt-SparseClusterCnt,-1,
               descridx+1)
    *)
    //RunlistView.Items[descridx].SubItems[0]:=
    //  IntToStr(RunCnt-SparseCnt)+' frags, '+
    //  IntToStr(ClusterCnt-SparseClusterCnt)+' clusters, '+
    //  IntToDottetInt((ClusterCnt-SparseCnt)*ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust)+' byte ('+
    //  IntToStr(SparseCnt)+' sparse frags in '+
    //  IntToStr(SparseClusterCnt)+' virtual clusters  '+
    //  IntToDottetInt(SparseCnt*ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust)+' byte)'
   end else begin
    RunlistView.Items[0].Caption:=ExtractFileNameCutStream(RunList[0].desc + ' ' +
                                         IntToDottetInt(FragCnt-1)+' frags - ' +
                                         IntToDottetInt(RunCnt)+' runs - ' +
                                         IntToDottetInt(ClusterCnt*b*c)+' byte');
    if RunCnt>1 then
    RunlistView.Items[0].SubItems[0]:='-'
    else
    RunlistView.Items[0].SubItems[0]:='-';
    RunlistView.Items[0].SubItems[1]:=IntToStr(ClusterCnt);
    RunlistView.Items[0].SubItems[2]:='-';
    RunlistView.Refresh;

    (*
    AddRunInfo(IntToStr(RunCnt)+' frags - '+
               IntToDottetInt(ClusterCnt*ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust)+' byte',
               -1,ClusterCnt,
               -1,
               descridx+1);
    *)
    //  IntToStr(RunCnt)+' frags, '+
    //  IntToStr(ClusterCnt)+' clusters ('+
    //  IntToDottetInt(ClusterCnt*ntfs.BootSec.BytesPerSec*ntfs.BootSec.SecPerClust)+' byte)';
   end;
  end else begin
   RunlistView.items.Clear;
  end;
  if visible then RunlistView.Refresh;
end;

procedure TNTFSMFTForm.doshortClick(Sender: TObject);
var id:integer;
begin
  TreeChange(sender,tree.Selected);
  ShowRunList2(StatusBar1.Panels[2].Text+StatusBar1.Panels[1].Text);
end;

procedure TNTFSMFTForm.UpdateNew;
var I: Integer;
    Node: TTreeNode;
    Sub1,Sub2: TTreeNode;
    parent:cardinal;
    pos:longint;
    sdir,sname:string;
    aDir:Boolean;
begin
  if NTFS=nil then exit;

  Screen.Cursor:=crHourGlass;
  //RefreshAttributes;
  Stop:=false;
  Tree.Items.Clear;
  List.Clear;
  RunlistView.Clear;
  RunlistView.Items.BeginUpdate;

  Application.ProcessMessages;
  edit2.Text := inttostr(CurMFT);

  if CurMFT>=NTFS.MaxMFT then CurMFT:=NTFS.MaxMFT-1;
  if CurMFT<0            then CurMFT:=0;
  edit2.Text := inttostr(CurMFT);

  Caption:=NTFS.partname + ' - MFT Record '+IntToStr(CurMFT) + ' / ' + IntToStr(NTFS.MaxMFT);
  Node := Tree.Items.Add(nil,'MFT #' + IntToStr(CurMFT) + ' / ' + IntToStr(NTFS.MaxMFT));
  Node.ImageIndex:=0;

  if Assigned(MFTEntry) then
    MFTEntry.Free;
  MFTEntry := nil;//just in case line below causes exception
  MFTEntry := NTFS.ReadMFTEntry(CurMFT);
  if MFTEntry<>nil then
  begin
    if (MFTEntry.FileRecHdr.FileID = 'BAAD') OR (MFTEntry.FileRecHdr.FileID = 'BAAD') then begin
      Tree.Items.AddChild(Node, 'BAAD');
      Tree.Items[0].Expand(True);
      StatusBar1.Panels[0].Text:='';
      StatusBar1.Panels[1].Text:='';
      StatusBar1.Panels[2].Text:='';

    end else if MFTEntry.AttribCount = 0 then begin
      Tree.Items.AddChild(Node, 'no valid data found');
      Tree.Items[0].Expand(True);
      StatusBar1.Panels[0].Text:='';
      StatusBar1.Panels[1].Text:='';
      StatusBar1.Panels[2].Text:='';

    end else begin
      pos:=MFTEntry.FileRecHdr.OffsetToFirstAttr;
      for I := 0 to MFTEntry.AttribCount - 1 do
      begin
        Sub1 := Tree.Items.AddChild(Node,NTFSAttribName(MFTEntry.Attrib[I].Hdr.AttributeType)+
          '    '+ inttostr(pos)+
          ' (+' +inttostr(MFTEntry.Attrib[I].Hdr.Size)+')');
        pos:=pos+MFTEntry.Attrib[I].Hdr.Size;
        Sub1.ImageIndex := 1;
        Sub1.SelectedIndex := 1;
        Tree.Items[0].Expand(True);
        Tree.Refresh;
        Sub2 := Tree.Items.AddChild(Sub1,'Header');
        Sub2.ImageIndex := 2;
        Sub2.SelectedIndex := 2;
        if MFTEntry.Attrib[I].Hdr.NonResident then
         begin
         Sub2 := Tree.Items.AddChild(Sub1,'Run List');
         Sub2.ImageIndex := 4;
         Sub2.SelectedIndex := 4;
         end;
        if (MFTEntry.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Data) or
           (MFTEntry.Attrib[I].Hdr.AttributeType = NTFS_AttribID_Bitmap)
        then
      	 Sub2 := Tree.Items.AddChild(Sub1,'save as...')
        else
      	 Sub2 := Tree.Items.AddChild(Sub1,'Attribute');
        Sub2.ImageIndex := 3;
        Sub2.SelectedIndex := 3;
        Tree.Items[0].Expand(True);
        Tree.Refresh;
      end;
      Tree.Items[0].Expand(True);
      Tree.Selected := Node;
      StatusBar1.Canvas.Font.Style:=[fsBold];
      StatusBar1.Panels[0].Text  := IntToSize(MFTEntry.GetFileSize);
      StatusBar1.Panels[0].Width := StatusBar1.Canvas.TextWidth(StatusBar1.Panels[0].Text)+60;


      //aDir:=(MFTentry.FileRecHdr.Flags and 2)=2;
      aDir:=fDir in MFTentry.FileRecHdr.Flags;
      parent:=0;
      sdir:=NTFS.GetReferencedDir(CurMFT, '',aDir);
      if RightStr(sdir,1)<>'\' then sdir := sdir + '\';
      //if (GetDirLevel(sdir)>2) then sdir  := ExtractFilePath(sdir+'\')
      //                         else sdir  := ExtractFilePath(sdir);
      sname:=MFTEntry.GetFileName(parent);

      (*
      if (fDir in MFTEntry.FileRecHdr.Flags) or (length(sname)=2) then
        sname:=sname+'\';
      //if fDir in MFTentry.FileRecHdr.Flags;
      *)

      if adir then StatusBar1.Panels[1].Text  := ExtractFileName(sname)+'\'
              else StatusBar1.Panels[1].Text  := ExtractFileName(sname);

      if StatusBar1.Panels[1].Text='' then StatusBar1.Panels[1].Text:=ExtractFileName(sdir);
      StatusBar1.Panels[1].Width := StatusBar1.Canvas.TextWidth(StatusBar1.Panels[1].Text)+60;

      if adir then StatusBar1.Panels[2].Text  := ExtractDirNoLastDir(sdir)+'\'
              else StatusBar1.Panels[2].Text  := sdir;

      StatusBar1.Panels[2].Width := StatusBar1.Canvas.TextWidth(StatusBar1.Panels[2].Text)+60;

      Edit2.SetFocus;
    end;
  end;
  ShowRunList2(StatusBar1.Panels[2].Text+StatusBar1.Panels[1].Text);
  Tree.Refresh;
  RunlistView.Items.EndUpdate;
  RunlistView.Refresh;
  Screen.Cursor:=crDefault;
end;

end.

