unit NTFSProps;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Main, NTFS;

type
  TNTFSPropsForm = class(TForm)
    Button1: TButton;
    BootSecList: TListView;
    Label1: TLabel;
    procedure AddBSInfo(Name, Value: string);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Button1Click(Sender: TObject);
    procedure GetNTFSINFO;
  private
    { Private declarations }
  public
    NTFS: TNTFS;
  end;

var
  NTFSPropsForm: TNTFSPropsForm;

implementation

{$R *.dfm}

procedure TNTFSPropsForm.AddBSInfo(Name, Value: string);
var I: TListItem;
begin
  I := BootSecList.Items.Add;
  I.Caption := Name;
  I.SubItems.Add(Value);
end;

function BootCodeToStr(Code: array of Byte): string;
var I: Integer;
begin
  for I := 0 to 2 do
    Result := Result + IntToHex(Code[I],2) + ' ';
end;

procedure TNTFSPropsForm.FormShow(Sender: TObject);
var
BootSec: TNTFSBootRecord;
begin
if (NTFS=nil) then exit;
BootSec := NTFS.BootSec;
AddBSInfo('--- BOOT_SECTOR ---','');
AddBSInfo('Boot Code',BootCodeToStr(BootSec.BootCode));
//AddBSInfo('OEM Name',BootSec.OemName);
AddBSInfo('Bytes per sector',IntToStr(BootSec.BytesPerSec));
AddBSInfo('Sectors per cluster',IntToStr(BootSec.SecPerClust));
AddBSInfo('Reserved sectors',IntToStr(BootSec.ReservedSectors));
AddBSInfo('Media descriptor',IntToHex(BootSec.MediaDescriptor,2) + ' (hex)');
AddBSInfo('Sectors per track',IntToStr(BootSec.SectorsPerTrack));
AddBSInfo('Numbers of heads',IntToStr(BootSec.NumOfHeads));
AddBSInfo('Hidden sectors',IntToStr(BootSec.HiddenSectors));
AddBSInfo('Sectors in partition',IntToStr(BootSec.SectorSize));
AddBSInfo('Master File Table cluster',IntToStr(BootSec.MFTClust));
AddBSInfo('Master File Table mirror cluster',IntToStr(BootSec.MFTMirrorClust));
AddBSInfo('Clusters per file record',IntToStr(BootSec.ClustersPerFileRec));
AddBSInfo('Clusters per index block',IntToStr(BootSec.ClustersPerIndexBlock));
AddBSInfo('Serial number',IntToHex(BootSec.SerialNumber,16) + ' (hex)');
AddBSInfo('Checksum',FloatToStr(BootSec.Checksum));
AddBSInfo('Boot signature',IntToHex(BootSec.BootSignature,8) + ' (hex)');
NTFS.GetNTFSInfo;
GetNTFSINFO;
end;

procedure TNTFSPropsForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
NTFS.Free;
Release;
end;

procedure TNTFSPropsForm.Button1Click(Sender: TObject);
begin
Close;
end;

procedure TNTFSPropsForm.GetNTFSINFO;
begin
    //  Get ntfs volume data
    AddBSInfo('--- NTFS_VOLUME_DATA ---','');
    AddBSInfo('Serial Number'    ,IntToHex(NTFS.VolumeInfo.SerialNumber.QuadPart,16) + ' (hex)');
    AddBSInfo('Number of Sectors',IntToStr(NTFS.VolumeInfo.NumberOfSectors.QuadPart));
    AddBSInfo('TotalClusters'    ,IntToStr(NTFS.VolumeInfo.TotalClusters.QuadPart));
    AddBSInfo('FreeClusters'     ,IntToStr(NTFS.VolumeInfo.FreeClusters.QuadPart));
    AddBSInfo('TotalReserved'    ,IntToStr(NTFS.VolumeInfo.Reserved.QuadPart));

    AddBSInfo('BytesPerSector'   ,IntToStr(NTFS.VolumeInfo.BytesPerSector));
    AddBSInfo('BytesPerCluster'  ,IntToStr(NTFS.VolumeInfo.BytesPerCluster));
    AddBSInfo('BytesPerMFTRecord',IntToStr(NTFS.VolumeInfo.BytesPerMFTRecord));
    AddBSInfo('ClustersPerMFTRecord',IntToStr(NTFS.VolumeInfo.ClustersPerMFTRecord));

    AddBSInfo('MFTLength'        ,IntToStr(NTFS.VolumeInfo.MFTLength.QuadPart));
    AddBSInfo('MFTStart'         ,IntToStr(NTFS.VolumeInfo.MFTStart.QuadPart));
    AddBSInfo('MFTMirrorStart'   ,IntToStr(NTFS.VolumeInfo.MFTMirrorStart.QuadPart));
    AddBSInfo('MFTZoneStart'     ,IntToStr(NTFS.VolumeInfo.MFTZoneStart.QuadPart));
    AddBSInfo('MFTZoneEnd'       ,IntToStr(NTFS.VolumeInfo.MFTZoneEnd.QuadPart));
end;

end.
