unit NTFSRecovery;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, ExtCtrls, NTFS;

type
  TNTFSRecoverForm = class(TForm)
    Bevel1: TBevel;
    Shape2: TShape;
    NextButton: TButton;
    Notebook1: TNotebook;
    Label3: TLabel;
    Label1: TLabel;
    DetailsButton: TButton;
    DetailMemo: TMemo;
    Label2: TLabel;
    ProgressBar1: TProgressBar;
    SaveDlg: TSaveDialog;
    procedure NextButtonClick(Sender: TObject);
    procedure Recover;
    procedure DetailsButtonClick(Sender: TObject);
  public
    NTFS: TNTFS;
    MFTNum: Cardinal;
    M: TMemoryStream;
  end;

var
  NTFSRecoverForm: TNTFSRecoverForm;

implementation

{$R *.dfm}

procedure TNTFSRecoverForm.NextButtonClick(Sender: TObject);
begin
case Notebook1.PageIndex of
0:
 begin
 NextButton.Enabled := False;
 Recover;
 NextButton.Enabled := True;
 end;
1:
 begin
 NextButton.Caption := 'Save...';
 end;
2:
 begin
 if SaveDlg.Execute then
  begin
  M.SaveToFile(SaveDlg.FileName);
  Close;
  end;
 end;
end;
Notebook1.PageIndex := Notebook1.PageIndex + 1;
end;

procedure TNTFSRecoverForm.Recover;
var
MFTEnt: TMFTEntry;
I: Integer;
B: Byte;
LengthBytes: Byte;
OffsetBytes: Byte;
RunLength: Cardinal;
RunOffset: Cardinal;
parent,OldOffset: Cardinal;
Run: TMemoryStream;
begin
MFTEnt := NTFS.ReadMFTEntry(MFTNum);
DetailMemo.Lines.Add('-Filename: ' + MFTEnt.GetFileName(parent));
DetailMemo.Lines.Add('-Filesize: ' + IntToStr(MFTEnt.GetFileSize) + ' bytes');
I := MFTEnt.IndexOfAttrib(NTFS_AttribID_Data);
if I = -1 then
 begin
 MessageDlg('Sorry, but DiskBuddy is unable to recover this file because file''s master file table record does not contain data attribute.',mtInformation,[mbOK],0);
 Close;
 Exit;
 end;
M := TMemoryStream.Create;
if not MFTEnt.Attrib[I].Hdr.NonResident then
 begin
 DetailMemo.Lines.Add('-File is resident inside MFT record ' + IntToStr(MFTNum) + ' (sectors ' + IntToStr(NTFS.MFTStartSec + MFTNum * 2) + ' - ' + IntToStr(NTFS.MFTStartSec + MFTNum * 2 + 2) + ')');
 MFTEnt.Attrib[I].Data.Position := MFTEnt.Attrib[I].HdrResident.AttributeOffset;
 M.CopyFrom(MFTEnt.Attrib[I].Data,MFTEnt.Attrib[I].HdrResident.AttributeLength);
 end
else
 begin
 DetailMemo.Lines.Add('-File is non-resident, using runlist from MFT record ' + IntToStr(MFTNum) + ' (sectors ' + IntToStr(NTFS.MFTStartSec + MFTNum * 2) + ' - ' + IntToStr(NTFS.MFTStartSec + MFTNum * 2 + 2) + ')');
 ProgressBar1.Max := MFTEnt.GetFileSize;
 with MFTEnt.Attrib[I] do
  begin
  Data.Position := HdrNonResident.DataRunsOffset;
  Data.ReadBuffer(B,1);
  LengthBytes := B and $0F;
  OffsetBytes := (B and $F0) shr 4;
  RunLength := 0;
  RunOffset := 0;
  Data.ReadBuffer(RunLength,LengthBytes);
  Data.ReadBuffer(RunOffset,OffsetBytes);
  while RunLength <> $00 do
   begin
   OldOffset := RunOffset;
   DetailMemo.Lines.Add('-reading clusters ' + IntToStr(RunOffset) + ' - ' + IntToStr(RunOffset + RunLength) + ' (sectors ' + IntToStr(NTFS.BootSec.HiddenSectors + NTFS.BootSec.SecPerClust * RunOffset) + ' - ' + IntToStr(NTFS.BootSec.HiddenSectors + NTFS.BootSec.SecPerClust * (RunOffset + RunLength)) + ')');
   Run := NTFS.ReadClusters(RunOffset,RunLength);
   M.Write(Run.Memory^,Run.Size);
   Run.Free;
   Data.ReadBuffer(B,1);
   LengthBytes := B and $0F;
   OffsetBytes := (B and $F0) shr 4;
   RunLength := 0;
   RunOffset := 0;
   Data.ReadBuffer(RunLength,LengthBytes);
   Data.ReadBuffer(RunOffset,OffsetBytes);
   RunOffset := OldOffset + ToSigned(RunOffset,OffsetBytes);
   ProgressBar1.Position := M.Size;
   Application.ProcessMessages;
   end;
  M.SetSize(HdrNonResident.RealSize);
  end;
 end;
end;

procedure TNTFSRecoverForm.DetailsButtonClick(Sender: TObject);
begin
DetailMemo.Visible := not DetailMemo.Visible;
if DetailMemo.Visible then
DetailsButton.Caption := '<< Hide Details'
else
DetailsButton.Caption := 'Show Details >>';
end;

end.
